package espresso;
final class Jblack_white {

	private static int white_head, white_tail;
	private static int black_head, black_tail;
	private static int forward_link, backward_link;
	private static int[] forward, backward;

	private static int[] stack_head, stack_tail;
	private static int stack_p;

	private static Jset_family BB = new Jset_family();

	private static void alloc_list(final int size) {
		forward = new int[ size ];
		backward = new int[ size ];
		if( null == forward || null == backward ) {
			System.err.println("alloc_list");
			Jmain.closeStreams();
			System.exit( 1 );
		}
	}

	private static void free_list() {
		forward = null;
		backward = null;
	}

	private static void init_list(final int size) {
		for( int i = 0; i < size; i++ ) {
			forward[ i ] = i + 1;
			backward[ i ] = i - 1;
		}
		forward[ size - 1 ] = -1;
		backward[ 0 ] = -1;
		white_head = 0;
		white_tail = size - 1;
		black_head = -1;
		black_tail = -1;
	}

	private static void delete(final int element) {
		forward_link = forward[ element ];
		backward_link = backward[ element ];
		if( forward_link != -1 ) {
			if( backward_link != -1 ) {
				forward[ backward_link ] = forward_link;
				backward[ forward_link ] = backward_link;
			} else{
				white_head = forward_link;
				backward[ forward_link ] = -1;
			}
		} else {
			if( backward_link != -1 ) {
				white_tail = backward_link;
				forward[ backward_link ] = -1;
			}
			else {
				white_head = white_tail = -1;
			}
		}
	}

	private static void insert(final int element) {
		if( black_head != -1 ) {
			forward[ element ] = black_head;
			backward[ element ] = -1;
			backward[ black_head ] = element;
			black_head = element;
		} else {
			black_head = black_tail = element;
			forward[ element ] = backward[ element ] = -1;
		}
	}

	static void merge_list() {
		if( white_head != -1 ) {
			if( black_head != -1 ) {
				forward[ white_tail ] = black_head;
				backward[ black_head ] = white_tail;
				white_tail = black_tail;
				black_head = black_tail = -1;
			}
		} else {
			white_head = black_head;
			white_tail = black_tail;
			black_head = black_tail = -1;
		}
	}

	@SuppressWarnings("boxing")
	private static void print_links(final int size, final int[] list) {
		for( int i = 0; i < size; i++ ) {
			System.out.printf("%d%c", list[ i ], ((i + 1) % 10) != 0 ? '\t' : '\n');
		}
		System.out.printf("\n");
	}

	@SuppressWarnings("boxing")
	static void print_bw(final int size) {
		System.out.printf("white_head %d\twhite_tail %d\tblack_head %d\tblack_tail %d\n",
							white_head, white_tail, black_head, black_tail );
		print_links( size, forward );
		print_links( size, backward );
	}

	private static void alloc_stack(final int size) {
		stack_head = new int[ size ];
		stack_tail = new int[ size ];
		if( null == stack_head || null == stack_tail ) {
			System.err.println("alloc_stack");
			Jmain.closeStreams();
			System.exit( 1 );
		}
	}

	private static void free_stack() {
		stack_head = null;
		stack_tail = null;
	}

	static void push_black_list() {
		stack_head[ stack_p ] = black_head;
		stack_tail[ stack_p++ ] = black_tail;
	}

	static void pop_black_list() {
		black_head = stack_head[ --stack_p ];
		black_tail = stack_tail[ stack_p ];
	}

	private static void clear() {
		stack_p = 0;
	}

	@SuppressWarnings("boxing")
	static void print_stack() {
		System.out.printf("head\n");
		for( int i = stack_p - 1; i >= 0; i-- ) {
			System.out.printf("%d%c", stack_head[ i ], (i % 10) != 0 ? '\t' : '\n');
		}
		System.out.printf("\ntail\n");
		for( int i = stack_p - 1; i >= 0; i-- ) {
			System.out.printf("%d%c", stack_tail[ i ], (i % 10) != 0 ? '\t' : '\n');
		}
		System.out.printf("\n");
	}

	static void setup_bw(final Jset_family R, final int[] c) {
		final int size = R.count;

		/* Allocate memory */
		alloc_list( size );
		alloc_stack( Jglobals.cube.num_binary_vars );
		BB = Jespresso.new_cover( size );
		BB.count = size;
		final int[] out_part_r = Jespresso.new_cube();

		init_list( size );
		clear();
		/* form BB */
		/* Blocking Matrix formed here is bit-complement of that in sigma  */
		final int[] data = R.data;
		for( int r = 0, i = 0; i < R.count; r += R.wsize, i++ ) {
			final int[] b_base = BB.data;
			final int b = BB.wsize * i;
			int last;
			if( (last = Jglobals.cube.inword) != -1 ) {
				/* Check the partial word of binary variables */
				int x = data[ r + last ] & c[ last ];
				x = ~(x | x >>> 1) & Jglobals.cube.inmask;
				b_base[ b + last ] = data[ r + last ] & (x | x << 1);
				/* Check the full words of binary variables */
				for( int w = 1; w < last; w++ ) {
					x = data[ r + w ] & c[ w ];
					x = ~(x | x >>> 1) & Jespresso.DISJOINT;
					b_base[ b + w ] = data[ r + w ] & (x | x << 1);
				}
			}
			Jespresso.PUTLOOP( b_base, b, Jespresso.LOOP( data, r ) );
			Jset.set_and( b_base, b, b_base, b, Jglobals.cube.binary_mask );
			Jset.set_and( out_part_r, Jglobals.cube.mv_mask, data, r );
			if( ! Jset.setp_implies( out_part_r, c ) ) {
				Jset.set_or( b_base, b, b_base, b, out_part_r );
			}
			Jsigma.set_not( b_base, b );
		}
		// out_part_r = null;
	}

	static void free_bw() {
		free_list();
		free_stack();
		Jespresso.free_cover( BB );
	}

	static boolean black_white() {
		for( int b_index = black_head; b_index != -1; b_index = forward[ b_index ] ) {
			boolean containment = false;
			for( int w_index = white_head; w_index != -1; w_index = forward[ w_index ] ) {
				if( Jset.setp_implies( BB.data, BB.wsize * b_index, BB.data, BB.wsize * w_index ) ) {
					containment = true;
					break;
				}
			}
			if( containment == false ) {
				return false;
			}
		}
		return true;
	}

	static void reset_black_list() {
		black_head = black_tail = -1;
	}

	static void split_list(final Jset_family R, final int v) {
		for( int next_index, index = white_head; index != -1; index = next_index ) {
			next_index = forward[ index ];
			if( ! Jespresso.is_in_set( R.data, R.wsize * index, v ) ) {
				delete( index );
				insert( index );
			}
		}
	}

	/* Data Structures for ordering variables in ess_test_and_reduction */
	/** Number of variables currently in the list */
	private static int variable_count;

	/** Next */
	private static int[] variable_forward_chain;

	/** Previous */
	private static int[] variable_backward_chain;

	/** first element in the list */
	private static int variable_head;

	/** last element in the list */
	private static int variable_tail;

	static void variable_list_alloc(final int size) {
		variable_forward_chain = new int[ size ];
		variable_backward_chain = new int[ size ];
		if( null == variable_forward_chain || null == variable_backward_chain ) {
			System.err.println("variable_list_alloc");
			Jmain.closeStreams();
			System.exit( 1 );
		}
	}

	static void variable_list_init(final int reduced_c_free_count, final int[] reduced_c_free_list) {
		variable_count = reduced_c_free_count;
		if( variable_count == 0 ) {
			variable_head = variable_tail = -1;
			return;
		}
		variable_head = reduced_c_free_list[ 0 ];
		variable_tail = reduced_c_free_list[ variable_count - 1 ];
		variable_forward_chain[ variable_tail ] = -1;
		variable_backward_chain[ variable_head ] = -1;

		int next_v = variable_head;
		for( int i = 1; i < variable_count; i++ ) {
			final int v = next_v;
			next_v = reduced_c_free_list[ i ];
			variable_forward_chain[ v ] = next_v;
			variable_backward_chain[ next_v ] = v;
		}
	}

	static void variable_list_delete(final int element) {
		variable_count--;
		forward_link = variable_forward_chain[ element ];
		backward_link = variable_backward_chain[ element ];
		if( forward_link != -1 ) {
			if( backward_link != -1 ) {
				variable_forward_chain[ backward_link ] = forward_link;
				variable_backward_chain[ forward_link ] = backward_link;
			} else {
				variable_head = forward_link;
				variable_backward_chain[ forward_link ] = -1;
			}
		} else {
			if( backward_link != -1 ) {
				variable_tail = backward_link;
				variable_forward_chain[ backward_link ] = -1;
			} else {
				variable_head = variable_tail = -1;
			}
		}
	}

	static void variable_list_insert(final int element) {
		variable_count++;
		if( variable_head != -1 ) {
			variable_forward_chain[ element ] = variable_head;
			variable_backward_chain[ element ] = -1;
			variable_backward_chain[ variable_head ] = element;
			variable_head = element;
		} else {
			variable_head = variable_tail = element;
			variable_forward_chain[ element ] = -1;
			variable_backward_chain[ element ] = -1;
		}
	}

	static boolean variable_list_empty() {
		return variable_count != 0 ? false : true;
	}

	static void get_next_variable(final int[] pv, final int[] pphase, final Jset_family R) {
		int max_variable = -1, max_phase = -1;// java fix may be not initialized

		int max_black_count = -1;
		for( int v = variable_head; v != -1; v = variable_forward_chain[ v ] ) {
			final int e0 = v << 1;
			final int e1 = e0 + 1;
			int e0_black_count = 0;
			int e1_black_count = 0;
			for( int w_index = white_head; w_index != -1; w_index = forward[ w_index ] ) {
				final int[] r_base = R.data;
				final int r = R.wsize * w_index;
				if( Jespresso.is_in_set( r_base, r, e0 ) ) {
					if( Jespresso.is_in_set( r_base, r, e1 ) ) {
						continue;
					}// else {
						e1_black_count++;
					//}
				} else {
					e0_black_count++;
				}
			}
			if( e0_black_count > e1_black_count ) {
				if( e0_black_count > max_black_count ) {
					max_black_count = e0_black_count;
					max_variable = v;
					max_phase = 0;
				}
			} else {
				if( e1_black_count > max_black_count ) {
					max_black_count = e1_black_count;
					max_variable = v;
					max_phase = 1;
				}
			}
		}
		pv[ 0 ] = max_variable;
		pphase[ 0 ] = max_phase;
	}

	static void print_variable_list() {
		System.out.printf("Variable_Forward_Chain:\n");
		print_links( Jglobals.cube.num_binary_vars, variable_forward_chain );
		System.out.printf("Variable_Backward_Chain:\n");
		print_links( Jglobals.cube.num_binary_vars, variable_backward_chain );
	}
}