package espresso;

/* Module:canonical.c
 *	contains routines for finding the canonical cover of the
 *	incompletely specified logic function.
 * Routine:
 * pcover find_canonical_cover():
 *	Finds canonical cover of the incompletely specified logic function
 *	by iteratively calling ess_test_and_reduction for each cube in the
 *	ON-set.
 */

final class Jcanonical {
	/**
	 * find_canonical_cover
	 * Objective: find canonical cover of the essential signature cube
	 * Input:
	 *	F: ONSET cover;
	 *	D: DC cover;
	 *	R: OFFSET cover;
	 * Output:
	 *	Return canonical cover of the essential signature cube
	 */
	static Jset_family find_canonical_cover(final Jset_family F1,final Jset_family D, final Jset_family R) {
		final Jset_family F = Jset.sf_save( F1 );
		final Jset_family E = Jespresso.new_cover( D.count );
		E.count = D.count;
		Jset.sf_copy( E, D );

		final Jset_family ESC = Jespresso.new_cover( F.count );

		while( F.count != 0 ) {
			int[] c_base = F.data;
			final int c = F.wsize * --F.count;
			Jespresso.RESET( c_base, c, Jespresso.NONESSEN );
			Jset[] extended_dc = Jcofactor.cube2list( E, F );
			int[] d = Jreduce.reduce_cube( extended_dc, c_base, c );
			extended_dc = null;
			if( Jset.setp_empty( d ) ) {
				d = null;
				continue;
			}
			c_base = Jsigma.get_sigma( R, d, 0 );
			final long t = System.currentTimeMillis();
			final Jset_family COVER = Jessentiality.etr_order( F, E, R, c_base, d );
			Jutil_signature.s_totals( t, Jespresso.ETR_TIME );
			d = null;
			if( Jespresso.TESTP( c_base, c, Jespresso.NONESSEN ) ) {
				Jset.sf_append( F, COVER );
			} else {
				Jespresso.free_cover( COVER );
				Jset.sf_addset( E, c_base, c );
				Jset.sf_addset( ESC, c_base, c );
			}
			c_base = null;
		}
		Jespresso.free_cover( F );
		Jespresso.free_cover( E );

		return ESC;
	}
}