package espresso;
import java.io.PrintStream;

final class Jcols {

	/** allocate a new col vector */
	static Jsm_col sm_col_alloc() {
		Jsm_col pcol;

if( Jespresso.FAST_AND_LOOSE ) {
		if( Jmatrix.sm_col_freelist == null ) {
			pcol = new Jsm_col();
		} else {
			pcol = Jmatrix.sm_col_freelist;
			Jmatrix.sm_col_freelist = pcol.next_col;
		}
} else {
		pcol = new Jsm_col();
}

		pcol.col_num = 0;
		pcol.length = 0;
		pcol.first_row = pcol.last_row = null;
		pcol.next_col = pcol.prev_col = null;
		pcol.flag = false;
		pcol.user_word = null;		/* for our user ... */
		return pcol;
	}

	/**
	 *  free a col vector -- for FAST_AND_LOOSE, this is real cheap for cols;
	 *  however, freeing a rowumn must still walk down the rowumn discarding
	 *  the elements one-by-one; that is the only use for the extra '-DCOLS'
	 *  compile flag ...
	 */
	static void sm_col_free(Jsm_col pcol) {
if( Jespresso.FAST_AND_LOOSE && ! Jespresso.COLS ) {
		if( pcol.first_row != null ) {
			/* Add the linked list of col items to the free list */
			pcol.last_row.next_row = Jmatrix.sm_element_freelist;
			Jmatrix.sm_element_freelist = pcol.first_row;
		}

		/* Add the col to the free list of cols */
		pcol.next_col = Jmatrix.sm_col_freelist;
		Jmatrix.sm_col_freelist = pcol;
} else {
		for( Jsm_element pnext, p = pcol.first_row; p != null; p = pnext ) {
			pnext = p.next_row;
			Jmatrix.sm_element_free( p );
		}
		pcol = null;
}
	}

	/** duplicate an existing col */
	static Jsm_col sm_col_dup(final Jsm_col pcol) {
		final Jsm_col pnew = sm_col_alloc();
		for( Jsm_element p = pcol.first_row; p != null; p = p.next_row ) {
			sm_col_insert( pnew, p.row_num );
		}
		return pnew;
	}


	/** insert an element into a col vector */
	static Jsm_element sm_col_insert(final Jsm_col pcol, final int row) {
		/* get a new item, save its address */
		final Jsm_element element = Jmatrix.sm_element_alloc();
		Jsm_element test = element;
		if( pcol.last_row == null ) {
			test.row_num = row;
			pcol.first_row = test;
			pcol.last_row = test;
			test.next_row = null;
			test.prev_row = null;
			pcol.length++;
		} else if( pcol.last_row.row_num < row ) {
			test.row_num = row;
			pcol.last_row.next_row = test;
			test.prev_row = pcol.last_row;
			pcol.last_row = test;
			test.next_row = null;
			pcol.length++;
	    } else if( pcol.first_row.row_num > row ) {
			test.row_num = row;
			pcol.first_row.prev_row = test;
			test.next_row = pcol.first_row;
			pcol.first_row = test;
			test.prev_row = null;
			pcol.length++;
	    } else {
	    	Jsm_element p;
	    	for( p = pcol.first_row; p.row_num < row; p = p.next_row ) {
				;
			}
			if( p.row_num > row ) {
			    test.row_num = row;
			    p = p.prev_row;
			    p.next_row.prev_row = test;
			    test.next_row = p.next_row;
			    p.next_row = test;
			    test.prev_row = p;
			    pcol.length++;
			} else {
			    test = p;
			}
	    }

		/* if item was not used, free it */
		if( element != test ) {
			Jmatrix.sm_element_free( element );
		}

		/* either way, return the current new value */
		return test;
	}


	/** remove an element from a col vector */
	static void sm_col_remove(final Jsm_col pcol, final int row) {
		Jsm_element p;

		for( p = pcol.first_row; p != null && p.row_num < row; p = p.next_row ) {
			;
		}
		if( p != null && p.row_num == row ) {
			if( p.prev_row == null ) {
				pcol.first_row = p.next_row;
			} else {
				p.prev_row.next_row = p.next_row;
			}
			if( p.next_row == null ) {
				pcol.last_row = p.prev_row;
			} else {
				p.next_row.prev_row = p.prev_row;
			}
			pcol.length--;
			Jmatrix.sm_element_free( p );
		}
	}


	/** find an element (if it is in the col vector) */
	static Jsm_element sm_col_find(final Jsm_col pcol, final int row) {
		Jsm_element p;

		for( p = pcol.first_row; p != null && p.row_num < row; p = p.next_row ) {
			;
		}
		if( p != null && p.row_num == row ) {
			return p;
		}
		return null;
	}

	/** return 1 if col p2 contains col p1; 0 otherwise */
	static boolean sm_col_contains(final Jsm_col p1, final Jsm_col p2) {
		Jsm_element q1 = p1.first_row;
		Jsm_element q2 = p2.first_row;
		while( q1 != null ) {
			if( q2 == null || q1.row_num < q2.row_num ) {
				return false;
			} else if( q1.row_num == q2.row_num ) {
				q1 = q1.next_row;
				q2 = q2.next_row;
			} else {
				q2 = q2.next_row;
			}
		}
		return true;
	}


	/** return 1 if col p1 and col p2 share an element in common */
	static boolean sm_col_intersects(final Jsm_col p1, final Jsm_col p2) {
		Jsm_element q1 = p1.first_row;
		Jsm_element q2 = p2.first_row;
		if( q1 == null || q2 == null ) {
			return false;
		}
		for( ; ; ) {
			if( q1.row_num < q2.row_num ) {
				if( (q1 = q1.next_row) == null ) {
					return false;
				}
			} else if( q1.row_num > q2.row_num ) {
				if( (q2 = q2.next_row) == null ) {
					return false;
				}
			} else {
				return true;
			}
		}
	}

	/** compare two cols, lexical ordering */
	static int Jsm_col_compare(final Jsm_col p1, final Jsm_col p2) {
		Jsm_element q1 = p1.first_row;
		Jsm_element q2 = p2.first_row;
		while( q1 != null && q2 != null ) {
			if( q1.row_num != q2.row_num ) {
				return q1.row_num - q2.row_num;
			}
			q1 = q1.next_row;
			q2 = q2.next_row;
		}

		if( q1 != null ) {
			return 1;
		} else if( q2 != null ) {
			return -1;
		} else {
			return 0;
		}
	}

	/** return the intersection */
	static Jsm_col sm_col_and(final Jsm_col p1, final Jsm_col p2) {
		final Jsm_col result = sm_col_alloc();
		Jsm_element q1 = p1.first_row;
		Jsm_element q2 = p2.first_row;
		if( q1 == null || q2 == null ) {
			return result;
		}
		for( ; ; ) {
			if( q1.row_num < q2.row_num ) {
				if( (q1 = q1.next_row) == null ) {
					return result;
				}
			} else if( q1.row_num > q2.row_num ) {
				if( (q2 = q2.next_row) == null ) {
					return result;
				}
			} else {
				sm_col_insert( result, q1.row_num );
				if( (q1 = q1.next_row) == null ) {
					return result;
				}
				if( (q2 = q2.next_row) == null ) {
					return result;
				}
			}
		}
	}

	static int sm_col_hash(final Jsm_col pcol, final int modulus) {
		int sum = 0;
		for( Jsm_element p = pcol.first_row; p != null; p = p.next_row ) {
			sum = (sum * 17 + p.row_num) % modulus;
		}
		return sum;
	}

	/** remove an element from a col vector (given a pointer to the element) */
	static void sm_col_remove_element(final Jsm_col pcol, final Jsm_element p) {
		if( p.prev_row == null ) {
			pcol.first_row = p.next_row;
		} else {
			p.prev_row.next_row = p.next_row;
		}
		if( p.next_row == null ) {
			pcol.last_row = p.prev_row;
		} else {
			p.next_row.prev_row = p.prev_row;
		}
		pcol.length--;
		Jmatrix.sm_element_free( p );
	}

	@SuppressWarnings("boxing")
	static void sm_col_print(final PrintStream fp, final Jsm_col pcol) {
		for( Jsm_element p = pcol.first_row; p != null; p = p.next_row ) {
			fp.printf(" %d", p.row_num);
		}
	}
}