package espresso;

/**
 *  The cube structure is a global structure which contains information
 *  on how a set maps into a cube -- i.e., number of parts per variable,
 *  number of variables, etc.  Also, many fields are pre-computed to
 *  speed up various primitive operations.
 */
final class Jcube_struct {
	static final int CUBE_TEMP = 10;

	/** set size of a cube */
	int size;

	/** number of variables in a cube */
	int num_vars;

	/** number of binary variables */
	int num_binary_vars;

	/** first element of each variable */
	int[] first_part;

	/** first element of each variable */
	int[] last_part;

	/** number of elements in each variable */
	int[] part_size;

	/** first word for each variable */
	int[] first_word;

	/** last word for each variable */
	int[] last_word;

	/** Mask to extract binary variables */
	int[] binary_mask;

	/** mask to get mv parts */
	int[] mv_mask;

	/** mask to extract a variable */
	int[][] var_mask;

	/** an array of temporary sets */
	int[][] temp;

	/** a full cube */
	int[] fullset;

	/** an empty cube */
	int[] emptyset;

	/** mask to get odd word of binary part */
	int inmask;

	/** which word number for above */
	int inword;

	/** should this variable be sparse? */
	boolean[] sparse;

	/** number of multiple-valued variables */
	int num_mv_vars;

	/** which variable is "output" (-1 if none) */
	int output;
}
