package espresso;
/*
    Module: cubestr.c -- routines for managing the global cube structure
*/

final class Jcubestr {

	/**
	    cube_setup -- assume that the fields "num_vars", "num_binary_vars", and
	    part_size[num_binary_vars .. num_vars-1] are setup, and initialize the
	    rest of cube and cdata.

	    If a part_size is < 0, then the field size is abs(part_size) and the
	    field read from the input is symbolic.
	*/
	static void cube_setup() {
		final Jcube_struct cube = Jglobals.cube;// java

		if( cube.num_binary_vars < 0 || cube.num_vars < cube.num_binary_vars ) {
			Jcvrmisc.fatal("cube size is silly, error in .i/.o or .mv");
		}

		cube.num_mv_vars = cube.num_vars - cube.num_binary_vars;
		cube.output = cube.num_mv_vars > 0 ? cube.num_vars - 1 : -1;

		cube.size = 0;
		cube.first_part = new int[ cube.num_vars ];
		cube.last_part = new int[ cube.num_vars ];
		cube.first_word = new int[ cube.num_vars ];
		cube.last_word = new int[ cube.num_vars ];
		for( int var = 0; var < cube.num_vars; var++ ) {
			if( var < cube.num_binary_vars ) {
				cube.part_size[ var ] = 2;
			}
			cube.first_part[ var ] = cube.size;
			cube.first_word[ var ] = Jespresso.WHICH_WORD( cube.size );
			cube.size += Math.abs( cube.part_size[ var ] );
			cube.last_part[ var ] = cube.size - 1;
			cube.last_word[ var ] = Jespresso.WHICH_WORD( cube.size - 1 );
		}

		cube.var_mask = new int[ cube.num_vars ][];
		cube.sparse = new boolean[ cube.num_vars ];
		cube.binary_mask = Jespresso.new_cube();
		cube.mv_mask = Jespresso.new_cube();
		for( int var = 0; var < cube.num_vars; var++ ) {
			cube.var_mask[ var ] = Jespresso.new_cube();
			final int[] p = cube.var_mask[ var ];
			for( int i = cube.first_part[ var ]; i <= cube.last_part[ var ]; i++ ) {
				Jespresso.set_insert( p, i );
			}
			if( var < cube.num_binary_vars ) {
				final int[] m = cube.binary_mask;
				Jset.set_or( m, m, p );
				cube.sparse[ var ] = false;
			} else {
				final int[] m = cube.mv_mask;
				Jset.set_or( m, m, p );
				cube.sparse[ var ] = true;
			}
		}
		if( cube.num_binary_vars == 0 ) {
			cube.inword = -1;
		} else {
			cube.inword = cube.last_word[ cube.num_binary_vars - 1 ];
			cube.inmask = cube.binary_mask[ cube.inword ] & Jespresso.DISJOINT;
		}

		cube.temp = new int[ Jcube_struct.CUBE_TEMP ][];
		for( int i = 0; i < Jcube_struct.CUBE_TEMP; i++ ) {
			cube.temp[ i ] = Jespresso.new_cube();
		}
		cube.fullset = Jespresso.new_cube();
		Jset.set_fill( cube.fullset, cube.size );
		cube.emptyset = Jespresso.new_cube();

		final Jcdata_struct cdata = Jglobals.cdata;// java
		cdata.part_zeros = new int[ cube.size ];
		cdata.var_zeros = new int[ cube.num_vars ];
		cdata.parts_active = new int[ cube.num_vars ];
		cdata.is_unate = new boolean[ cube.num_vars ];
	}

	/**
	setdown_cube -- free memory allocated for the cube/cdata structs
	(free's all but the part_size array)

	(I wanted to call this cube_setdown, but that violates the 8-character
	external routine limit on the IBM !)
	*/
	static void setdown_cube() {
		final Jcube_struct cube = Jglobals.cube;// java
		final Jcdata_struct cdata = Jglobals.cdata;// java

		cube.first_part = cube.last_part = null;
		cube.first_word = cube.last_word = null;
		cube.sparse = null;
		cube.binary_mask = cube.mv_mask = null;
		cube.fullset = cube.emptyset = null;
		cube.var_mask = cube.temp = null;

		cdata.part_zeros = cdata.var_zeros = cdata.parts_active = null;
		cdata.is_unate = null;
	}

	static void save_cube_struct() {
		final Jcube_struct cube = Jglobals.cube;// java
		final Jcdata_struct cdata = Jglobals.cdata;// java

		Jglobals.temp_cube_save = cube;              /* structure copy ! */
		Jglobals.temp_cdata_save = Jglobals.cdata;            /*      ""          */

		cube.first_part = cube.last_part = null;
		cube.first_word = cube.last_word = null;
		cube.part_size = null;
		cube.binary_mask = cube.mv_mask = null;
		cube.fullset = cube.emptyset = null;
		cube.var_mask = cube.temp = null;

		cdata.part_zeros = cdata.var_zeros = cdata.parts_active = null;
		cdata.is_unate = null;
	}

	static void restore_cube_struct() {
		Jglobals.cube = Jglobals.temp_cube_save;              /* structure copy ! */
		Jglobals.cdata = Jglobals.temp_cdata_save;            /*      ""          */
	}
}