package espresso;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.io.PushbackInputStream;

/*
    module: cvrin.c
    purpose: cube and cover input routines
*/

final class Jcvrin {

	private static boolean line_length_error;
	private static int lineno;

	static void skip_line(final InputStream fpin, final PrintStream fpout, final boolean echo) {
		try {// TODO java: throw exception
			int ch;
			while( (ch = fpin.read()) != -1 && ch != '\n') {
				if( echo ) {
					fpout.print( (char)ch );
				}
			}
		} catch(final IOException ie) {
		}
		if( echo ) {
			fpout.print('\n');
		}
		lineno++;
	}

	static String get_word(final InputStream fp, final StringBuilder word) {
		word.setLength( 0 );
		try {// TODO java: throw exception
			int ch;
			while( (ch = fp.read()) != -1 && Character.isWhitespace( ch ) ) {
				;
			}
			word.append( (char)ch );
			while( (ch = fp.read()) != -1 && ! Character.isWhitespace( ch ) ) {
				word.append( (char)ch );
			}
			// word.append( '\0' );
		} catch(final IOException ie) {
		}
		return word.toString();
	}

	// java doesn't have sscanf
	/**
	 * Reads an integer value from an text input stream.
	 *
	 * @param fp the input stream to read.
	 * @return the integer value.
	 * @throws IOException an error occurs while reading or parsing.
	 */
	private static int scanInt(final StringBuilder s) throws IOException {
		final StringBuilder sb = new StringBuilder();
		int i = 0;
		while( i < s.length() && Character.isWhitespace( s.charAt( i ) ) ) {
			i++;
		}
		while( i < s.length() ) {
			final char ch = s.charAt( i );
			if( Character.isWhitespace( ch ) ) {
				break;
			}
			sb.append( i );
		}
		try {
			return Integer.parseInt( sb.toString() );
		} catch(final NumberFormatException ne) {
			throw new IOException( ne.getMessage() );
		}
		// return 0;
	}

	/**
	 * Reads a string from an text input stream.
	 *
	 * @param fp the input stream to read.
	 * @return the read text data.
	 * @throws IOException an error occurs while reading or parsing.
	 */
	private static String scanString(final InputStream fp, final StringBuilder sb) throws IOException {
		sb.setLength( 0 );
		int ch;
		while( (ch = fp.read()) != -1 && Character.isWhitespace( ch ) ) {
			;
		}
		sb.append( (char)ch );
		while( (ch = fp.read()) != -1 && ! Character.isWhitespace( ch ) ) {
			sb.append( (char)ch );
		}
		return sb.toString();
	}
	/**
	 * Reads and checks a specified string form an input stream.
	 * For example, to parse strings like: "var=%d"
	 *
	 * @param fp the input stream.
	 * @param s the string to check
	 * @throws IOException an error occurs while reading or comparing.
	 */
	private static void skipString(final InputStream fp, final String s) throws IOException {
		int ch;
		while( (ch = fp.read()) != -1 && Character.isWhitespace( ch ) ) {
			;
		}
		int i = 0;
		while( (ch = fp.read()) != -1 ) {
			if( i >= s.length() ) {
				return;
			}
			if( s.charAt( i++ ) != ch ) {
				break;
			}
		}
		throw new IOException();
	}

	// label bad_char
	@SuppressWarnings("boxing")
	private static void badChar(final InputStream fp) {
		System.err.printf("(warning): input line #%d ignored\n", lineno);
		skip_line( fp, System.out, true );
	}

	/**
	 *  Yes, I know this routine is a mess
	 */
	@SuppressWarnings("boxing")
	static void read_cube(final InputStream fp, final JpPLA PLA) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int[] cf = cube.temp[0];
		final int[] cr = cube.temp[1];
		final int[] cd = cube.temp[2];
		boolean savef = false, saved = false, saver = false;
		final StringBuilder token = new StringBuilder();/* for kiss read hack */

		Jset.set_clear( cf, cube.size );

		/* Loop and read binary variables */
		try {
			int var;
			for( var = 0; var < cube.num_binary_vars; var++) {
				switch( fp.read() ) {
				case -1:
					badChar( fp );
					return;
				case '\n':
					if( ! line_length_error ) {
						// System.err.printf("product term(s) %s\nspan more than one line (warning only)");
						System.err.printf("product term(s) span more than one line (warning only)");
					}
					line_length_error = true;
					lineno++;
					var--;
					break;
				case ' ': case '|': case '\t':
					var--;
					break;
				case '2': case '-':
					Jespresso.set_insert( cf, var * 2 + 1 );
				case '0':
					Jespresso.set_insert( cf, var * 2 );
					break;
				case '1':
					Jespresso.set_insert( cf, var * 2 + 1 );
					break;
				case '?':
					break;
				default:
					badChar( fp );
					return;
				}
			}

			/* Loop for the all but one of the multiple-valued variables */
			for( var = cube.num_binary_vars; var < cube.num_vars - 1; var++ ) {
				if( cube.part_size[ var ] < 0 ) {
					final String str = scanString( fp, token );
					if( str.equals("-") || str.equals("ANY") ) {
						if( Jglobals.kiss && var == cube.num_vars - 2 ) {
							/* leave it empty */
						} else {
							/* make it full */
							Jset.set_or( cf, cf, cube.var_mask[var] );
						}
					} else if( str.equals("~") ) {
						;
						/* leave it empty ... (?) */
					} else {
						int i, varx, offset;/* for kiss read hack */
						if( Jglobals.kiss && var == cube.num_vars - 2 ) {
							varx = var - 1;
							offset = Math.abs( cube.part_size[var - 1] );
						} else {
							varx = var;
							offset = 0;
						}
						/* Find the symbolic label in the label table */
						final int first = cube.first_part[ varx ];
						final int last = cube.last_part[ varx ];
						for( i = first; i <= last; i++ ) {
							if( PLA.label[i] == null ) {
								PLA.label[i] = str;	/* add new label */
								Jespresso.set_insert( cf, i + offset );
								break;
							} else if( PLA.label[i].equals( str ) ) {
								Jespresso.set_insert( cf, i + offset );	/* use column i */
								break;
							}
						}
						if( i > last ) {
							System.err.printf("declared size of variable %d (counting from variable 0) is too small\n", var);
							Jmain.closeStreams();
							System.exit( -1 );
						}
					}

				} else {
					for( int i = cube.first_part[var]; i <= cube.last_part[var]; i++ ) {
						switch( fp.read() ) {
						case -1:
							badChar( fp );
							return;
						case '\n':
							if( ! line_length_error ) {
								System.err.printf("product term(s) %s\nspan more than one line (warning only)");
							}
							line_length_error = true;
							lineno++;
							i--;
							break;
						case ' ': case '|': case '\t':
							i--;
							break;
						case '1':
							Jespresso.set_insert( cf, i );
						case '0':
							break;
						default:
							badChar( fp );
							return;
						}
					}
				}
			}

			/* Loop for last multiple-valued variable */
			if( Jglobals.kiss ) {
				saver = savef = true;
				Jset.set_xor( cr, cf, cube.var_mask[ cube.num_vars - 2 ] );
			} else {
				Jset.set_copy( cr, cf );
			}
			Jset.set_copy( cd, cf );
			for( int i = cube.first_part[ var ]; i <= cube.last_part[ var ]; i++ ) {
				switch( fp.read() ) {
				case -1:
					badChar( fp );
					return;
				case '\n':
					if( ! line_length_error ) {
						// System.err.printf("product term(s) %s\nspan more than one line (warning only)");
						System.err.printf("product term(s) span more than one line (warning only)");
					}
					line_length_error = true;
					lineno++;
					i--;
					break;
				case ' ': case '|': case '\t':
					i--;
					break;
				case '4': case '1':
					if( (PLA.pla_type & Jespresso.F_type) != 0 ) {
						Jespresso.set_insert( cf, i );
						savef = true;
					}
					break;
				case '3': case '0':
					if( (PLA.pla_type & Jespresso.R_type) != 0 ) {
						Jespresso.set_insert( cr, i );
						saver = true;
					}
					break;
				case '2': case '-':
					if( (PLA.pla_type & Jespresso.D_type) != 0 ) {
						Jespresso.set_insert( cd, i );
						saved = true;
					}
				case '~':
					break;
				default:
					badChar( fp );
					return;
				}
			}
			if( savef ) {
				PLA.F = Jset.sf_addset( PLA.F, cf );
			}
			if( saved ) {
				PLA.D = Jset.sf_addset( PLA.D, cd );
			}
			if( saver ) {
				PLA.R = Jset.sf_addset( PLA.R, cr );
			}
		} catch(final IOException ie) {
			badChar( fp );
			return;
		}
/*		return;

bad_char:
		System.err.printf("(warning): input line #%d ignored\n", lineno);
		skip_line( fp, System.out, true );
		return; */
	}

	@SuppressWarnings("boxing")
	static void parse_pla(final PushbackInputStream fp, final JpPLA PLA) {
		final Jcube_struct cube = Jglobals.cube;// java;
		final StringBuilder word = new StringBuilder();

		lineno = 1;
		line_length_error = false;

// loop:
		do {
			try {
				int ch;
				switch( ch = fp.read() ) {
				case -1:
					return;

				case '\n':
					lineno++;

				case ' ': case '\t': case '\f': case '\r':
					break;

				case '#':
					fp.unread( ch );
					skip_line( fp, System.out, Jglobals.echo_comments );
					break;

				case '.':
					/* .i gives the cube input size (binary-functions only) */
					String str = get_word( fp, word );// java
					if( str.equals("i") ) {
						if( cube.fullset != null ) {
							System.err.printf("extra .i ignored\n");
							skip_line( fp, System.out, /* echo */ false );
						} else {
							try {
								cube.num_binary_vars = Jsupport.scanInt( fp, 10 );
							} catch( final IOException e ) {
								Jcvrmisc.fatal("error reading .i");
							}
							cube.num_vars = cube.num_binary_vars + 1;
							cube.part_size = new int[ cube.num_vars ];
						}

						/* .o gives the cube output size (binary-functions only) */
					} else if( str.equals("o") ) {
						if( cube.fullset != null ) {
							System.err.printf("extra .o ignored\n");
							skip_line( fp, System.out, /* echo */ false );
						} else {
							if( cube.part_size == null ) {
								Jcvrmisc.fatal(".o cannot appear before .i");
							}
							try {
								cube.part_size[ cube.num_vars - 1 ] = Jsupport.scanInt( fp, 10 );
							} catch( final IOException e ) {
								Jcvrmisc.fatal("error reading .o");
							}
							Jcubestr.cube_setup();
							PLA_labels( PLA );
						}

						/* .mv gives the cube size for a multiple-valued function */
					} else if( str.equals("mv") ) {
						if( cube.fullset != null ) {
							System.err.printf("extra .mv ignored\n");
							skip_line( fp, System.out, /* echo */ false );
						} else {
							if( cube.part_size != null ) {
								Jcvrmisc.fatal("cannot mix .i and .mv");
							}
							try {
								cube.num_vars = Jsupport.scanInt( fp, 10 );
								cube.num_binary_vars = Jsupport.scanInt( fp, 10 );
							} catch( final IOException e ) {
								Jcvrmisc.fatal("error reading .mv");
							}
							if( cube.num_binary_vars < 0 ) {
								Jcvrmisc.fatal("num_binary_vars (second field of .mv) cannot be negative");
							}
							if( cube.num_vars < cube.num_binary_vars ) {
								Jcvrmisc.fatal("num_vars (1st field of .mv) must exceed num_binary_vars (2nd field of .mv)");
							}
							cube.part_size = new int[ cube.num_vars ];
							for( int var = cube.num_binary_vars; var < cube.num_vars; var++ ) {
								try {
									cube.part_size[var] = Jsupport.scanInt( fp, 10 );
								} catch( final IOException e ) {
									Jcvrmisc.fatal("error reading .mv");
								}
							}
							Jcubestr.cube_setup();
							PLA_labels( PLA );
						}

						/* .p gives the number of product terms -- we ignore it */
					} else if( str.equals("p") ) {
						Jsupport.scanInt( fp, 10 );
					} else if( str.equals("e") || str.equals("end") ) {
						return;
					} else if( str.equals("kiss") ) {
						Jglobals.kiss = true;
					} else if( str.equals("type") ) {
						str = get_word( fp, word );
						int i;
						for( i = 0; i < Jglobals.pla_types.length; i++ ) {
							if( Jglobals.pla_types[i].key.endsWith( str ) ) {
								PLA.pla_type = Jglobals.pla_types[i].value;
								break;
							}
						}
						if( i >= Jglobals.pla_types.length ) {
							Jcvrmisc.fatal("unknown type in .type command");
						}
						/* parse the labels */
					} else if( str.equals("ilb") ) {
						if( cube.fullset == null ) {
							Jcvrmisc.fatal("PLA size must be declared before .ilb or .ob");
						}
						if( PLA.label == null ) {
							PLA_labels( PLA );
						}
						for( int var = 0; var < cube.num_binary_vars; var++ ) {
							str = get_word( fp, word );
							final int i = cube.first_part[ var ];
							PLA.label[ i + 1 ] = str;
							PLA.label[ i ] = String.format( "%s.bar", str );
						}
					} else if( str.equals("ob") ) {
						if( cube.fullset == null ) {
							Jcvrmisc.fatal("PLA size must be declared before .ilb or .ob");
						}
						if( PLA.label == null ) {
							PLA_labels( PLA );
						}
						final int var = cube.num_vars - 1;
						for( int i = cube.first_part[var]; i <= cube.last_part[var]; i++ ) {
							str = get_word( fp, word );
							PLA.label[i] = str;
						}
						/* .label assigns labels to multiple-valued variables */
					} else if( str.equals("label") ) {
						if( cube.fullset == null ) {
							Jcvrmisc.fatal("PLA size must be declared before .label");
						}
						if( PLA.label == null ) {
							PLA_labels( PLA );
						}
						int var;
						try {
							skipString( fp, "var=" );// java fscanf(fp, "var=%d", &var)
							var = Jsupport.scanInt( fp, 10 );
							for( int i = cube.first_part[var]; i <= cube.last_part[var]; i++ ) {
								str = get_word( fp, word );
								PLA.label[i] = str;
							}
						} catch(final IOException ie) {
							Jcvrmisc.fatal("Error reading labels");
						}
					} else if( str.equals("symbolic") ) {
						final Jsymbolic_t[] newlist = new Jsymbolic_t[1];// java holder
						if( read_symbolic( fp, PLA, word, newlist ) ) {
							if( PLA.symbolic == null ) {
								PLA.symbolic = newlist[0];
							} else {
								Jsymbolic_t p1;
								for( p1 = PLA.symbolic; p1.next != null; p1 = p1.next ) {
								}
								p1.next = newlist[0];
							}
						} else {
							Jcvrmisc.fatal("error reading .symbolic");
						}

					} else if( str.equals("symbolic-output") ) {
						final Jsymbolic_t[] newlist = new Jsymbolic_t[1];// java holder

						if( read_symbolic( fp, PLA, word, newlist ) ) {
							if( PLA.symbolic_output == null ) {
								PLA.symbolic_output = newlist[1];
							} else {
								Jsymbolic_t p1;
								for( p1 = PLA.symbolic_output; p1.next != null; p1 = p1.next ) {
								}
								p1.next = newlist[1];
							}
						} else {
							Jcvrmisc.fatal("error reading .symbolic-output");
						}

						/* .phase allows a choice of output phases */
					} else if( str.equals("phase") ) {
						if( cube.fullset == null ) {
							Jcvrmisc.fatal("PLA size must be declared before .phase");
						}
						if( PLA.phase != null ) {
							System.err.printf("extra .phase ignored\n");
							skip_line( fp, System.out, /* echo */ false );
						} else {
							do {
								ch = fp.read();
							} while( ch == ' ' || ch == '\t' );
							fp.unread( ch );
							PLA.phase = Jset.set_save( cube.fullset );
							final int last = cube.last_part[cube.num_vars - 1];
							for( int i = cube.first_part[cube.num_vars - 1]; i <= last; i++ ) {
								if( (ch = fp.read()) == '0' ) {
									Jespresso.set_remove( PLA.phase, i );
								} else if( ch != '1') {
									Jcvrmisc.fatal("only 0 or 1 allowed in phase description");
								}
							}
						}

						/* .pair allows for bit-pairing input variables */
					} else if( str.equals("pair") ) {
						if( PLA.pair != null ) {
							System.err.printf("extra .pair ignored\n");
						} else {
							Jpair pair;
							PLA.pair = pair = new Jpair();
							try {
								pair.cnt = Jsupport.scanInt( fp, 10 );
							} catch( final IOException e ) {
								Jcvrmisc.fatal("syntax error in .pair");
							}
							pair.var1 = new int[ pair.cnt ];
							pair.var2 = new int[ pair.cnt ];
							final int[] j = new int[1];// java holder
							final int[] var = new int[1];// java holder
							for( int i = 0; i < pair.cnt; i++ ) {
								str = get_word( fp, word );
								if( word.charAt( 0 ) == '(' ) {
									word.delete( 0, 1 );
									str = word.toString();
								}
								if( label_index( PLA, word, var, j ) ) {
									pair.var1[ i ] = var[0] + 1;
								} else {
									Jcvrmisc.fatal("syntax error in .pair");
								}

								str = get_word( fp, word );
								if( word.charAt( word.length() - 1 ) == ')' ) {
									word.delete( word.length() - 1, word.length() );
									str = word.toString();
								}
								if( label_index( PLA, word, var, j ) ) {
									pair.var2[ i ] = var[0] + 1;
								} else {
									Jcvrmisc.fatal("syntax error in .pair");
								}
							}
						}
					} else {
						if( Jglobals.echo_unknown_commands ) {
							System.out.printf("%c%s ", ch, word );
						}
						skip_line( fp, System.out, Jglobals.echo_unknown_commands );
					}
					break;
				default:
					fp.unread( ch );
					if( cube.fullset == null ) {
						/*		fatal("unknown PLA size, need .i/.o or .mv");*/
						if( Jglobals.echo_comments ) {
							System.out.print('#');
						}
						skip_line( fp, System.out, Jglobals.echo_comments );
						break;
					}
					if( PLA.F == null ) {
						PLA.F = Jespresso.new_cover( 10 );
						PLA.D = Jespresso.new_cover( 10 );
						PLA.R = Jespresso.new_cover( 10 );
					}
					read_cube( fp, PLA );
				}
			} catch(final IOException ie) {
				return;
			}
		// goto loop;
		} while( true );
	}

	/**
	read_pla -- read a PLA from a file

	Input stops when ".e" is encountered in the input file, or upon reaching
	end of file.

	Returns the PLA in the variable PLA after massaging the "symbolic"
	representation into a positional cube notation of the ON-set, OFF-set,
	and the DC-set.

	needs_dcset and needs_offset control the computation of the OFF-set
	and DC-set (i.e., if either needs to be computed, then it will be
	computed via complement only if the corresponding option is TRUE.)
	pla_type specifies the interpretation to be used when reading the
	PLA.

	The phase of the output functions is adjusted according to the
	global option "pos" or according to an imbedded .phase option in
	the input file.  Note that either phase option implies that the
	OFF-set be computed regardless of whether the caller needs it
	explicitly or not.

	Bit pairing of the binary variables is performed according to an
	imbedded .pair option in the input file.

	The global cube structure also reflects the sizes of the PLA which
	was just read.  If these fields have already been set, then any
	subsequent PLA must conform to these sizes.

	The global flags trace and summary control the output produced
	during the read.

	Returns a status code as a result:
	EOF (-1) : End of file reached before any data was read
	> 0	 : Operation successful
	*/
	static int read_pla(final PushbackInputStream fp, final boolean needs_dcset, boolean needs_offset, final int pla_type, final JpPLA[] PLA_return) {
		final Jcube_struct cube = Jglobals.cube;// java

		/* Allocate and initialize the PLA structure */
		final JpPLA PLA = PLA_return[ 0 ] = new_PLA();
		PLA.pla_type = pla_type;

		/* Read the pla */
		long time = System.currentTimeMillis();
		parse_pla( fp, PLA );

		/* Check for nothing on the file -- implies reached EOF */
		if( PLA.F == null ) {
			return -1;// FIXME free allocated PLA
		}

		/* This hack merges the next-state field with the outputs */
		for( int i = 0; i < cube.num_vars; i++ ) {
			cube.part_size[ i ] = Math.abs( cube.part_size[ i ] );
		}
		if( Jglobals.kiss ) {
			final int third = cube.num_vars - 3;
			final int second = cube.num_vars - 2;
			if( cube.part_size[third] != cube.part_size[second] ) {
				System.err.printf(" with .kiss option, third to last and second\n");
				System.err.printf("to last variables must be the same size.\n");
				return -1;// FIXME free allocated PLA
			}
			for( int i = 0; i < cube.part_size[second]; i++ ) {
				PLA.label[ i + cube.first_part[ second ] ] = PLA.label[ i + cube.first_part[ third ] ];
			}
			cube.part_size[ second ] += cube.part_size[ cube.num_vars - 1 ];
			cube.num_vars--;
			Jcubestr.setdown_cube();
			Jcubestr.cube_setup();
		}

		final Jcost_t cost = new Jcost_t();
		if( Jglobals.trace ) {
			Jcvrmisc.totals( time, Jespresso.READ_TIME, PLA.F, cost );
		}

		/* Decide how to break PLA into ON-set, OFF-set and DC-set */
		time = System.currentTimeMillis();
		if( Jglobals.pos || PLA.phase != null || PLA.symbolic_output != null ) {
			needs_offset = true;
		}
		if( needs_offset && (PLA.pla_type == Jespresso.F_type || PLA.pla_type == Jespresso.FD_type) ) {
			Jespresso.free_cover( PLA.R );
			PLA.R = Jcompl.complement( Jcofactor.cube2list( PLA.F, PLA.D ) );
		} else if( needs_dcset && PLA.pla_type == Jespresso.FR_type ) {
			Jespresso.free_cover( PLA.D );
			/* hack, why not? */
			final Jset_family X = Jcontain.d1merge( Jset.sf_join( PLA.F, PLA.R ), cube.num_vars - 1);
			PLA.D = Jcompl.complement( Jcofactor.cube1list( X ) );
			Jespresso.free_cover( X );
		} else if( PLA.pla_type == Jespresso.R_type || PLA.pla_type == Jespresso.DR_type ) {
			Jespresso.free_cover( PLA.F );
			PLA.F = Jcompl.complement( Jcofactor.cube2list( PLA.D, PLA.R ) );
		}

		if( Jglobals.trace ) {
			Jcvrmisc.totals( time, Jespresso.COMPL_TIME, PLA.R, cost );
		}

		/* Check for phase rearrangement of the functions */
		if( Jglobals.pos ) {
			final Jset_family onset = PLA.F;
			PLA.F = PLA.R;
			PLA.R = onset;
			PLA.phase = Jespresso.new_cube();
			Jset.set_diff( PLA.phase, cube.fullset, cube.var_mask[ cube.num_vars - 1 ] );
		} else if( PLA.phase != null ) {
			Jopo.set_phase( PLA );
		}

		/* Setup minimization for two-bit decoders */
		if( PLA.pair != null ) {
			Jpair.set_pair( PLA );
		}

		if( PLA.symbolic != null ) {
		    {
		    	final long t = System.currentTimeMillis();
		    	Jhack.map_symbolic( PLA );
		    	if( Jglobals.trace ) {
					Jcvrmisc.print_trace( PLA.F, "MAP-INPUT  ", System.currentTimeMillis() - t );
				}
		    }
		}
		if( PLA.symbolic_output != null ) {
			{
		    	final long t = System.currentTimeMillis();
		    	Jhack.map_output_symbolic( PLA );
		    	if( Jglobals.trace ) {
					Jcvrmisc.print_trace( PLA.F, "MAP-OUTPUT ", System.currentTimeMillis() - t );
				}
		    }
			if( needs_offset ) {
				Jespresso.free_cover( PLA.R );
			    {
			    	final long t = System.currentTimeMillis();
			    	PLA.R = Jcompl.complement( Jcofactor.cube2list( PLA.F, PLA.D ) );
			    	Jcvrmisc.totals( t, Jespresso.COMPL_TIME, PLA.R, cost );
			    }
			}
		}

		return 1;
	}

	@SuppressWarnings("boxing")
	static void PLA_summary(final JpPLA PLA) {
		final Jcube_struct cube = Jglobals.cube;// java

		System.out.printf("# PLA is %s", PLA.filename );
		if( cube.num_binary_vars == cube.num_vars - 1 ) {
			System.out.printf(" with %d inputs and %d outputs\n",
		cube.num_binary_vars, cube.part_size[cube.num_vars - 1]);
		} else {
			System.out.printf(" with %d variables (%d binary, mv sizes",
			cube.num_vars, cube.num_binary_vars);
			for( int var = cube.num_binary_vars; var < cube.num_vars; var++) {
				System.out.printf(" %d", cube.part_size[var]);
			}
			System.out.printf(")\n");
		}
		System.out.printf("# ON-set cost is  %s\n", Jcvrmisc.print_cost( PLA.F ) );
		System.out.printf("# OFF-set cost is %s\n", Jcvrmisc.print_cost( PLA.R ) );
		System.out.printf("# DC-set cost is  %s\n", Jcvrmisc.print_cost( PLA.D ) );
		if( PLA.phase != null ) {
			System.out.printf("# phase is %s\n", Jcvrout.pc1( PLA.phase, 0 ) );
		}
		if( PLA.pair != null ) {
			System.out.printf("# two-bit decoders:");
			for( int i = 0; i < PLA.pair.cnt; i++) {
				System.out.printf(" (%d %d)", PLA.pair.var1[i], PLA.pair.var2[i]);
			}
			System.out.printf("\n");
		}
		if( PLA.symbolic != null ) {
			for( Jsymbolic_t p1 = PLA.symbolic; p1 != null; p1 = p1.next ) {
				System.out.printf("# symbolic: ");
				for( Jsymbolic_list_t p2 = p1.symbolic_list; p2 != null; p2 = p2.next ) {
					System.out.printf(" %d", p2.variable);
				}
				System.out.printf("\n");
			}
		}
		if( PLA.symbolic_output != null ) {
			for( Jsymbolic_t p1 = PLA.symbolic_output; p1 != null; p1 = p1.next ) {
				System.out.printf("# output symbolic: ");
				for( Jsymbolic_list_t p2 = p1.symbolic_list; p2 != null; p2 = p2.next ) {
					System.out.printf(" %d", p2.pos);
				}
				System.out.printf("\n");
			}
		}
		System.out.flush();
	}

	static JpPLA new_PLA() {
		final JpPLA PLA = new JpPLA();
		PLA.F = PLA.D = PLA.R = null;
		PLA.phase = null;
		PLA.pair = null;
		PLA.label = null;
		PLA.filename = null;
		PLA.pla_type = 0;
		PLA.symbolic = null;
		PLA.symbolic_output = null;
		return PLA;
	}

	static void PLA_labels(final JpPLA PLA) {
		final Jcube_struct cube = Jglobals.cube;// java

		PLA.label = new String[ cube.size ];
		for( int i = 0; i < cube.size; i++ ) {
			PLA.label[i] = null;
		}
	}

	static void free_PLA(JpPLA PLA) {
		if( PLA.F != null ) {
			Jespresso.free_cover( PLA.F );
		}
		if( PLA.R != null ) {
			Jespresso.free_cover( PLA.R );
		}
		if( PLA.D != null ) {
			Jespresso.free_cover( PLA.D );
		}
		if( PLA.phase != null ) {
			PLA.phase = null;
		}
		if( PLA.pair != null ) {
			PLA.pair.var1 = null;
			PLA.pair.var2 = null;
			PLA.pair = null;
		}
		if( PLA.label != null ) {
			for( int i = 0; i < Jglobals.cube.size; i++) {
				if( PLA.label[i] != null ) {
					PLA.label[i] = null;
				}
			}
			PLA.label = null;
		}
		if( PLA.filename != null ) {
			PLA.filename = null;
		}
		for( Jsymbolic_t p1next, p1 = PLA.symbolic; p1 != null; p1 = p1next ) {
			for( Jsymbolic_list_t p2next, p2 = p1.symbolic_list; p2 != null; p2 = p2next ) {
				p2next = p2.next;
				p2 = null;
			}
			p1next = p1.next;
			p1 = null;
		}
		PLA.symbolic = null;
		for( Jsymbolic_t p1next, p1 = PLA.symbolic_output; p1 != null; p1 = p1next ) {
			for( Jsymbolic_list_t p2next, p2 = p1.symbolic_list; p2 != null; p2 = p2next ) {
				p2next = p2.next;
				p2 = null;
			}
			p1next = p1.next;
			p1 = null;
		}
		PLA.symbolic_output = null;
		PLA = null;
	}

	/**
	 *
	 * @param fp
	 * @param PLA
	 * @param word scratch string for words
	 * @param retval
	 * @return
	 */
	static boolean read_symbolic(final InputStream fp, final JpPLA PLA, final StringBuilder word, final Jsymbolic_t[] retval) {
		final int[] i = new int[1];// java holder
		final int[] var = new int[1];// java holder

		final Jsymbolic_t newlist = new Jsymbolic_t();
		newlist.next = null;
		newlist.symbolic_list = null;
		newlist.symbolic_list_length = 0;
		newlist.symbolic_label = null;
		newlist.symbolic_label_length = 0;
		Jsymbolic_list_t prev_listp = null;
		Jsymbolic_label_t prev_labelp = null;

		for( ;; ) {
			final String str = get_word( fp, word );
			if( str.equals(";") ) {
				break;
			}
			if( label_index( PLA, word, var, i ) ) {
				final Jsymbolic_list_t listp = new Jsymbolic_list_t();
				listp.variable = var[ 0 ];
				listp.pos = i[ 0 ];
				listp.next = null;
				if( prev_listp == null ) {
					newlist.symbolic_list = listp;
				} else {
					prev_listp.next = listp;
				}
				prev_listp = listp;
				newlist.symbolic_list_length++;
			} else {
				return false;
			}
		}

		for( ;; ) {
			final String str = get_word( fp, word );
			if( str.equals(";") ) {
				break;
			}
			final Jsymbolic_label_t labelp = new Jsymbolic_label_t();
			labelp.label = str;
			labelp.next = null;
			if( prev_labelp == null ) {
				newlist.symbolic_label = labelp;
			} else {
				prev_labelp.next = labelp;
			}
			prev_labelp = labelp;
			newlist.symbolic_label_length++;
		}

		retval[ 0 ] = newlist;
		return true;
	}

	static boolean label_index(final JpPLA PLA, final StringBuilder word, final int[] varp, final int[] ip) {
		if( PLA.label == null || PLA.label[0] == null ) {
			try {
				varp[0] = scanInt( word );
				ip[0] = varp[0];
				return true;
			} catch( final IOException e ) {
			}
		} else {
			final Jcube_struct cube = Jglobals.cube;// java
			final String s = word.toString();
			for( int var = 0; var < cube.num_vars; var++ ) {
				for( int i = 0; i < cube.part_size[ var ]; i++ ) {
					if( PLA.label[ cube.first_part[ var ] + i ].equals( s ) ) {
						varp[ 0 ] = var;
						ip[ 0 ] = i;
						return true;
					}
				}
			}
		}
		return false;
	}
}