package espresso;

final class Jcvrmisc {

	/** cost -- compute the cost of a cover */
	static void cover_cost(final Jset_family F, final Jcost_t cost) {
		final Jcube_struct cube = Jglobals.cube;// java
		final Jcdata_struct cdata = Jglobals.cdata;// java

		/* use the routine used by cofactor to decide splitting variables */
		Jcofactor.massive_count( Jcofactor.cube1list( F ) );

		cost.cubes = F.count;
		cost.total = cost.in = cost.out = cost.mv = cost.primes = 0;

		/* Count transistors (zeros) for each binary variable (inputs) */
		for( int var = 0; var < cube.num_binary_vars; var++ ) {
			cost.in += cdata.var_zeros[var];
		}

		/* Count transistors for each mv variable based on sparse/dense */
		for( int var = cube.num_binary_vars; var < cube.num_vars - 1; var++ ) {
			if( cube.sparse[var] ) {
				cost.mv += F.count * cube.part_size[var] - cdata.var_zeros[var];
			} else {
				cost.mv += cdata.var_zeros[var];
			}
		}

		/* Count the transistors (ones) for the output variable */
		if( cube.num_binary_vars != cube.num_vars ) {
			final int var = cube.num_vars - 1;
			cost.out = F.count * cube.part_size[var] - cdata.var_zeros[var];
		}

		/* Count the number of nonprime cubes */
		final int[] data = F.data;// java
	    for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
			cost.primes += Jespresso.TESTP( data, p, Jespresso.PRIME ) ? 1 : 0;
		}

		/* Count the total number of literals */
		cost.total = cost.in + cost.out + cost.mv;
	}

	/** fmt_cost -- return a string which reports the "cost" of a cover */
	@SuppressWarnings("boxing")
	static String fmt_cost(final Jcost_t cost) {
		final Jcube_struct cube = Jglobals.cube;// java
		String s;

		if( cube.num_binary_vars == cube.num_vars - 1 ) {
			s = String.format("c=%d(%d) in=%d out=%d tot=%d",
				cost.cubes, cost.cubes - cost.primes, cost.in,
				cost.out, cost.total );
		} else {
			s = String.format("c=%d(%d) in=%d mv=%d out=%d",
				cost.cubes, cost.cubes - cost.primes, cost.in,
				cost.mv, cost.out );
		}
		return s;
	}

	static String print_cost(final Jset_family F) {
		final Jcost_t cost = new Jcost_t();
		cover_cost( F, cost );
		return fmt_cost( cost );
	}

	/** copy_cost -- copy a cost function from s to d */
	static void copy_cost(final Jcost_t s, final Jcost_t d) {
		d.cubes = s.cubes;
		d.in = s.in;
		d.out = s.out;
		d.mv = s.mv;
		d.total = s.total;
		d.primes = s.primes;
	}

	/** size_stamp -- print single line giving the size of a cover */
	static void size_stamp(final Jset_family T, final String name) {
		System.out.printf("# %s\tCost is %s\n", name, print_cost( T ) );
		System.out.flush();
	}

	/** print_trace -- print a line reporting size and time after a function */
	static void print_trace(final Jset_family T, final String name, final long time) {
		System.out.printf("# %s\tTime was %s, cost is %s\n",
				name, Jespresso.print_time( time ), print_cost( T ) );
		System.out.flush();
	}

	/** totals -- add time spent in the function into the totals */
	static void totals(long time, final int i, final Jset_family T, final Jcost_t cost) {
		time = System.currentTimeMillis() - time;
		Jglobals.total_time[ i ] += time;
		Jglobals.total_calls[ i ]++;
		cover_cost( T, cost );
		if( Jglobals.trace ) {
			System.out.printf("# %s\tTime was %s, cost is %s\n",
					Jglobals.total_name[ i ], Jespresso.print_time( time ), fmt_cost( cost ) );
			System.out.flush();
		}
	}

	/** fatal -- report fatal error message and take a dive */
	static void fatal(final String s) {
		System.err.printf("espresso: %s\n", s);
		Jmain.closeStreams();
		System.exit( 1 );
	}
}