package espresso;

import java.io.PrintStream;

/**
    module: cvrout.c
    purpose: cube and cover output routines
*/

final class Jcvrout {

	@SuppressWarnings("boxing")
	static void fprint_pla(final PrintStream fp, final JpPLA PLA, int output_type) {

		if( (output_type & Jespresso.CONSTRAINTS_type) != 0 ) {
			output_symbolic_constraints( fp, PLA, false );
			output_type &= ~ Jespresso.CONSTRAINTS_type;
			if( output_type == 0 ) {
				return;
			}
		}

		if( (output_type & Jespresso.SYMBOLIC_CONSTRAINTS_type) != 0 ) {
			output_symbolic_constraints( fp, PLA, true );
			output_type &= ~ Jespresso.SYMBOLIC_CONSTRAINTS_type;
			if( output_type == 0 ) {
				return;
			}
		}

		if( output_type == Jespresso.PLEASURE_type ) {
			pls_output( PLA );
		} else if( output_type == Jespresso.EQNTOTT_type ) {
			eqn_output( PLA );
		} else if( output_type == Jespresso.KISS_type ) {
			kiss_output( fp, PLA );
		} else {
			fpr_header( fp, PLA, output_type );

			int num = 0;
			if( (output_type & Jespresso.F_type) != 0 ) {
				num += PLA.F.count;
			}
			if( (output_type & Jespresso.D_type) != 0 ) {
				num += PLA.D.count;
			}
			if( (output_type & Jespresso.R_type) != 0 ) {
				num += PLA.R.count;
			}
			fp.printf(".p %d\n", num);

			/* quick patch 01/17/85 to support TPLA ! */
			if( output_type == Jespresso.F_type ) {
				final int[] data = PLA.F.data;// java
			    for( int p = 0, last = p + PLA.F.count * PLA.F.wsize; p < last; p += PLA.F.wsize ) {
					print_cube( fp, data, p, "01");
				}
				fp.printf(".e\n");
			} else {
				if( (output_type & Jespresso.F_type) != 0 ) {
					final int[] data = PLA.F.data;// java
				    for( int p = 0, last = p + PLA.F.count * PLA.F.wsize; p < last; p += PLA.F.wsize ) {
						print_cube( fp, data, p, "~1");
					}
				}
				if( (output_type & Jespresso.D_type) != 0 ) {
					final int[] data = PLA.D.data;// java
				    for( int p = 0, last = p + PLA.D.count * PLA.D.wsize; p < last; p += PLA.D.wsize ) {
						print_cube( fp, data, p, "~2");
					}
				}
				if( (output_type & Jespresso.R_type) != 0 ) {
					final int[] data = PLA.R.data;// java
				    for( int p = 0, last = p + PLA.R.count * PLA.R.wsize; p < last; p += PLA.R.wsize ) {
						print_cube( fp, data, p, "~0");
					}
				}
				fp.printf(".end\n");
			}
		}
	}

	@SuppressWarnings("boxing")
	static void fpr_header(final PrintStream fp, final JpPLA PLA, final int output_type) {
		/* .type keyword gives logical type */
		if( output_type != Jespresso.F_type ) {
			fp.printf(".type ");
			if( (output_type & Jespresso.F_type) != 0 ) {
				fp.print('f');
			}
			if( (output_type & Jespresso.D_type) != 0 ) {
				fp.print('d');
			}
			if( (output_type & Jespresso.R_type) != 0 ) {
				fp.print('r');
			}
			fp.print('\n');
		}

		/* Check for binary or multiple-valued labels */
		final Jcube_struct cube = Jglobals.cube;// java
		if( cube.num_mv_vars <= 1 ) {
			fp.printf(".i %d\n", cube.num_binary_vars );
			if( cube.output != -1 ) {
				fp.printf(".o %d\n", cube.part_size[cube.output] );
			}
		} else {
			fp.printf(".mv %d %d", cube.num_vars, cube.num_binary_vars );
			for( int var = cube.num_binary_vars; var < cube.num_vars; var++ ) {
				fp.printf(" %d", cube.part_size[var] );
			}
			fp.printf("\n");
		}

		/* binary valued labels */
		if( PLA.label != null && PLA.label[1] != null
				&& cube.num_binary_vars > 0 ) {
			fp.printf(".ilb");
			for( int var = 0; var < cube.num_binary_vars; var++ ) {
				fp.printf(" %s", Jespresso.INLABEL( PLA, var ) );
			}
			fp.print('\n');
		}

		/* output-part (last multiple-valued variable) labels */
		if( PLA.label != null &&
				PLA.label[ cube.first_part[ cube.output ] ] != null
				&& cube.output != -1 ) {
			fp.printf(".ob");
			for( int i = 0; i < cube.part_size[ cube.output ]; i++ ) {
				fp.printf(" %s", Jespresso.OUTLABEL( PLA, i ) );
			}
			fp.print('\n');
		}

		/* multiple-valued labels */
		for( int var = cube.num_binary_vars; var < cube.num_vars - 1; var++ ) {
			final int first = cube.first_part[ var ];
			final int last = cube.last_part[ var ];
			if( PLA.label != null && PLA.label[ first ] != null ) {
				fp.printf(".label var=%d", var);
				for( int i = first; i <= last; i++ ) {
					fp.printf(" %s", PLA.label[ i ] );
				}
				fp.print('\n');
			}
		}

		if( PLA.phase != null ) {
			final int first = cube.first_part[ cube.output ];
			final int last = cube.last_part[ cube.output ];
			fp.printf("#.phase ");
			for( int i = first; i <= last; i++ ) {
				fp.print( Jespresso.is_in_set( PLA.phase, 0, i ) ? '1' : '0');
			}
			fp.printf("\n");
		}
	}

	@SuppressWarnings("boxing")
	static void pls_output(final JpPLA PLA) {
		System.out.printf(".option unmerged\n");
		makeup_labels( PLA );
		pls_label( PLA, System.out );
		pls_group( PLA, System.out );
		System.out.printf(".p %d\n", PLA.F.count );
		final int[] data = PLA.F.data;// java
		for( int p = 0, last = p + PLA.F.count * PLA.F.wsize; p < last; p += PLA.F.wsize ) {
			print_expanded_cube( System.out, data, p, PLA.phase );
		}
		System.out.printf(".end\n");
	}


	static void pls_group(final JpPLA PLA, final PrintStream fp) {
		final Jcube_struct cube = Jglobals.cube;// java
		fp.printf("\n.group");
		int col = 6;
		for( int var = 0; var < cube.num_vars - 1; var++ ) {
			fp.printf(" (");
			col += 2;
			for( int i = cube.first_part[ var ]; i <= cube.last_part[ var ]; i++ ) {
				final int len = PLA.label[ i ].length();
				if( col + len > 75 ) {
					fp.printf(" \\\n");
					col = 0;
				} else if( i != 0 ) {
					fp.print(' ');
					col += 1;
				}
				fp.printf("%s", PLA.label[ i ] );
				col += len;
			}
			fp.printf(")");
			col += 1;
		}
		fp.printf("\n");
	}

	static void pls_label(final JpPLA PLA, final PrintStream fp) {
		fp.printf(".label");
		final Jcube_struct cube = Jglobals.cube;// java
		int col = 6;
		for( int var = 0; var < cube.num_vars; var++ ) {
			for( int i = cube.first_part[ var ]; i <= cube.last_part[ var ]; i++ ) {
				final int len = PLA.label[ i ].length();
				if( col + len > 75 ) {
					fp.printf(" \\\n");
					col = 0;
				} else {
					fp.print(' ');
					col += 1;
				}
				fp.printf("%s", PLA.label[ i ] );
				col += len;
			}
		}
	}

	/** eqntott output mode -- output algebraic equations */
	static void eqn_output(final JpPLA PLA) {
		final Jcube_struct cube = Jglobals.cube;// java
		if( cube.output == -1 ) {
			Jcvrmisc.fatal("Cannot have no-output function for EQNTOTT output mode");
		}
		if( cube.num_mv_vars != 1 ) {
			Jcvrmisc.fatal("Must have binary-valued function for EQNTOTT output mode");
		}
		makeup_labels( PLA );

		/* Write a single equation for each output */
		for( int i = 0; i < cube.part_size[ cube.output ]; i++ ) {
			System.out.printf("%s = ", Jespresso.OUTLABEL( PLA, i ) );
			int col = Jespresso.OUTLABEL( PLA, i ).length() + 3;
			boolean firstor = true;

			/* Write product terms for each cube in this output */
			final int[] data = PLA.F.data;// java
			for( int p = 0, last = p + PLA.F.count * PLA.F.wsize; p < last; p += PLA.F.wsize ) {
				if( Jespresso.is_in_set( data, p, i + cube.first_part[ cube.output ] ) ) {
					if( firstor ) {
						System.out.printf("(");
						col += 1;
					} else {
						System.out.printf(" | (");
						col += 4;
					}
					firstor = false;
					boolean firstand = true;

					/* print out a product term */
					for( int var = 0; var < cube.num_binary_vars; var++ ) {
						int x;
						if( (x = Jespresso.GETINPUT( data, p, var ) ) != Jespresso.DASH ) {
							final int len = Jespresso.INLABEL( PLA, var ).length();
							if( col+len > 72 ) {
								System.out.printf("\n    ");
								col = 4;
							}
							if( ! firstand ) {
								System.out.printf("&");
								col += 1;
							}
							firstand = false;
							if( x == Jespresso.ZERO ) {
								System.out.printf("!");
								col += 1;
							}
							System.out.printf("%s", Jespresso.INLABEL( PLA, var ) );
							col += len;
						}
					}
					System.out.printf(")");
					col += 1;
				}
			}
			System.out.printf(";\n\n");
		}
	}

	private static String fmt_cube(final int[] c, final int off, final char[] out_map) {
		final StringBuilder s = new StringBuilder();// java
		final Jcube_struct cube = Jglobals.cube;// java
		for( int var = 0; var < cube.num_binary_vars; var++ ) {
			s.append( new char[] {'?','0','1','-'}[ Jespresso.GETINPUT( c, off, var ) ] );
		}
		for( int var = cube.num_binary_vars; var < cube.num_vars - 1; var++ ) {
			s.append(' ');
			for( int i = cube.first_part[ var ]; i <= cube.last_part[ var ]; i++ ) {
				s.append( Jespresso.is_in_set( c, off, i ) ? '1' : '0' );
			}
		}
		if( cube.output != -1 ) {
			final int last = cube.last_part[ cube.output ];
			s.append(' ');
			for( int i = cube.first_part[ cube.output ]; i <= last; i++ ) {
				s.append( out_map[ Jespresso.is_in_set( c, off, i ) ? 1 : 0 ] );
			}
		}
		// s[len] = '\0';
		return s.toString();
	}

	static void print_cube(final PrintStream fp, final int[] c, final int off_c, final String out_map) {
		final Jcube_struct cube = Jglobals.cube;// java
		for( int var = 0; var < cube.num_binary_vars; var++ ) {
			final char ch = new char[] {'?','0','1','-'}[ Jespresso.GETINPUT( c, off_c, var ) ];
			fp.print( ch );
		}
		for( int var = cube.num_binary_vars; var < cube.num_vars - 1; var++ ) {
			fp.print(' ');
			for( int i = cube.first_part[var]; i <= cube.last_part[ var ]; i++ ) {
				final char ch = Jespresso.is_in_set( c, off_c, i ) ? '1' : '0';
				fp.print( ch );
			}
		}
		if( cube.output != -1 ) {
			final int last = cube.last_part[ cube.output ];
			fp.print(' ');
			for( int i = cube.first_part[cube.output]; i <= last; i++ ) {
				final char ch = out_map.charAt( Jespresso.is_in_set( c, off_c, i ) ? 1 : 0 );
				fp.print( ch );
			}
		}
		fp.print('\n');
	}

	static void print_expanded_cube(final PrintStream fp, final int[] c, final int off_c, final int[] phase) {
		final Jcube_struct cube = Jglobals.cube;// java
		for( int var = 0; var < cube.num_binary_vars; var++ ) {
			for( int i = cube.first_part[var]; i <= cube.last_part[var]; i++ ) {
				final char ch = Jespresso.is_in_set( c, off_c, i ) ? '1' : '~';
				fp.print( ch );
			}
		}
		for( int var = cube.num_binary_vars; var < cube.num_vars - 1; var++ ) {
			for( int i = cube.first_part[var]; i <= cube.last_part[var]; i++ ) {
				final char ch = Jespresso.is_in_set( c, off_c, i ) ? '~' : '1';
				fp.print( ch );
			}
		}
		if( cube.output != -1 ) {
			final int var = cube.num_vars - 1;
			fp.print(' ');
			for( int i = cube.first_part[var]; i <= cube.last_part[var]; i++ ) {
				char ch = '~';
				if( phase == null || Jespresso.is_in_set( phase, i ) ) {
					if( Jespresso.is_in_set( c, off_c, i ) ) {
						ch = '1';
					}
				} else {
					if( Jespresso.is_in_set( c, off_c, i ) ) {
						ch = '0';
					}
				}
				fp.print( ch );
			}
		}
		fp.print('\n');
	}

	static String pc1(final int[] c, final int off) {
		return fmt_cube( c, off, new char[] {'0','1'} );
	}

	static String pc2(final int[] c) {
		return fmt_cube( c, 0, new char[] {'0','1'} );
	}


	@SuppressWarnings("boxing")
	static void debug_print(final Jset[] T, final String name, final int level) {
		int cnt = Jespresso.CUBELISTSIZE( T );
		int[] temp = Jespresso.new_cube();
		if( Jglobals.verbose_debug && level == 0 ) {
			System.out.printf("\n");
		}
		System.out.printf("%s[%d]: ord(T)=%d\n", name, level, cnt );
		if( Jglobals.verbose_debug ) {
			System.out.printf("cofactor=%s\n", pc1( T[0].base, T[0].offset ) );
			int T1;
			Jset p;
			for( T1 = 2, cnt = 1; (p = T[ T1++ ]) != null; cnt++ ) {
				Jset.set_or( temp, p.base, p.offset, T[0].base, T[0].offset );
				System.out.printf("%4d. %s\n", cnt, pc1( temp, 0 ) );
			}
		}
		temp = null;
	}


	@SuppressWarnings("boxing")
	static void debug1_print(final Jset_family T, final String name, final int num) {
		if( Jglobals.verbose_debug && num == 0 ) {
			System.out.printf("\n");
		}
		System.out.printf("%s[%d]: ord(T)=%d\n", name, num, T.count );
		if( Jglobals.verbose_debug ) {
			int cnt = 1;
			final int[] data = T.data;// java
		    for( int p = 0, last = p + T.count * T.wsize; p < last; p += T.wsize ) {
				System.out.printf("%4d. %s\n", cnt++, pc1( data, p ) );
			}
		}
	}

	static void cprint(final Jset_family T) {
		final int[] data = T.data;// java
	    for( int p = 0, last = p + T.count * T.wsize; p < last; p += T.wsize ) {
			System.out.printf("%s\n", pc1( data, p ) );
		}
	}


	@SuppressWarnings("boxing")
	static void makeup_labels(final JpPLA PLA) {
		if( PLA.label == null ) {
			Jcvrin.PLA_labels( PLA );
		}
		final Jcube_struct cube = Jglobals.cube;// java
		for( int var = 0; var < cube.num_vars; var++ ) {
			for( int i = 0; i < cube.part_size[ var ]; i++ ) {
				final int ind = cube.first_part[ var ] + i;
				if( PLA.label[ ind ] == null ) {
					if( var < cube.num_binary_vars ) {
						if( (i % 2) == 0 ) {
							PLA.label[ ind ] = String.format("v%d.bar", var);
						} else {
							PLA.label[ ind ] = String.format("v%d", var);
						}
					} else {
						PLA.label[ ind ] = String.format("v%d.%d", var, i);
					}
				}
			}
		}
	}

	static void kiss_output(final PrintStream fp, final JpPLA PLA) {
		int[] data = PLA.F.data;// java
		for( int p = 0, last = p + PLA.F.count * PLA.F.wsize; p < last; p += PLA.F.wsize ) {
			kiss_print_cube( fp, PLA, data, p, "~1");
		}
		data = PLA.D.data;// java
		for( int p = 0, last = p + PLA.D.count * PLA.D.wsize; p < last; p += PLA.D.wsize ) {
			kiss_print_cube( fp, PLA, data, p, "~2");
		}
	}

	static void kiss_print_cube(final PrintStream fp, final JpPLA PLA, final int[] p, final int off, final String out_string) {
		final Jcube_struct cube = Jglobals.cube;// java
		int var;
		for( var = 0; var < cube.num_binary_vars; var++ ) {
			final char x = new char[] {'?','0','1','-'}[ Jespresso.GETINPUT( p, off, var ) ];
			fp.print( x );
		}

		for( var = cube.num_binary_vars; var < cube.num_vars - 1; var++ ) {
			fp.print(' ');
			if( Jset.setp_implies( cube.var_mask[ var ], p, off ) ) {
				fp.print('-');
			} else {
				int part = -1;
				for( int i = cube.first_part[ var ]; i <= cube.last_part[ var ]; i++ ) {
					if( Jespresso.is_in_set( p, off, i ) ) {
						if( part != -1 ) {
							Jcvrmisc.fatal("more than 1 part in a symbolic variable\n");
						}
						part = i;
					}
				}
				if( part == -1 ) {
					fp.print('~');	/* no parts, hope its an output ... */
				} else {
					fp.print( PLA.label[ part ] );
				}
			}
		}

		if( (var = cube.output) != -1 ) {
			fp.print(' ');
			for( int i = cube.first_part[ var ]; i <= cube.last_part[ var ]; i++ ) {
				final char x = out_string.charAt( Jespresso.is_in_set( p, off, i ) ? 1 : 0 );
				fp.print( x );
			}
		}

		fp.print('\n');
	}

	@SuppressWarnings("boxing")
	static void output_symbolic_constraints(final PrintStream fp, final JpPLA PLA, final boolean output_symbolic) {
		final Jcube_struct cube = Jglobals.cube;// java
		if( (cube.num_vars - cube.num_binary_vars) <= 1 ) {
			return;
		}
		makeup_labels( PLA );

		for( int var = cube.num_binary_vars; var < cube.num_vars - 1; var++ ) {

			/* pull out the columns for variable "var" */
			final int npermute = cube.part_size[var];
			int[] permute = new int[ npermute ];
			for( int i = 0; i < npermute; i++ ) {
				permute[i] = cube.first_part[var] + i;
			}
			final Jset_family A = Jset.sf_permute( Jset.sf_save( PLA.F ), permute, npermute );
			permute = null;

			/* Delete the singletons and the full sets */
			int noweight = 0;
			for( int i = 0; i < A.count; i++ ) {
				final int size = Jset.set_ord( A.data, A.wsize * i );
				if( size == 1 || size == A.sf_size ) {
					Jset.sf_delset( A, i-- );
					noweight++;
				}
			}

			/* Count how many times each is duplicated */
			int[] weight = new int[ A.count ];
			for( int i = 0; i < A.count; i++ ) {
				Jespresso.RESET( A.data, A.wsize * i, Jespresso.COVERED );
			}
			for( int i = 0; i < A.count; i++ ) {
				weight[i] = 0;
				if( ! Jespresso.TESTP( A.data, A.wsize * i, Jespresso.COVERED ) ) {
					weight[i] = 1;
					for( int j = i + 1; j < A.count; j++ ) {
						if( Jset.setp_equal( A.data, A.wsize * i, A.data, A.wsize * j ) ) {
							weight[i]++;
							Jespresso.SET( A.data, A.wsize * j, Jespresso.COVERED );
						}
					}
				}
			}

			/* Print out the constraints */
			if( ! output_symbolic ) {
				fp.printf("# Symbolic constraints for variable %d (Numeric form)\n", var );
				fp.printf("# unconstrained weight = %d\n", noweight);
				fp.printf("num_codes=%d\n", cube.part_size[ var ]);
				for( int i = 0; i < A.count; i++ ) {
					if( weight[ i ] > 0 ) {
						fp.printf("weight=%d: ", weight[ i ] );
						for( int j = 0; j < A.sf_size; j++ ) {
							if( Jespresso.is_in_set( A.data, A.wsize * i, j ) ) {
								fp.printf( " %d", j );
							}
						}
						fp.printf( "\n");
					}
				}
				weight = null;// FIXME weight must be free
			} else {
				fp.printf("# Symbolic constraints for variable %d (Symbolic form)\n", var );
				for( int i = 0; i < A.count; i++ ) {
					if( weight[i] > 0 ) {
						fp.printf("#   w=%d: (", weight[ i ] );
						for( int j = 0; j < A.sf_size; j++ ) {
							if( Jespresso.is_in_set( A.data, A.wsize * i, j ) ) {
								fp.printf(" %s", PLA.label[ cube.first_part[ var ] + j ] );
							}
						}
						fp.printf(" )\n");
					}
				}
				weight = null;
			}
		}
	}
}