package espresso;

import java.util.Comparator;

/**
 * Java: setc.c, descend
 */
final class Jdescend_order implements Comparator<Jset> {
	// static int descend(final Jset[] a, final Jset[] b)
	/*
	These routines compare two sets (cubes) for the qsort() routine and
	return:

	-1 if set a is to precede set b
	0 if set a and set b are equal
	1 if set a is to follow set b

	Usually the SIZE field of the set is assumed to contain the size
	of the set (which will save recomputing the set size during the
	sort).  For distance-1 merging, the global variable cube.temp[0] is
	a mask which mask's-out the merging variable.
	*/

	/** descend -- comparison for descending sort on set size */
	@Override
	public final int compare(final Jset a, final Jset b) {
		if( Jespresso.SIZE( a ) > Jespresso.SIZE( b ) ) {
			return -1;
		} else if( Jespresso.SIZE( a ) < Jespresso.SIZE( b ) ) {
			return 1;
		} else {
			int i = Jespresso.LOOP( a );
			do {
				final long x1 = ((long)a.base[ a.offset + i ] & 0xffffffffL);
				final long x2 = ((long)b.base[ b.offset + i ] & 0xffffffffL);
				if( x1 > x2 ) {
					return -1;
				} else if( x1 < x2 ) {
					return 1;
				}
			} while( --i > 0 );
		}
		return 0;
	}
}
