package espresso;
final class Jirred {

	private static int Rp_current;

	/**
	*   irredundant -- Return a minimal subset of F
	*/
	static Jset_family irredundant(final Jset_family F, final Jset_family D) {
		mark_irredundant( F, D );
		return Jset.sf_inactive( F );
	}

	/**
	*   mark_irredundant -- find redundant cubes, and mark them "INACTIVE"
	*/
	@SuppressWarnings("boxing")
	static void mark_irredundant(final Jset_family F, final Jset_family D) {
		final Jset_family[] E = new Jset_family[1];// java helper
		final Jset_family[] Rt = new Jset_family[1];// java helper
		final Jset_family[] Rp = new Jset_family[1];// java helper

		/* extract a minimum cover */
		irred_split_cover( F, D, E, Rt, Rp );
		final Jsm_matrix table = irred_derive_table( D, E[ 0 ], Rp[ 0 ] );
		final Jsm_row cover = Jmincov.sm_minimum_cover( table, null, /* heuristic */ true, /* debug */ 0 );

		/* mark the cubes for the result */
		final int[] fdata = F.data;// java
		for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
			Jespresso.RESET( fdata, p, Jespresso.ACTIVE );
			Jespresso.RESET( fdata, p, Jespresso.RELESSEN );
		}

		final int[] edata = E[ 0 ].data;// java
		for( int p = 0, last = p + E[ 0 ].count * E[ 0 ].wsize; p < last; p += E[ 0 ].wsize ) {
			final int p1 = F.wsize * Jespresso.SIZE( edata, p );
			// assert( Jset.setp_equal( fdata, p1, edata, p ) );
			Jespresso.SET( fdata, p1, Jespresso.ACTIVE );
			Jespresso.SET( fdata, p1, Jespresso.RELESSEN );		/* for essen(), mark as rel. ess. */
		}

		for( Jsm_element pe = cover.first_col; pe != null; pe = pe.next_col ) {
			Jespresso.SET( fdata, F.wsize * pe.col_num, Jespresso.ACTIVE );
		}

		if( (Jglobals.debug & Jespresso.IRRED) != 0 ) {
			System.out.printf("# IRRED: F=%d E=%d R=%d Rt=%d Rp=%d Rc=%d Final=%d Bound=%d\n",
					F.count, E[ 0 ].count, Rt[ 0 ].count + Rp[ 0 ].count, Rt[ 0 ].count, Rp[ 0 ].count,
					cover.length, E[ 0 ].count + cover.length, 0);
		}

		Jespresso.free_cover( E[0] );
		Jespresso.free_cover( Rt[0] );
		Jespresso.free_cover( Rp[0] );
		Jmatrix.sm_free( table );
		Jrows.sm_row_free( cover );
	}

	/**
	*  irred_split_cover -- find E, Rt, and Rp from the cover F, D
	*
	*	E  -- relatively essential cubes
	*	Rt  -- totally redundant cubes
	*	Rp  -- partially redundant cubes
	*/
	@SuppressWarnings("boxing")
	static void irred_split_cover(final Jset_family F, final Jset_family D, final Jset_family[] E, final Jset_family[] Rt, final Jset_family[] Rp) {
		/* number the cubes of F -- these numbers track into E, Rp, Rt, etc. */
		int index = 0;
		// foreach_set(F, last, p) {
		int[] data = F.data;// java
		for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
			Jespresso.PUTSIZE( data, p, index );
			index++;
		}

		E[ 0 ] = Jespresso.new_cover( 10 );
		Rt[ 0 ] = Jespresso.new_cover( 10 );
		Rp[ 0 ] = Jespresso.new_cover( 10 );
		Jset_family R = Jespresso.new_cover( 10 );

		/* Split F into E and R */
		Jset[] FD = Jcofactor.cube2list( F, D );
		data = F.data;
		for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
			if( cube_is_covered( FD, data, p ) ) {
				R = Jset.sf_addset( R, data, p );
			} else {
				E[ 0 ] = Jset.sf_addset( E[ 0 ], data, p );
			}
			if( (Jglobals.debug & Jespresso.IRRED1) != 0 ) {
				System.out.printf("IRRED1: zr=%d ze=%d to-go=%d time=%s\n",
						R.count, E[ 0 ].count, F.count - (R.count + E[ 0 ].count),
						Jespresso.print_time( System.currentTimeMillis() ));
			}
		}
		FD = null;

		/* Split R into Rt and Rp */
		Jset[] ED = Jcofactor.cube2list( E[ 0 ], D );
		data = R.data;// java
		for( int p = 0, last = p + R.count * R.wsize; p < last; p += R.wsize ) {
			if( cube_is_covered( ED, data, p ) ) {
				Rt[ 0 ] = Jset.sf_addset( Rt[ 0 ], data, p );
			} else {
				Rp[ 0 ] = Jset.sf_addset( Rp[ 0 ], data, p );
			}
			if( (Jglobals.debug & Jespresso.IRRED1) != 0 ) {
				System.out.printf("IRRED1: zr=%d zrt=%d to-go=%d time=%s\n",
						Rp[ 0 ].count, Rt[ 0 ].count,
						R.count - (Rp[ 0 ].count + Rt[ 0 ].count), Jespresso.print_time( System.currentTimeMillis() ) );
			}
		}
		ED = null;

		Jespresso.free_cover( R );
	}

	/**
	*  irred_derive_table -- given the covers D, E and the set of
	*  partially redundant primes Rp, build a covering table showing
	*  possible selections of primes to cover Rp.
	*/
	@SuppressWarnings("boxing")
	static Jsm_matrix irred_derive_table(final Jset_family D, final Jset_family E, final Jset_family Rp) {
		/* Mark each cube in DE as not part of the redundant set */
		int[] data = D.data;// java
		for( int p = 0, last = p + D.count * D.wsize; p < last; p += D.wsize ) {
			Jespresso.RESET( data, p, Jespresso.REDUND );
		}
		data = E.data;// java
		for( int p = 0, last = p + E.count * E.wsize; p < last; p += E.wsize ) {
			Jespresso.RESET( data, p, Jespresso.REDUND );
		}

		/* Mark each cube in Rp as partially redundant */
		data = Rp.data;// java
		for( int p = 0, last = p + Rp.count * Rp.wsize; p < last; p += Rp.wsize ) {
			Jespresso.SET( data, p, Jespresso.REDUND );             /* belongs to redundant set */
		}

		/* For each cube in Rp, find ways to cover its minterms */
		Jset[] list = Jcofactor.cube3list( D, E, Rp );
		final Jsm_matrix table = Jmatrix.sm_alloc();
		int size_last_dominance = 0;
		int i = 0;

		// data = Rp.data;// java
		for( int p = 0, last = p + Rp.count * Rp.wsize; p < last; p += Rp.wsize ) {
			Rp_current = Jespresso.SIZE( data, p );
			fcube_is_covered( list, data, p, table );
			Jespresso.RESET( data, p, Jespresso.REDUND );	/* can now consider this cube redundant */
			if( (Jglobals.debug & Jespresso.IRRED1) != 0 ) {
				System.out.printf("IRRED1: %d of %d to-go=%d, table=%dx%d time=%s\n",
						i, Rp.count, Rp.count - i,
						table.nrows, table.ncols, Jespresso.print_time( System.currentTimeMillis() ) );
			}
			/* try to keep memory limits down by reducing table as we go along */
			if( table.nrows - size_last_dominance > 1000 ) {
				Jdominate.sm_row_dominance( table );
				size_last_dominance = table.nrows;
				if( (Jglobals.debug & Jespresso.IRRED1) != 0 ) {
					System.out.printf("IRRED1: delete redundant rows, now %dx%d\n",
							table.nrows, table.ncols);
				}
			}
			i++;
		}
		list = null;

		return table;
	}

	/** cube_is_covered -- determine if a cubelist "covers" a single cube */
	static boolean cube_is_covered(final Jset[] T, final int[] c, final int off_c) {
		return tautology( Jcofactor.cofactor( T, c, off_c ) ) != 0;
	}


	private static int taut_level = 0;

	/* tautology -- answer the tautology question for T
	 *
	 * @param T T will be disposed of
	 */
	@SuppressWarnings("boxing")
	static int tautology(final Jset[] T) {

		if( (Jglobals.debug & Jespresso.TAUT) != 0 ) {
			Jcvrout.debug_print( T, "TAUTOLOGY", taut_level++ );
		}

		int result;
		if( (result = taut_special_cases( T )) == Jespresso.MAYBE ) {
			int[] cl = Jespresso.new_cube();
			int[] cr = Jespresso.new_cube();
			final int best = Jcofactor.binate_split_select( T, cl, cr, Jespresso.TAUT );
			result = (tautology( Jcofactor.scofactor( T, cl, best ) ) != 0 &&
					tautology( Jcofactor.scofactor( T, cr, best ) ) != 0 ) ? 1 : 0;// TODO may be TRUE : FALSE ?
			T[0] = null;
			cl = null;
			cr = null;
		}

		if( (Jglobals.debug & Jespresso.TAUT) != 0 ) {
			System.out.printf("exit TAUTOLOGY[%d]: %s\n", --taut_level, Jespresso.print_bool( result ) );
		}
		return result;
	}

	/**
	*  taut_special_cases -- check special cases for tautology
	*
	* @param T will be disposed if answer is determined
	* @return 3 state boolean
	*/
	@SuppressWarnings("boxing")
	static int taut_special_cases(final Jset[] T) {
		final Jcdata_struct cdata = Jglobals.cdata;// java
		final Jcube_struct cube = Jglobals.cube;// java
		Jset p;
		final int[] ceil = cube.temp[0], temp = cube.temp[1];

		/* Check for a row of all 1's which implies tautology */
		for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
			if( Jsetc.full_row( p, T[0] ) ) {
				T[0] = null;
				return Jespresso.TRUE;
			}
		}

		/* Check for a column of all 0's which implies no tautology */
		boolean is_continue = false;
//start:
		do {
			is_continue = false;
			Jset.set_copy( ceil, T[0].base, T[0].offset );
			for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
				Jset.set_or( ceil, ceil, p.base, p.offset );
			}
			if( ! Jset.setp_equal( ceil, cube.fullset ) ) {
				T[0] = null;
				return Jespresso.FALSE;
			}

			/* Collect column counts, determine unate variables, etc. */
			Jcofactor.massive_count( T );

			/* If function is unate (and no row of all 1's), then no tautology */
			if( cdata.vars_unate == cdata.vars_active ) {
				T[0] = null;
				return Jespresso.FALSE;

			/* If active in a single variable (and no column of 0's) then tautology */
			} else if( cdata.vars_active == 1 ) {
				T[0] = null;
				return Jespresso.TRUE;

				/* Check for unate variables, and reduce cover if there are any */
			} else if( cdata.vars_unate != 0 ) {
				/* Form a cube "ceil" with full variables in the unate variables */
				Jset.set_copy( ceil, cube.emptyset );
				for( int var = 0; var < cube.num_vars; var++ ) {
					if( cdata.is_unate[var] ) {
						Jset.set_or( ceil, ceil, cube.var_mask[ var ] );
					}
				}

				/* Save only those cubes that are "full" in all unate variables */
				int Tsave, T1;// java T[ Tsave ]
				for( Tsave = T1 = 2; (p = T[ T1++ ]) != null; ) {
					Jset.set_or( temp, p.base, p.offset, T[0].base, T[0].offset );
					if( Jset.setp_implies( ceil, temp ) ) {
						T[ Tsave++ ] = p;
					}
				}
				T[ Tsave++ ] = null;
				T[ 1 ] = new Jset( null, Tsave );

				if( (Jglobals.debug & Jespresso.TAUT) != 0 ) {
					System.out.printf("UNATE_REDUCTION: %d unate variables, reduced to %d\n",
							cdata.vars_unate, Jespresso.CUBELISTSIZE( T ) );
				}
				is_continue = true;// final goto start;

				/* Check for component reduction */
			} else if( cdata.var_zeros[cdata.best] < Jespresso.CUBELISTSIZE( T ) / 2 ) {
				final Jset[][] A = new Jset[1][];// java holder
				Jset[][] B = new Jset[1][];// java holder
				if( Jcvrm.cubelist_partition( T, A, B, Jglobals.debug & Jespresso.TAUT ) == 0 ) {
					return Jespresso.MAYBE;
				}// else {
					T[0] = null;
					if( tautology( A[0] ) != 0 ) {
						B = null;
						return Jespresso.TRUE;
					}// else {
						return tautology( B[0] );
					//}
				//}
			}
		}
		while( is_continue );

		/* We tried as hard as we could, but must recurse from here on */
		return Jespresso.MAYBE;
	}

	/** fcube_is_covered -- determine exactly how a cubelist "covers" a cube */
	private static void fcube_is_covered(final Jset[] T, final int[] c, final int off, final Jsm_matrix table) {
		ftautology( Jcofactor.cofactor( T, c, off ), table );
	}

	private static int ftaut_level = 0;

	/** ftautology -- find ways to make a tautology
	 *
	 * @param T T will be disposed of
	 */
	@SuppressWarnings("boxing")
	private static void ftautology(final Jset[] T, final Jsm_matrix table) {
		if( (Jglobals.debug & Jespresso.TAUT) != 0 ) {
			Jcvrout.debug_print( T, "FIND_TAUTOLOGY", ftaut_level++ );
		}

		if( ftaut_special_cases( T, table ) == Jespresso.MAYBE ) {
			int[] cl = Jespresso.new_cube();
			int[] cr = Jespresso.new_cube();
			final int best = Jcofactor.binate_split_select( T, cl, cr, Jespresso.TAUT );

			ftautology( Jcofactor.scofactor( T, cl, best ), table );
			ftautology( Jcofactor.scofactor( T, cr, best ), table );

			T[0] = null;
			cl = null;
			cr = null;
		}

		if( (Jglobals.debug & Jespresso.TAUT) != 0 ) {
			System.out.printf("exit FIND_TAUTOLOGY[%d]: table is %d by %d\n",
					--ftaut_level, table.nrows, table.ncols );
		}
	}

	/**
	 *
	 * @param T will be disposed if answer is determined
	 * @param table
	 * @return
	 */
	@SuppressWarnings("boxing")
	private static int ftaut_special_cases(final Jset[] T, final Jsm_matrix table) {
		final Jcube_struct cube = Jglobals.cube;// java
		final Jcdata_struct cdata = Jglobals.cdata;// java
		final int[] temp = cube.temp[0], ceil = cube.temp[1];

		Jset p;
		/* Check for a row of all 1's in the essential cubes */
		for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
			if( ! Jespresso.TESTP( p, Jespresso.REDUND ) ) {
				if( Jsetc.full_row( p, T[0] ) ) {
					/* subspace is covered by essentials -- no new rows for table */
					T[0] = null;
					return Jespresso.TRUE;
				}
			}
		}

		// start:
		do {
			Jcofactor.massive_count( T );

			/* If function is unate, find the rows of all 1's */
			if( cdata.vars_unate == cdata.vars_active ) {
				/* find which nonessentials cover this subspace */
				final int rownum = table.last_row != null ? table.last_row.row_num + 1 : 0;
				Jmatrix.sm_insert( table, rownum, Rp_current );
				for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
					if( Jespresso.TESTP( p, Jespresso.REDUND ) ) {
						/* See if a redundant cube covers this leaf */
						if( Jsetc.full_row( p, T[0] ) ) {
							Jmatrix.sm_insert( table, rownum, (int) Jespresso.SIZE( p ) );
						}
					}
				}
				T[0] = null;
				return Jespresso.TRUE;

				/* Perform unate reduction if there are any unate variables */
			} else if( cdata.vars_unate != 0 ) {
				/* Form a cube "ceil" with full variables in the unate variables */
				Jset.set_copy( ceil, cube.emptyset );
				for( int var = 0; var < cube.num_vars; var++ ) {
					if( cdata.is_unate[ var ] ) {
						Jset.set_or( ceil, ceil, cube.var_mask[ var ] );
					}
				}

				/* Save only those cubes that are "full" in all unate variables */
				int Tsave, T1;
				for( Tsave = T1 = 2; (p = T[ T1++ ]) != null; ) {
					Jset.set_or( temp, p.base, p.offset, T[0].base, T[0].offset );
					if( Jset.setp_implies( ceil, temp ) ) {
						T[ Tsave++ ] = p;
					}
				}
				T[ Tsave++ ] = null;
				T[1] = new Jset( null, Tsave );

				if( (Jglobals.debug & Jespresso.TAUT) != 0 ) {
					System.out.printf("UNATE_REDUCTION: %d unate variables, reduced to %d\n",
							cdata.vars_unate, Jespresso.CUBELISTSIZE( T ) );
				}
			}
		} while( false );

		/* Not much we can do about it */
		return Jespresso.MAYBE;
	}
}