package espresso;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.PushbackInputStream;

/*
 * Revision Control Information
 *
 * $Source$
 * $Author$
 * $Revision$
 * $Date$
 *
 */
/**
 *  Main driver for espresso
 *
 *  Old style -do xxx, -out xxx, etc. are still supported.
 */
public final class Jmain {

	private enum keys {
		KEY_ESPRESSO, KEY_PLA_verify, KEY_check, KEY_contain, KEY_d1merge,
		KEY_disjoint, KEY_dsharp, KEY_echo, KEY_essen, KEY_exact, KEY_expand,
		KEY_gasp, KEY_intersect, KEY_irred, KEY_lexsort, KEY_make_sparse,
		KEY_map, KEY_mapdc, KEY_minterms, KEY_opo, KEY_opoall,
		KEY_pair, KEY_pairall, KEY_primes, KEY_qm, KEY_reduce, KEY_sharp,
		KEY_simplify, KEY_so, KEY_so_both, KEY_stats, KEY_super_gasp, KEY_taut,
		KEY_test, KEY_equiv, KEY_union, KEY_verify, KEY_MANY_ESPRESSO,
		KEY_separate, KEY_xor, KEY_d1merge_in, KEY_fsm, KEY_signature,
		KEY_unknown
	}

	private static class Joption_table {
		private final String name;
		private final keys key;
		private final int num_plas;
		private final boolean needs_offset;
		private final boolean needs_dcset;

		private Joption_table(final String n, final keys k, final int p, final boolean is_off, final boolean is_set) {
			name = n;
			key = k;
			num_plas = p;
			needs_offset = is_off;
			needs_dcset = is_set;
		}
	}
	/** Lookup table for program options */
	private static final Joption_table option_table[] = {
		/* ways to minimize functions */
		new Joption_table("ESPRESSO", keys.KEY_ESPRESSO, 1, true, true),    /* must be first */
		new Joption_table("many", keys.KEY_MANY_ESPRESSO, 1, true, true),
		new Joption_table("exact", keys.KEY_exact, 1, true, true),
		new Joption_table("qm", keys.KEY_qm, 1, true, true),
		new Joption_table("single_output", keys.KEY_so, 1, true, true),
		new Joption_table("so", keys.KEY_so, 1, true, true),
		new Joption_table("so_both", keys.KEY_so_both, 1, true, true),
		new Joption_table("simplify", keys.KEY_simplify, 1, false, false),
		new Joption_table("echo", keys.KEY_echo, 1, false, false),
		new Joption_table("signature", keys.KEY_signature, 1, true, true),

		/* output phase assignment and assignment of inputs to two-bit decoders */
		new Joption_table("opo", keys.KEY_opo, 1, true, true),
		new Joption_table("opoall", keys.KEY_opoall, 1, true, true),
		new Joption_table("pair", keys.KEY_pair, 1, true, true),
		new Joption_table("pairall", keys.KEY_pairall, 1, true, true),

		/* Ways to check covers */
		new Joption_table("check", keys.KEY_check, 1, true, true),
		new Joption_table("stats", keys.KEY_stats, 1, false, false),
		new Joption_table("verify", keys.KEY_verify, 2, false, true),
		new Joption_table("PLAverify", keys.KEY_PLA_verify, 2, false, true),

		/* hacks */
		new Joption_table("equiv", keys.KEY_equiv, 1, true, true),
		new Joption_table("map", keys.KEY_map, 1, false, false),
		new Joption_table("mapdc", keys.KEY_mapdc, 1, false, false),
		new Joption_table("fsm", keys.KEY_fsm, 1, false, true),

		/* the basic boolean operations on covers */
		new Joption_table("contain", keys.KEY_contain, 1, false, false),
		new Joption_table("d1merge", keys.KEY_d1merge, 1, false, false),
		new Joption_table("d1merge_in", keys.KEY_d1merge_in, 1, false, false),
		new Joption_table("disjoint", keys.KEY_disjoint, 1, true, false),
		new Joption_table("dsharp", keys.KEY_dsharp, 2, false, false),
		new Joption_table("intersect", keys.KEY_intersect, 2, false, false),
		new Joption_table("minterms", keys.KEY_minterms, 1, false, false),
		new Joption_table("primes", keys.KEY_primes, 1, false, true),
		new Joption_table("separate", keys.KEY_separate, 1, true, true),
		new Joption_table("sharp", keys.KEY_sharp, 2, false, false),
		new Joption_table("union", keys.KEY_union, 2, false, false),
		new Joption_table("xor", keys.KEY_xor, 2, true, true),

		/* debugging only -- call each step of the espresso algorithm */
		new Joption_table("essen", keys.KEY_essen, 1, false, true),
		new Joption_table("expand", keys.KEY_expand, 1, true, false),
		new Joption_table("gasp", keys.KEY_gasp, 1, true, true),
		new Joption_table("irred", keys.KEY_irred, 1, false, true),
		new Joption_table("make_sparse", keys.KEY_make_sparse, 1, true, true),
		new Joption_table("reduce", keys.KEY_reduce, 1, false, true),
		new Joption_table("taut", keys.KEY_taut, 1, false, false),
		new Joption_table("super_gasp", keys.KEY_super_gasp, 1, true, true),
		new Joption_table("lexsort", keys.KEY_lexsort, 1, false, false),
		new Joption_table("test", keys.KEY_test, 1, true, true),
		new Joption_table( null, keys.KEY_unknown, 0, false, false)             /* must be last */
	};


	private static final class Jdebug_struct {
		private final String name;
		private final int value;
		private Jdebug_struct(final String n, final int v) {
			name = n;
			value = v;
		}
	}
	private static final Jdebug_struct debug_table[] = {
		new Jdebug_struct("", Jespresso.EXPAND + Jespresso.ESSEN + Jespresso.IRRED + Jespresso.REDUCE + Jespresso.SPARSE + Jespresso.GASP + Jespresso.SHARP + Jespresso.MINCOV),
		new Jdebug_struct("compl",  Jespresso.COMPL),  new Jdebug_struct("essen",   Jespresso.ESSEN),
		new Jdebug_struct("expand", Jespresso.EXPAND),  new Jdebug_struct("expand1", Jespresso.EXPAND1|Jespresso.EXPAND),
		new Jdebug_struct("irred",  Jespresso.IRRED),  new Jdebug_struct("irred1",  Jespresso.IRRED1|Jespresso.IRRED),
		new Jdebug_struct("reduce", Jespresso.REDUCE),  new Jdebug_struct("reduce1", Jespresso.REDUCE1|Jespresso.REDUCE),
		new Jdebug_struct("mincov", Jespresso.MINCOV),  new Jdebug_struct("mincov1", Jespresso.MINCOV1|Jespresso.MINCOV),
		new Jdebug_struct("sparse", Jespresso.SPARSE),  new Jdebug_struct("sharp",   Jespresso.SHARP),
		new Jdebug_struct("taut",   Jespresso.TAUT),  new Jdebug_struct("gasp",    Jespresso.GASP),
		new Jdebug_struct("exact",  Jespresso.EXACT),
		new Jdebug_struct( null,  0 )
	};

/* java: inplace
	private static final class Jesp_opt {
		private final String name;
		private final int[] variable;
		private final int value;
		private Jesp_opt(final String n, final int[] vars, final int v) {
			name = n;
			variable = vars;
			value = v;
		}
	}
	private static final Jesp_opt esp_opt_table[] = {
		new Jesp_opt("eat", Jglobals.echo_comments, false),
		new Jesp_opt("eatdots", Jglobals.echo_unknown_commands, false),
		new Jesp_opt("fast", Jglobals.single_expand, true),
		new Jesp_opt("kiss", Jglobals.kiss, true),
		new Jesp_opt("ness", Jglobals.remove_essential, false),
		new Jesp_opt("nirr", Jglobals.force_irredundant, false),
		new Jesp_opt("nunwrap", Jglobals.unwrap_onset, false),
		new Jesp_opt("onset", Jglobals.recompute_onset, true),
		new Jesp_opt("pos", Jglobals.pos, true),
		new Jesp_opt("random", Jglobals.use_random_order, true),
		new Jesp_opt("strong", Jglobals.use_super_gasp, true)
	};
*/

	private static PushbackInputStream last_fp = null;// TODO java check closing the stream
	private static int input_type = Jespresso.FD_type;

	static void closeStreams() {// java added to fix the c code bug
		if( last_fp != null ) {
			try {
				last_fp.close();
			} catch( final IOException e ) {
			}
		}
	}

	/**
	 * The app entry point
	 *
	 * @param argv
	 */
	public static void main(final String[] argv)
	{
		final int argc = argv.length;// java
		int i, j, first, last, strategy, out_type, option;
		JpPLA[] PLA, PLA1;
		Jset_family F;
		Jset_family Fold;
		Jset_family Dold;
		final Jcost_t cost = new Jcost_t();
		boolean exact_cover;
		// extern int optind;// java

		final long start = System.currentTimeMillis();

		boolean error = false;
		init_runtime();
if( Jespresso.RANDOM ) {
		Jespresso.random.setSeed( 314973 );
}

		option = 0;			/* default -D: ESPRESSO */
		out_type = Jespresso.F_type;		/* default -o: default is ON-set only */
		Jglobals.debug = 0;			/* default -d: no debugging info */
		Jglobals.verbose_debug = false;	/* default -v: not verbose */
		Jglobals.print_solution = true;	/* default -x: print the solution (!) */
		Jglobals.summary = false;		/* default -s: no summary */
		Jglobals.trace = false;		/* default -t: no trace information */
		strategy = 0;		/* default -S: strategy number */
		first = -1;			/* default -R: select range */
		last = -1;
		Jglobals.remove_essential = true;	/* default -e: */
		Jglobals.force_irredundant = true;
		Jglobals.unwrap_onset = true;
		Jglobals.single_expand = false;
		Jglobals.pos = false;
		Jglobals.recompute_onset = false;
		Jglobals.use_super_gasp = false;
		Jglobals.use_random_order = false;
		Jglobals.kiss = false;
		Jglobals.echo_comments = true;
		Jglobals.echo_unknown_commands = true;
		exact_cover = false;	/* for -qm option, the default */

		// backward_compatibility_hack( &argc, argv, &option, &out_type );

		/* parse command line options*/
		final Jgetopt getopt = new Jgetopt();
		while( (i = getopt.getopt( argc, argv, "D:S:de:o:r:stv:x" )) != -1 ) {
			final String optarg = getopt.getString();
			switch( i ) {
			case 'D':		/* -Dcommand invokes a subcommand */
				for( j = 0; option_table[j].name != null; j++ ) {
					if( optarg.compareTo( option_table[j].name ) == 0 ) {
						option = j;
						break;
					}
				}
				if( option_table[j].name == null ) {
					System.err.printf("%s: bad subcommand \"%s\"\n",
							argv[0], optarg);
					closeStreams();
					System.exit( 1 );
				}
				break;

			case 'o':		/* -ooutput selects and output option */
				for( j = 0; j < Jglobals.pla_types.length; j++ ) {
					if( optarg.compareTo( Jglobals.pla_types[ j ].key.substring( 1 ) ) == 0 ) {
						out_type = Jglobals.pla_types[ j ].value;
						break;
					}
				}
				if( j >= Jglobals.pla_types.length ) {
					System.err.printf("%s: bad output type \"%s\"\n", argv[0], optarg );
					closeStreams();
					System.exit( 1 );
				}
				break;

			case 'e':		/* -eespresso selects an option for espresso */
				switch( optarg ) {
				case "eat": Jglobals.echo_comments = false; break;
				case "eatdots": Jglobals.echo_unknown_commands = false; break;
				case "fast": Jglobals.single_expand = true; break;
				case "kiss": Jglobals.kiss = true; break;
				case "ness": Jglobals.remove_essential = false; break;
				case "nirr": Jglobals.force_irredundant = false; break;
				case "nunwrap": Jglobals.unwrap_onset = false; break;
				case "onset": Jglobals.recompute_onset = true; break;
				case "pos": Jglobals.pos = true; break;
				case "random": Jglobals.use_random_order = true; break;
				case "strong": Jglobals.use_super_gasp = true; break;
				default:
					System.err.printf("%s: bad espresso option \"%s\"\n", argv[0], optarg );
					closeStreams();
					System.exit( 1 );
					break;
				}
				break;

			case 'd':		/* -d turns on (softly) all debug switches */
				Jglobals.debug = debug_table[ 0 ].value;
				Jglobals.trace = true;
				Jglobals.summary = true;
				break;

			case 'v':		/* -vdebug invokes a debug option */
				Jglobals.verbose_debug = true;
				for( j = 0; debug_table[j].name != null; j++ ) {
					if( optarg.compareTo( debug_table[j].name ) == 0 ) {
						Jglobals.debug |= debug_table[j].value;
						break;
					}
				}
				if( debug_table[j].name == null ) {
					System.err.printf("%s: bad debug type \"%s\"\n",
							argv[0], optarg);
					closeStreams();
					System.exit( 1 );
				}
				break;

			case 't':
				Jglobals.trace = true;
				break;

			case 's':
				Jglobals.summary = true;
				break;

			case 'x':		/* -x suppress printing of results */
				Jglobals.print_solution = false;
				break;

			case 'S':		/* -S sets a strategy for several cmds */
				try {
					strategy = Integer.parseInt( optarg );
				} catch(final NumberFormatException ne) {
					strategy = 0;
				}
				break;

			case 'r':		/* -r selects range (outputs or vars) */
				try {// "%d-%d"
					final String[] s = optarg.split("-");
					if( s.length < 2 ) {
						throw new NumberFormatException();
					}
					first = Integer.parseInt( s[ 0 ] );
					last = Integer.parseInt( s[ 1 ] );
				} catch(final NumberFormatException ne) {
					System.err.printf("%s: bad output range \"%s\"\n", argv[0], optarg );
					closeStreams();
					System.exit( 1 );
				}
				break;

			default:
				usage();
				closeStreams();
				System.exit( 1 );
			}
		}

		/* provide version information and summaries */
		if( Jglobals.summary || Jglobals.trace ) {
			/* echo command line and arguments */
			System.out.printf("#");
			for( i = 0; i < argc; i++ ) {
				System.out.printf(" %s", argv[ i ] );
			}
			System.out.printf("\n");
			System.out.printf("# %s\n", Jespresso.VERSION );
		}

		/* the remaining arguments are argv[optind ... argc-1] */
		PLA = new JpPLA[ 1 ];// PLA = PLA1 = null;
		PLA1 = new JpPLA[ 1 ];// PLA = PLA1 = null;
		int optind = getopt.getIndex();// java
		switch( option_table[ option ].num_plas ) {
		case 2:
			if( optind + 2 + 1 < argc ) {// +1 java
				Jcvrmisc.fatal("trailing arguments on command line");
			}
			getPLA( optind++, argc, argv, option, PLA, out_type );
			getPLA( optind++, argc, argv, option, PLA1, out_type );
			break;
		case 1:
			if( optind + 1 + 1 < argc ) {// +1 java
				Jcvrmisc.fatal("trailing arguments on command line");
			}
			getPLA( optind++, argc, argv, option, PLA, out_type );
			break;
		}
		if( optind + 1 < argc ) {// +1 java
			Jcvrmisc.fatal("trailing arguments on command line");
		}

		if( Jglobals.summary || Jglobals.trace ) {
			if( PLA[ 0 ] != null ) {
				Jcvrin.PLA_summary( PLA[ 0 ] );
			}
			if( PLA1[ 0 ] != null ) {
				Jcvrin.PLA_summary( PLA1[ 0 ] );
			}
		}

		/* Now a case-statement to decide what to do */
		switch( option_table[ option ].key ) {
			case KEY_unknown:// java added to suppress the warning
				break;
			/******************** Espresso operations ********************/
			case KEY_ESPRESSO:
				Fold = Jset.sf_save( PLA[ 0 ].F );
				PLA[ 0 ].F = Jespresso.espresso( PLA[ 0 ].F, PLA[ 0 ].D, PLA[ 0 ].R );
			    {
			    	final long t = System.currentTimeMillis();
			    	error = Jverify.verify( PLA[ 0 ].F, Fold, PLA[ 0 ].D );
			    	Jcvrmisc.totals( t, Jespresso.VERIFY_TIME, PLA[ 0 ].F, cost );
			    }
				if( error ) {
					Jglobals.print_solution = false;
					PLA[ 0 ].F = Fold;
					Jverify.check_consistency( PLA[ 0 ] );
				} else {
					Jespresso.free_cover( Fold );
				}
				break;

			case KEY_MANY_ESPRESSO: {
				int pla_type;
				do {
				    {
				    	final long t = System.currentTimeMillis();
				    	PLA[ 0 ].F = Jespresso.espresso( PLA[ 0 ].F, PLA[ 0 ].D, PLA[ 0 ].R );
				    	if( Jglobals.trace ) {
				    		Jcvrmisc.print_trace( PLA[ 0 ].F, "ESPRESSO   ", System.currentTimeMillis() - t);
				    	}
				    }
					if( Jglobals.print_solution ) {
						Jcvrout.fprint_pla( System.out, PLA[ 0 ], out_type );
						System.out.flush();
					}
					pla_type = PLA[ 0 ].pla_type;
					Jcvrin.free_PLA( PLA[ 0 ] );
					Jcubestr.setdown_cube();
					Jglobals.cube.part_size = null;
				} while( Jcvrin.read_pla( last_fp, true, true, pla_type, PLA ) != -1 );
				closeStreams();// FIXME PLA and PLA1 must be free
				System.exit( 0 );
			}

			case KEY_simplify:
			    {
			    	final long t = System.currentTimeMillis();
			    	PLA[ 0 ].F = Jcompl.simplify( Jcofactor.cube1list( PLA[ 0 ].F ) );
			    	if( Jglobals.trace ) {
			    		Jcvrmisc.print_trace( PLA[ 0 ].F, "SIMPLIFY  ", System.currentTimeMillis() - t);
			    	}
			    }
				break;

			case KEY_so:	        /* minimize all functions as single-output */
				if( strategy < 0 || strategy > 1 ) {
					strategy = 0;
				}
				Jcvrm.so_espresso( PLA[ 0 ], strategy );
				break;

			case KEY_so_both:		/* minimize all functions as single-output */
				if( strategy < 0 || strategy > 1 ) {
					strategy = 0;
				}
				Jcvrm.so_both_espresso( PLA[ 0 ], strategy );
				break;

			case KEY_expand:            /* execute expand */
			    {
			    	final long t = System.currentTimeMillis();
			    	PLA[ 0 ].F = Jexpand.expand( PLA[ 0 ].F, PLA[ 0 ].R, false );
			    	Jcvrmisc.totals( t, Jespresso.EXPAND_TIME, PLA[ 0 ].F, cost );
			    }
				break;

			case KEY_irred:             /* extract minimal irredundant subset */
			    {
			    	final long t = System.currentTimeMillis();
			    	PLA[ 0 ].F = Jirred.irredundant( PLA[ 0 ].F, PLA[ 0 ].D );
			    	Jcvrmisc.totals( t, Jespresso.IRRED_TIME, PLA[ 0 ].F, cost );
			    }
				break;

			case KEY_reduce:            /* perform reduction */
			    {
			    	final long t = System.currentTimeMillis();
			    	PLA[ 0 ].F = Jreduce.reduce( PLA[ 0 ].F, PLA[ 0 ].D );
			    	Jcvrmisc.totals( t, Jespresso.REDUCE_TIME, PLA[ 0 ].F, cost );
			    }
				break;

			case KEY_essen:             /* check for essential primes */
				{
					final int[] data = PLA[ 0 ].F.data;// java
					for( int p = 0, last1 = p + PLA[ 0 ].F.count * PLA[ 0 ].F.wsize; p < last1; p += PLA[ 0 ].F.wsize ) {
						Jespresso.SET( data, p, Jespresso.RELESSEN );
						Jespresso.RESET( data, p, Jespresso.NONESSEN );
					}
				    {
				    	final long t = System.currentTimeMillis();
				    	final Jset_family f_holder[] = new Jset_family[] { PLA[ 0 ].F };// java
				    	final Jset_family d_holder[] = new Jset_family[] { PLA[ 0 ].D };// java
				    	F = Jessen.essential( f_holder, d_holder );
				    	PLA[ 0 ].F = f_holder[ 0 ];// java
				    	PLA[ 0 ].D = d_holder[ 0 ];// java
				    	Jcvrmisc.totals( t, Jespresso.ESSEN_TIME, PLA[ 0 ].F, cost );
				    }
					Jespresso.free_cover( F );
				}
				break;

			case KEY_super_gasp:
				PLA[ 0 ].F = Jgasp.super_gasp( PLA[ 0 ].F, PLA[ 0 ].D, PLA[ 0 ].R, cost );
				break;

			case KEY_gasp:
				PLA[ 0 ].F = Jgasp.last_gasp( PLA[ 0 ].F, PLA[ 0 ].D, PLA[ 0 ].R, cost );
				break;

			case KEY_make_sparse:       /* make_sparse step of Espresso */
				PLA[ 0 ].F = Jsparse.make_sparse( PLA[ 0 ].F, PLA[ 0 ].D, PLA[ 0 ].R );
				break;

			case KEY_exact:
				exact_cover = true;

			case KEY_qm:
				Fold = Jset.sf_save( PLA[ 0 ].F );
				PLA[ 0 ].F = Jexact.minimize_exact( PLA[ 0 ].F, PLA[ 0 ].D, PLA[ 0 ].R, exact_cover );
			    {
			    	final long t = System.currentTimeMillis();
			    	error = Jverify.verify( PLA[ 0 ].F, Fold, PLA[ 0 ].D );
			    	Jcvrmisc.totals( t, Jespresso.VERIFY_TIME, PLA[ 0 ].F, cost );
			    }
				if( error ) {
					Jglobals.print_solution = false;
					PLA[ 0 ].F = Fold;
					Jverify.check_consistency( PLA[ 0 ] );
				}
				Jespresso.free_cover( Fold );
				break;

			case KEY_primes:            /* generate all prime implicants */
			    {
			    	final long t = System.currentTimeMillis();
			    	PLA[ 0 ].F = Jprimes.primes_consensus( Jcofactor.cube2list( PLA[ 0 ].F, PLA[ 0 ].D ) );
			    	if( Jglobals.trace ) {
			    		Jcvrmisc.print_trace( PLA[ 0 ].F, "PRIMES     ", System.currentTimeMillis() - t);
			    	}
			    }
				break;

			case KEY_map:               /* print out a Karnaugh map of function */
				Jmap.map( PLA[ 0 ].F );
				Jglobals.print_solution = false;
				break;

			case KEY_signature:
				Fold = Jset.sf_save( PLA[ 0 ].F );
				PLA[ 0 ].F = Jsignature.signature( PLA[ 0 ].F, PLA[ 0 ].D, PLA[ 0 ].R );
			    {
			    	final long t = System.currentTimeMillis();
			    	error = Jverify.verify( PLA[ 0 ].F, Fold, PLA[ 0 ].D );
			    	Jcvrmisc.totals( t, Jespresso.VERIFY_TIME, PLA[ 0 ].F, cost );
			    }
				if( error ) {
					Jglobals.print_solution = false;
					PLA[ 0 ].F = Fold;
					Jverify.check_consistency( PLA[ 0 ] );
				} else {
					Jespresso.free_cover( Fold );
				}
				break;

		/******************** Output phase and bit pairing ********************/

			case KEY_opo:               /* sasao output phase assignment */
				Jopo.phase_assignment( PLA[ 0 ], strategy );
				break;

			case KEY_opoall:		/* try all phase assignments (!) */
				if( first < 0 || first >= Jglobals.cube.part_size[ Jglobals.cube.output ] ) {
					first = 0;
				}
				if( last < 0 || last >= Jglobals.cube.part_size[ Jglobals.cube.output ] ) {
					last = Jglobals.cube.part_size[ Jglobals.cube.output ] - 1;
				}
				Jopo.opoall( PLA[ 0 ], first, last, strategy );
				break;

			case KEY_pair:              /* find an optimal pairing */
				Jpair.find_optimal_pairing( PLA[ 0 ], strategy );
				break;

			case KEY_pairall:		/* try all pairings !! */
				Jpair.pair_all( PLA[ 0 ], strategy );
				break;

		/******************** Simple cover operations ********************/

			case KEY_echo:				/* echo the PLA */
				break;

			case KEY_taut:				/* tautology check */
				System.out.printf("ON-set is%sa tautology\n",
						Jirred.tautology( Jcofactor.cube1list( PLA[ 0 ].F ) ) != 0 ? " " : " not ");
				Jglobals.print_solution = false;
				break;

			case KEY_contain:				/* single cube containment */
				PLA[ 0 ].F = Jcontain.sf_contain( PLA[ 0 ].F );
				break;

			case KEY_intersect:				/* cover intersection */
				PLA[ 0 ].F = Jsharp.cv_intersect( PLA[ 0 ].F, PLA1[ 0 ].F );
				break;

			case KEY_union:				/* cover union */
				PLA[ 0 ].F = Jcontain.sf_union( PLA[ 0 ].F, PLA1[ 0 ].F );
				break;

			case KEY_disjoint:				/* make cover disjoint */
				PLA[ 0 ].F = Jsharp.make_disjoint( PLA[ 0 ].F );
				break;

			case KEY_dsharp:				/* cover disjoint-sharp */
				PLA[ 0 ].F = Jsharp.cv_dsharp( PLA[ 0 ].F, PLA1[ 0 ].F );
				break;

			case KEY_sharp:				/* cover sharp */
				PLA[ 0 ].F = Jsharp.cv_sharp( PLA[ 0 ].F, PLA1[ 0 ].F );
				break;

			case KEY_lexsort:				/* lexical sort order */
				PLA[ 0 ].F = Jcvrm.lex_sort(PLA[ 0 ].F);
				break;

			case KEY_stats:				/* print info on size */
				if( ! Jglobals.summary ) {
					Jcvrin.PLA_summary( PLA[ 0 ] );
				}
				Jglobals.print_solution = false;
				break;

			case KEY_minterms:				/* explode into minterms */
				if( first < 0 || first >= Jglobals.cube.num_vars ) {
					first = 0;
				}
				if( last < 0 || last >= Jglobals.cube.num_vars ) {
					last = Jglobals.cube.num_vars - 1;
				}
				PLA[ 0 ].F = Jcontain.sf_dupl( Jcvrm.unravel_range( PLA[ 0 ].F, first, last ) );
				break;

			case KEY_d1merge:				/* distance-1 merge */
				if( first < 0 || first >= Jglobals.cube.num_vars ) {
					first = 0;
				}
				if( last < 0 || last >= Jglobals.cube.num_vars ) {
					last = Jglobals.cube.num_vars - 1;
				}
				for( i = first; i <= last; i++ ) {
					PLA[ 0 ].F = Jcontain.d1merge( PLA[ 0 ].F, i );
				}
				break;

			case KEY_d1merge_in:		/* distance-1 merge inputs only */
				for( i = 0; i < Jglobals.cube.num_binary_vars; i++ ) {
					PLA[ 0 ].F = Jcontain.d1merge( PLA[ 0 ].F, i );
				}
				break;

			case KEY_PLA_verify:		/* check two PLAs for equivalence */
			    {
			    	final long t = System.currentTimeMillis();
			    	error = Jverify.PLA_verify( PLA[ 0 ], PLA1[ 0 ] );
			    	Jcvrmisc.totals( t, Jespresso.VERIFY_TIME, PLA[ 0 ].F, cost );
			    }
				if( error ) {// FIXME PLA and PLA1 must be free
					System.out.printf("PLA comparison failed; the PLA's are not equivalent\n");
					closeStreams();
					System.exit( 1 );
				} else {
					System.out.printf("PLA's compared equal\n");
					closeStreams();
					System.exit( 0 );
				}
				break;	/* silly */

			case KEY_verify:			/* check two covers for equivalence */
				Fold = PLA[ 0 ].F;	Dold = PLA[ 0 ].D;	F = PLA1[ 0 ].F;
			    {
			    	final long t = System.currentTimeMillis();
			    	error = Jverify.verify( F, Fold, Dold );
			    	Jcvrmisc.totals( t, Jespresso.VERIFY_TIME, PLA[ 0 ].F, cost );
			    }
				if( error ) {// FIXME PLA and PLA1 must be free
					System.out.printf("PLA comparison failed; the PLA's are not equivalent\n");
					closeStreams();
					System.exit( 1 );
				} else {
					System.out.printf("PLA's compared equal\n");
					closeStreams();
					System.exit( 0 );
				}
				break;	/* silly */

			case KEY_check:			/* check consistency */
				Jverify.check_consistency( PLA[ 0 ] );
				Jglobals.print_solution = false;
				break;

			case KEY_mapdc:			/* compute don't care set */
				Jhack.map_dcset( PLA[ 0 ] );
				out_type = Jespresso.FD_type;
				break;

			case KEY_equiv:
				Jequiv.find_equiv_outputs( PLA[ 0 ] );
				Jglobals.print_solution = false;
				break;

			case KEY_separate:			/* remove PLA[ 0 ].D from PLA[ 0 ].F */
				PLA[ 0 ].F = Jcompl.complement( Jcofactor.cube2list( PLA[ 0 ].D, PLA[ 0 ].R ) );
				break;

			case KEY_xor: {
				final Jset_family T1 = Jsharp.cv_intersect( PLA[ 0 ].F, PLA1[ 0 ].R );
				final Jset_family T2 = Jsharp.cv_intersect( PLA1[ 0 ].F, PLA[ 0 ].R );
				Jespresso.free_cover( PLA[ 0 ].F );
				PLA[ 0 ].F = Jcontain.sf_contain( Jset.sf_join( T1, T2 ) );
				Jespresso.free_cover( T1 );
				Jespresso.free_cover( T2 );
				break;
			}

			case KEY_fsm: {
				Jhack.disassemble_fsm( PLA[ 0 ], Jglobals.summary );
				Jglobals.print_solution = false;
				break;
			}

			case KEY_test: {
				Jset_family T, E;
				T = Jset.sf_join( PLA[ 0 ].D, PLA[ 0 ].R );
				E = Jespresso.new_cover( 10 );
				Jset.sf_free( PLA[ 0 ].F );
			    {
			    	final long t = System.currentTimeMillis();
			    	PLA[ 0 ].F = Jcompl.complement( Jcofactor.cube1list( T ) );
			    	Jcvrmisc.totals( t, Jespresso.COMPL_TIME, PLA[ 0 ].F, cost );
			    }
			    {
			    	final long t = System.currentTimeMillis();
			    	PLA[ 0 ].F = Jexpand.expand( PLA[ 0 ].F, T, false );
			    	Jcvrmisc.totals( t, Jespresso.EXPAND_TIME, PLA[ 0 ].F, cost );
			    }
			    {
			    	final long t = System.currentTimeMillis();
			    	PLA[ 0 ].F = Jirred.irredundant( PLA[ 0 ].F, E );
			    	Jcvrmisc.totals( t, Jespresso.IRRED_TIME, PLA[ 0 ].F, cost );
			    }
				Jset.sf_free( T );
				T = Jset.sf_join(PLA[ 0 ].F, PLA[ 0 ].R);
			    {
			    	final long t = System.currentTimeMillis();
			    	PLA[ 0 ].D = Jexpand.expand( PLA[ 0 ].D, T, false );
			    	Jcvrmisc.totals( t, Jespresso.EXPAND_TIME, PLA[ 0 ].D, cost );
			    }
			    {
			    	final long t = System.currentTimeMillis();
			    	PLA[ 0 ].D = Jirred.irredundant( PLA[ 0 ].D, E );
			    	Jcvrmisc.totals( t, Jespresso.IRRED_TIME, PLA[ 0 ].D, cost );
			    }
				Jset.sf_free( T );
				Jset.sf_free( E );
				break;
			}
		}

		/* Print a runtime summary if trace mode enabled */
		if( Jglobals.trace ) {
			runtime();
		}

		/* Print total runtime */
		if( Jglobals.summary || Jglobals.trace ) {
			Jcvrmisc.print_trace( PLA[ 0 ].F, option_table[ option ].name, System.currentTimeMillis() - start );
		}

		/* Output the solution */
		if( Jglobals.print_solution ) {
		    {
		    	final long t = System.currentTimeMillis();
		    	Jcvrout.fprint_pla( System.out, PLA[ 0 ], out_type );
		    	Jcvrmisc.totals( t, Jespresso.WRITE_TIME, PLA[ 0 ].F, cost );
		    }
		}

		/* Crash and burn if there was a verify error */
		if( error ) {
			Jcvrmisc.fatal("cover verification failed");
		}

		/* cleanup all used memory */
		if( PLA[ 0 ] != null ) {
			Jcvrin.free_PLA( PLA[ 0 ] );
		}
		if( PLA1[ 0 ] != null ) {
			Jcvrin.free_PLA( PLA1[ 0 ] );// FIXME PLA1 must be free
		}
		Jglobals.cube.part_size = null;
		Jcubestr.setdown_cube();             /* free the cube/cdata structure data */
		Jset.sf_cleanup();               /* free unused set structures */
		Jmatrix.sm_cleanup();               /* sparse matrix cleanup */

		closeStreams();
		System.exit( 0 );
	}


	private static void getPLA(final int opt, final int argc, final String[] argv, final int option, final JpPLA[] PLA, final int out_type)
	{
		PushbackInputStream fp = null;
		String fname;

		if( opt >= argc ) {
			fp = new PushbackInputStream( System.in );
			fname = "(stdin)";
		} else {
			fname = argv[opt];
			if( fname.compareTo("-") == 0 ) {
				fp = new PushbackInputStream( System.in );
			} else {
				try {
					fp = new PushbackInputStream( new FileInputStream( argv[opt] ) );
				} catch(final Exception e) {
					System.err.printf("%s: Unable to open %s\n", argv[0], fname );
					Jmain.closeStreams();
					System.exit( 1 );
				}
			}
		}
		boolean needs_dcset, needs_offset;
		if( option_table[option].key == keys.KEY_echo ) {
			needs_dcset = (out_type & Jespresso.D_type) != 0;
			needs_offset = (out_type & Jespresso.R_type) != 0;
		} else {
			needs_dcset = option_table[ option ].needs_dcset;
			needs_offset = option_table[ option ].needs_offset;
		}

		if( Jcvrin.read_pla( fp, needs_dcset, needs_offset, input_type, PLA ) == -1 ) {
			System.err.printf("%s: Unable to find PLA on file %s\n", argv[0], fname);
			Jmain.closeStreams();
			System.exit( 1 );
		}
		PLA[ 0 ].filename = fname;
		Jglobals.filename = PLA[ 0 ].filename;
		/*    (void) fclose(fp);*/
		/* hackto support -Dmany */
		last_fp = fp;// FIXME the input stream must be closed !
	}

	@SuppressWarnings("boxing")
	private static void runtime( ) {
		final long[] total_time = Jglobals.total_time;// java
		final int[] total_calls = Jglobals.total_calls;// java
		final String[] total_name = Jglobals.total_name;// java
		long total = 1, temp;

		for( int i = 0; i < Jespresso.TIME_COUNT; i++ ) {
			total += total_time[ i ];
		}
		for( int i = 0; i < Jespresso.TIME_COUNT; i++ ) {
			if( total_calls[i] != 0 ) {
				temp = 100 * total_time[ i ];
				System.out.printf("# %s\t%2d call(s) for %s (%2d.%01d%%)\n",
						total_name[ i ], total_calls[ i ], Jespresso.print_time( total_time[ i ] ),
						temp / total, (10 * (temp % total)) / total );
			}
		}
	}

	private static void init_runtime( ) {
		final String total_name[] = Jglobals.total_name;// java
		total_name[Jespresso.READ_TIME] =     "READ       ";
		total_name[Jespresso.WRITE_TIME] =    "WRITE      ";
		total_name[Jespresso.COMPL_TIME] =    "COMPL      ";
		total_name[Jespresso.REDUCE_TIME] =   "REDUCE     ";
		total_name[Jespresso.EXPAND_TIME] =   "EXPAND     ";
		total_name[Jespresso.ESSEN_TIME] =    "ESSEN      ";
		total_name[Jespresso.IRRED_TIME] =    "IRRED      ";
		total_name[Jespresso.GREDUCE_TIME] =  "REDUCE_GASP";
		total_name[Jespresso.GEXPAND_TIME] =  "EXPAND_GASP";
		total_name[Jespresso.GIRRED_TIME] =   "IRRED_GASP ";
		total_name[Jespresso.MV_REDUCE_TIME] ="MV_REDUCE  ";
		total_name[Jespresso.RAISE_IN_TIME] = "RAISE_IN   ";
		total_name[Jespresso.VERIFY_TIME] =   "VERIFY     ";
		total_name[Jespresso.PRIMES_TIME] =   "PRIMES     ";
		total_name[Jespresso.MINCOV_TIME] =   "MINCOV     ";
		// FIXME 22 names, 18 are used, but only 15 are initialized
		total_name[Jespresso.ONSET_TIME] =    "ONSET      ";// is not used
		total_name[Jespresso.FCC_TIME] =      "FCC        ";
		total_name[Jespresso.ETR_TIME] =      "ETR        ";
		total_name[Jespresso.ETRAUX_TIME] =   "ETRAUX     ";

	}

	private static void subcommands( ) {
		System.out.printf("                ");
		int col = 16;
		for( int i = 0; option_table[i].name != null; i++ ) {
			if( (col + option_table[i].name.length() + 1) > 76 ) {
				System.out.printf(",\n                ");
				col = 16;
			} else if( i != 0 ) {
				System.out.printf(", ");
			}
			System.out.printf("%s", option_table[ i ].name );
			col += option_table[ i ].name.length() + 2;
		}
		System.out.printf("\n");
	}


	private static void usage( ) {
		System.out.printf("%s\n\n", Jespresso.VERSION);
		System.out.printf("SYNOPSIS: espresso [options] [file]\n\n");
		System.out.printf("  -d        Enable debugging\n");
		System.out.printf("  -e[opt]   Select espresso option:\n");
		System.out.printf("                fast, ness, nirr, nunwrap, onset, pos, strong,\n");
		System.out.printf("                eat, eatdots, kiss, random\n");
		System.out.printf("  -o[type]  Select output format:\n");
		System.out.printf("                f, fd, fr, fdr, pleasure, eqntott, kiss, cons\n");
		System.out.printf("  -rn-m     Select range for subcommands:\n");
		System.out.printf("                d1merge: first and last variables (0 ... m-1)\n");
		System.out.printf("                minterms: first and last variables (0 ... m-1)\n");
		System.out.printf("                opoall: first and last outputs (0 ... m-1)\n");
		System.out.printf("  -s        Provide short execution summary\n");
		System.out.printf("  -t        Provide longer execution trace\n");
		System.out.printf("  -x        Suppress printing of solution\n");
		System.out.printf("  -v[type]  Verbose debugging detail (-v '' for all)\n");
		System.out.printf("  -D[cmd]   Execute subcommand 'cmd':\n");
		subcommands();
		System.out.printf("  -Sn       Select strategy for subcommands:\n");
		System.out.printf("                opo: bit2=exact bit1=repeated bit0=skip sparse\n");
		System.out.printf("                opoall: 0=minimize, 1=exact\n");
		System.out.printf("                pair: 0=algebraic, 1=strongd, 2=espresso, 3=exact\n");
		System.out.printf("                pairall: 0=minimize, 1=exact, 2=opo\n");
		System.out.printf("                so_espresso: 0=minimize, 1=exact\n");
		System.out.printf("                so_both: 0=minimize, 1=exact\n");
	}

	/**
	 *  Hack for backward compatibility (ACK! )
	 */
	/* private static void backward_compatibility_hack(final int[] argc, final String[][] argv, int[] option, int[] out_type ) {
		// Scan the argument list for something to do (default is ESPRESSO)
		*option = 0;
		for( int i = 1; i < (*argc) - 1; i++ ) {
			if( strcmp(argv[i], "-do") == 0 ) {
				for( int j = 0; option_table[j].name != 0; j++) {
					if( strcmp(argv[i+1], option_table[j].name) == 0 ) {
						*option = j;
						delete_arg(argc, argv, i+1);
						delete_arg(argc, argv, i);
						break;
					}
				}
					if( option_table[j].name == 0 ) {
						System.err.printf(
								"espresso: bad keyword \"%s\" following -do\n",argv[i+1]);
						System.exit( 1 );
					}
				break;
			}
		}

		for( int i = 1; i < (*argc)-1; i++ ) {
			if( strcmp(argv[i], "-out") == 0 ) {
				for( int j = 0; pla_types[j].key != 0; j++ ) {
					if( strcmp(pla_types[j].key+1, argv[i+1]) == 0 ) {
						*out_type = pla_types[j].value;
						delete_arg(argc, argv, i+1);
						delete_arg(argc, argv, i);
						break;
					}
				}
				if( pla_types[j].key == 0 ) {
					System.err.printf(
							"espresso: bad keyword \"%s\" following -out\n",argv[i+1]);
					System.exit( 1 );
				}
				break;
			}
		}

		for( int i = 1; i < (*argc); i++ ) {
			if( argv[i][0] == '-' ) {
				for( int j = 0; esp_opt_table[j].name != 0; j++ ) {
					if( strcmp(argv[i]+1, esp_opt_table[j].name) == 0 ) {
						delete_arg(argc, argv, i);
						*(esp_opt_table[j].variable) = esp_opt_table[j].value;
						break;
					}
				}
			}
		}

		if( check_arg(argc, argv, "-fdr") ) {
			input_type = FDR_type;
		}
		if( check_arg(argc, argv, "-fr") ) {
			input_type = FR_type;
		}
		if( check_arg(argc, argv, "-f") ) {
			input_type = F_type;
		}
	} */

	/** delete_arg -- delete an argument from the argument list
	 * java: a new value of argc is returned
	 */
	/* private static int delete_arg(int argc, final String[] argv, final int num) {
		argc--;
		for( int i = num; i < argc; i++ ) {
			argv[ i ] = argv[i + 1];
		}
		return argc;
	} */


	/** check_arg -- scan argv for an argument, and return true if found
	 * java: -1: instead of false, a new value of argc
	 */
	/* private static int check_arg(int argc, final String argv[], final String s) {
		for( int i = 1; i < argc; i++ ) {
			if( argv[ i ].compareTo( s ) == 0 ) {
				argc = delete_arg( argc, argv, i );
				return argc;
			}
		}
		return -1;
	} */
}
