package espresso;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;

final class Jmatrix {
/*
 *  free-lists are only used if 'FAST_AND_LOOSE' is set; this is because
 *  we lose the debugging capability of libmm_t which trashes objects when
 *  they are free'd.  However, FAST_AND_LOOSE is much faster if matrices
 *  are created and freed frequently.
 */

// if( Jmain.FAST_AND_LOOSE ) {
	static Jsm_element sm_element_freelist;
	static Jsm_row sm_row_freelist;
	static Jsm_col sm_col_freelist;
// }

	// sparse_int.h
	static Jsm_element sm_element_alloc() {
		final Jsm_element newobj;
if( Jespresso.FAST_AND_LOOSE ) {
		if( sm_element_freelist == null ) {
			newobj = new Jsm_element();
		} else {
			newobj = sm_element_freelist;
			sm_element_freelist = sm_element_freelist.next_col;
		}
} else {
		newobj = new Jsm_element();
}
		newobj.user_word = null;
		return newobj;
	}

	static void sm_element_free(final Jsm_element e) {
if( Jespresso.FAST_AND_LOOSE ) {
	    e.next_col = sm_element_freelist;
	    sm_element_freelist = e;
} else {
		// e = null;
}
	}
	// end sparse_int.h

	static Jsm_matrix sm_alloc() {
		final Jsm_matrix A = new Jsm_matrix();
		A.rows = null;
		A.cols = null;
		A.nrows = A.ncols = 0;
		A.rows_size = A.cols_size = 0;
		A.first_row = A.last_row = null;
		A.first_col = A.last_col = null;
		A.user_word = null;		/* for our user ... */
		return A;
	}


	static Jsm_matrix sm_alloc_size(final int row, final int col) {
		final Jsm_matrix A = sm_alloc();
		sm_resize( A, row, col );
		return A;
	}

	static void sm_free(final Jsm_matrix A) {
if( Jespresso.FAST_AND_LOOSE ) {
		if( A.first_row != null ) {
			for( Jsm_row prow = A.first_row; prow != null; prow = prow.next_row ) {
				/* add the elements to the free list of elements */
				prow.last_col.next_col = sm_element_freelist;
				sm_element_freelist = prow.first_col;
			}

			/* Add the linked list of rows to the row-free-list */
			A.last_row.next_row = sm_row_freelist;
			sm_row_freelist = A.first_row;

			/* Add the linked list of cols to the col-free-list */
			A.last_col.next_col = sm_col_freelist;
			sm_col_freelist = A.first_col;
		}
} else {
		for( Jsm_row pnext_row, prow = A.first_row; prow != null; prow = pnext_row ) {
			pnext_row = prow.next_row;
			Jrows.sm_row_free( prow );
		}
		for( Jsm_col pnext_col, pcol = A.first_col; pcol != null; pcol = pnext_col ) {
			pnext_col = pcol.next_col;
			pcol.first_row = pcol.last_row = null;
			Jcols.sm_col_free( pcol );
		}
}

		/* Free the arrays to map row/col numbers into pointers */
		A.rows = null;
		A.cols = null;
		// A = null;
	}

	static Jsm_matrix sm_dup(final Jsm_matrix A) {
		final Jsm_matrix B = sm_alloc();
		if( A.last_row != null ) {
			sm_resize( B, A.last_row.row_num, A.last_col.col_num );
			for( Jsm_row prow = A.first_row; prow != null; prow = prow.next_row ) {
				for( Jsm_element p = prow.first_col; p != null; p = p.next_col ) {
					sm_insert( B, p.row_num, p.col_num );
				}
			}
		}
		return B;
	}

	static void sm_resize(final Jsm_matrix A, final int row, final int col) {
		if( row >= A.rows_size ) {
			final int new_size = Math.max( A.rows_size << 1, row + 1 );
			final Jsm_row[] r = new Jsm_row[ new_size ];
			if( A.rows != null ) {
				System.arraycopy( A.rows, 0, r, 0, A.rows.length );
			}
			A.rows = r;// java: extras are already nulled
			//for( int i = A.rows_size; i < new_size; i++ ) {
			//	A.rows[i] = null;
			//}
			A.rows_size = new_size;
		}

		if( col >= A.cols_size ) {
			final int new_size = Math.max( A.cols_size << 1, col + 1 );
			final Jsm_col[] c = new Jsm_col[ new_size ];
			if( A.cols != null ) {
				System.arraycopy( A.cols, 0, c, 0, A.cols.length );
			}
			A.cols = c;// java: extras are already nulled
			//for( int i = A.cols_size; i < new_size; i++ ) {
			//	A.cols[i] = null;
			//}
			A.cols_size = new_size;
		}
	}

	/** insert -- insert a value into the matrix */
	static Jsm_element sm_insert(final Jsm_matrix A, final int row, final int col) {
		Jsm_col pcol;
		Jsm_element element;
		Jsm_element save_element;

		if( row >= A.rows_size || col >= A.cols_size ) {
			sm_resize( A, row, col );
		}

		Jsm_row prow = A.rows[row];
		if( prow == null ) {
			prow = A.rows[ row ] = Jrows.sm_row_alloc();
			prow.row_num = row;
			if( A.last_row == null ) {
				prow.row_num = row;
				A.first_row = prow;
				A.last_row = prow;
				prow.next_row = null;
				prow.prev_row = null;
				A.nrows++;
			} else if( A.last_row.row_num < row ) {
				prow.row_num = row;
				A.last_row.next_row = prow;
				prow.prev_row = A.last_row;
				A.last_row = prow;
				prow.next_row = null;
				A.nrows++;
		    } else if( A.first_row.row_num > row ) {
				prow.row_num = row;
				A.first_row.prev_row = prow;
				prow.next_row = A.first_row;
				A.first_row = prow;
				prow.prev_row = null;
				A.nrows++;
		    } else {
		    	Jsm_row p;
		    	for( p = A.first_row; p.row_num < row; p = p.next_row ) {
					;
				}
				if( p.row_num > row ) {
				    prow.row_num = row;
				    p = p.prev_row;
				    p.next_row.prev_row = prow;
				    prow.next_row = p.next_row;
				    p.next_row = prow;
				    prow.prev_row = p;
				    A.nrows++;
				} else {
				    prow = p;
				}
		    }
		}

		pcol = A.cols[col];
		if( pcol == null ) {
			pcol = A.cols[col] = Jcols.sm_col_alloc();
			pcol.col_num = col;
			if( A.last_col == null ) {
				pcol.col_num = col;
				A.first_col = pcol;
				A.last_col = pcol;
				pcol.next_col = null;
				pcol.prev_col = null;
				A.ncols++;
			} else if( A.last_col.col_num < col ) {
				pcol.col_num = col;
				A.last_col.next_col = pcol;
				pcol.prev_col = A.last_col;
				A.last_col = pcol;
				pcol.next_col = null;
				A.ncols++;
		    } else if( A.first_col.col_num > col ) {
				pcol.col_num = col;
				A.first_col.prev_col = pcol;
				pcol.next_col = A.first_col;
				A.first_col = pcol;
				pcol.prev_col = null;
				A.ncols++;
		    } else {
		    	Jsm_col p;
		    	for( p = A.first_col; p.col_num < col; p = p.next_col ) {
					;
				}
				if( p.col_num > col ) {
				    pcol.col_num = col;
				    p = p.prev_col;
				    p.next_col.prev_col = pcol;
				    pcol.next_col = p.next_col;
				    p.next_col = pcol;
				    pcol.prev_col = p;
				    A.ncols++;
				} else {
				    pcol = p;
				}
		    }
		}

		/* get a new item, save its address */
		element = Jmatrix.sm_element_alloc();
		save_element = element;

		/* insert it into the row list */
		if( prow.last_col == null ) {
			element.col_num = col;
			prow.first_col = element;
			prow.last_col = element;
			element.next_col = null;
			element.prev_col = null;
			prow.length++;
		} else if( prow.last_col.col_num < col ) {
			element.col_num = col;
			prow.last_col.next_col = element;
			element.prev_col = prow.last_col;
			prow.last_col = element;
			element.next_col = null;
			prow.length++;
	    } else if( prow.first_col.col_num > col ) {
			element.col_num = col;
			prow.first_col.prev_col = element;
			element.next_col = prow.first_col;
			prow.first_col = element;
			element.prev_col = null;
			prow.length++;
	    } else {
	    	Jsm_element p;
	    	for( p = prow.first_col; p.col_num < col; p = p.next_col ) {
				;
			}
			if( p.col_num > col ) {
			    element.col_num = col;
			    p = p.prev_col;
			    p.next_col.prev_col = element;
			    element.next_col = p.next_col;
			    p.next_col = element;
			    element.prev_col = p;
			    prow.length++;
			} else {
			    element = p;
			}
	    }

		/* if it was used, also insert it into the column list */
		if( element == save_element ) {
			if( pcol.last_row == null ) {
				element.row_num = row;
				pcol.first_row = element;
				pcol.last_row = element;
				element.next_row = null;
				element.prev_row = null;
				pcol.length++;
			} else if( pcol.last_row.row_num < row ) {
				element.row_num = row;
				pcol.last_row.next_row = element;
				element.prev_row = pcol.last_row;
				pcol.last_row = element;
				element.next_row = null;
				pcol.length++;
		    } else if( pcol.first_row.row_num > row ) {
				element.row_num = row;
				pcol.first_row.prev_row = element;
				element.next_row = pcol.first_row;
				pcol.first_row = element;
				element.prev_row = null;
				pcol.length++;
		    } else {
		    	Jsm_element p;
		    	for( p = pcol.first_row; p.row_num < row; p = p.next_row ) {
					;
				}
				if( p.row_num > row ) {
				    element.row_num = row;
				    p = p.prev_row;
				    p.next_row.prev_row = element;
				    element.next_row = p.next_row;
				    p.next_row = element;
				    element.prev_row = p;
				    pcol.length++;
				} else {
				    element = p;
				}
		    }
		} else {
			/* otherwise, it was already in matrix -- free element we allocated */
			Jmatrix.sm_element_free( save_element );
		}
		return element;
	}


	static Jsm_element sm_find(final Jsm_matrix A, final int rownum, final int colnum) {
		final Jsm_row prow = Jsparse.sm_get_row( A, rownum );
		if( prow == null ) {
			return null;
		}// else {
			final Jsm_col pcol = Jsparse.sm_get_col( A, colnum );
			if( pcol == null ) {
				return null;
			}
			if( prow.length < pcol.length ) {
				return Jrows.sm_row_find( prow, colnum );
			}// else {
				return Jcols.sm_col_find( pcol, rownum );
			//}
		//}
	}


	static void sm_remove(final Jsm_matrix A, final int rownum, final int colnum) {
		sm_remove_element( A, sm_find( A, rownum, colnum ) );
	}

	static void sm_remove_element(final Jsm_matrix A, final Jsm_element p) {
		if( p == null ) {
			return;
		}

		/* Unlink the element from its row */
		final Jsm_row prow = Jsparse.sm_get_row( A, p.row_num );
		if( p.prev_col == null ) {
			prow.first_col = p.next_col;
		} else {
			p.prev_col.next_col = p.next_col;
		}
		if( p.next_col == null ) {
			prow.last_col = p.prev_col;
		} else {
			p.next_col.prev_col = p.prev_col;
		}
		prow.length--;

		/* if no more elements in the row, discard the row header */
		if( prow.first_col == null ) {
			sm_delrow( A, p.row_num );
		}

		/* Unlink the element from its column */
		final Jsm_col pcol = Jsparse.sm_get_col( A, p.col_num );
		if( p.prev_row == null ) {
			pcol.first_row = p.next_row;
		} else {
			p.prev_row.next_row = p.next_row;
		}
		if( p.next_row == null ) {
			pcol.last_row = p.prev_row;
		} else {
			p.next_row.prev_row = p.prev_row;
		}
		pcol.length--;

		/* if no more elements in the column, discard the column header */
		if( pcol.first_row == null ) {
			sm_delcol( A, p.col_num );
		}

		Jmatrix.sm_element_free( p );
	}

	static void sm_delrow(final Jsm_matrix A, final int i) {
		Jsm_element p;
		Jsm_element pnext;
		Jsm_col pcol;

		final Jsm_row prow = Jsparse.sm_get_row( A, i );
		if( prow != null ) {
			/* walk across the row */
			for( p = prow.first_col; p != null; p = pnext ) {
				pnext = p.next_col;

				/* unlink the item from the column (and delete it) */
				pcol = Jsparse.sm_get_col( A, p.col_num );
				Jcols.sm_col_remove_element( pcol, p );

				/* discard the column if it is now empty */
				if( pcol.first_row == null ) {
					sm_delcol( A, pcol.col_num );
				}
			}

			/* discard the row -- we already threw away the elements */
			A.rows[ i ] = null;
			if( prow.prev_row == null ) {
				A.first_row = prow.next_row;
			} else {
				prow.prev_row.next_row = prow.next_row;
			}
			if( prow.next_row == null ) {
				A.last_row = prow.prev_row;
			} else {
				prow.next_row.prev_row = prow.prev_row;
			}
			A.nrows--;
			prow.first_col = prow.last_col = null;
			Jrows.sm_row_free( prow );
		}
	}

	static void sm_delcol(final Jsm_matrix A, final int i) {
		Jsm_element p, pnext;

		final Jsm_col pcol = Jsparse.sm_get_col( A, i );
		if( pcol != null ) {
			/* walk down the column */
			for( p = pcol.first_row; p != null; p = pnext ) {
				pnext = p.next_row;

				/* unlink the element from the row (and delete it) */
				final Jsm_row prow = Jsparse.sm_get_row( A, p.row_num );
				Jrows.sm_row_remove_element( prow, p );

				/* discard the row if it is now empty */
				if( prow.first_col == null ) {
					sm_delrow( A, prow.row_num );
				}
			}

			/* discard the column -- we already threw away the elements */
			A.cols[i] = null;
			if( pcol.prev_col == null ) {
				A.first_col = pcol.next_col;
			} else {
				pcol.prev_col.next_col = pcol.next_col;
			}
			if( pcol.next_col == null ) {
				A.last_col = pcol.prev_col;
			} else {
				pcol.next_col.prev_col = pcol.prev_col;
			}
			A.ncols--;
			pcol.first_row = pcol.last_row = null;
			Jcols.sm_col_free( pcol );
		}
	}

	static void sm_copy_row(final Jsm_matrix dest, final int dest_row, final Jsm_row prow) {
		for( Jsm_element p = prow.first_col; p != null; p = p.next_col ) {
			sm_insert( dest, dest_row, p.col_num );
		}
	}

	static void sm_copy_col(final Jsm_matrix dest, final int dest_col, final Jsm_col pcol) {
		for( Jsm_element p = pcol.first_row; p != null; p = p.next_row ) {
			sm_insert( dest, dest_col, p.row_num );
		}
	}

	static Jsm_row sm_longest_row(final Jsm_matrix A) {
		int max_length = 0;
		Jsm_row large_row = null;
		for( Jsm_row prow = A.first_row; prow != null; prow = prow.next_row ) {
			if( prow.length > max_length ) {
				max_length = prow.length;
				large_row = prow;
			}
		}
		return large_row;
	}

	static Jsm_col sm_longest_col(final Jsm_matrix A) {
		int max_length = 0;
		Jsm_col large_col = null;
		for( Jsm_col pcol = A.first_col; pcol != null; pcol = pcol.next_col ) {
			if( pcol.length > max_length ) {
				max_length = pcol.length;
				large_col = pcol;
			}
		}
		return large_col;
	}

	static int sm_num_elements(final Jsm_matrix A) {
		int num = 0;
		// sm_foreach_row( A, prow ) {
		for( Jsm_row prow = A.first_row; prow != null; prow = prow.next_row ) {
			num += prow.length;
		}
		return num;
	}

	static boolean sm_read(final InputStream fp, final Jsm_matrix[] A) {
		A[ 0 ] = sm_alloc();
		while( true ) {// java ! EOF
			// err = fscanf(fp, "%d %d", &i, &j);
			try {
				final int i = Jsupport.scanInt( fp, 10 );
				try {
					final int j = Jsupport.scanInt( fp, 10 );
					sm_insert( A[ 0 ], i, j );
				} catch(final IOException ie) {
					return false;// not two
				}
			} catch(final IOException ie) {
				return true;// EOF
			}
		}
	}

	static boolean sm_read_compressed(final InputStream fp, final Jsm_matrix[] A) {
		A[ 0 ] = sm_alloc();
		try {// fscanf(fp, "%d %d", &nrows, &ncols)
			final int nrows = Jsupport.scanInt( fp, 10 );
			final int ncols = Jsupport.scanInt( fp, 10 );
			sm_resize( A[ 0 ], nrows, ncols );
			for( int i = 0; i < nrows; i++ ) {
				int x = Jsupport.scanInt( fp, 16 );// fscanf(fp, "%lx", &x)
				for( int j = 0; j < ncols; j += 32 ) {
					x = Jsupport.scanInt( fp, 16 );// fscanf(fp, "%lx", &x)
					for( int k = j; x != 0; x >>>= 1, k++ ) {
						if( (x & 1) != 0 ) {
							sm_insert( A[ 0 ], i, k );
						}
					}
				}
			}
		} catch(final IOException ie) {
			return false;
		}
		return true;
	}

	@SuppressWarnings("boxing")
	static void sm_write(final PrintStream fp, final Jsm_matrix A) {
		for( Jsm_row prow = A.first_row; prow != null; prow = prow.next_row ) {
			for( Jsm_element p = prow.first_col; p != null; p = p.next_col ) {
				fp.printf("%d %d\n", p.row_num, p.col_num );
			}
		}
	}

	@SuppressWarnings("boxing")
	static void sm_print(final PrintStream fp, final Jsm_matrix A) {
		if( A.last_col.col_num >= 100 ) {
			fp.printf("    ");
			for( Jsm_col pcol = A.first_col; pcol != null; pcol = pcol.next_col ) {
				fp.printf("%d", (pcol.col_num / 100) % 10 );
			}
			fp.print('\n');
		}

		if( A.last_col.col_num >= 10 ) {
			fp.printf("    ");
			for( Jsm_col pcol = A.first_col; pcol != null; pcol = pcol.next_col ) {
				fp.printf("%d", (pcol.col_num / 10) % 10 );
			}
			fp.print('\n');
		}

		fp.printf("    ");
		for( Jsm_col pcol = A.first_col; pcol != null; pcol = pcol.next_col ) {
			fp.printf("%d", pcol.col_num % 10 );
		}
		fp.print('\n');

		fp.printf("    ");
		for( Jsm_col pcol = A.first_col; pcol != null; pcol = pcol.next_col ) {
			fp.printf("-");
		}
		fp.print('\n');

		for( Jsm_row prow = A.first_row; prow != null; prow = prow.next_row ) {
			fp.printf("%3d:", prow.row_num );

			for( Jsm_col pcol = A.first_col; pcol != null; pcol = pcol.next_col ) {
				final int c = Jrows.sm_row_find( prow, pcol.col_num ) != null ? '1' : '.';
				fp.print( c );
			}
			fp.print('\n');
		}
	}

	@SuppressWarnings("boxing")
	static void sm_dump(final Jsm_matrix A, final String s, final int max) {
		final PrintStream fp = System.out;

		fp.printf( "%s %d rows by %d cols\n", s, A.nrows, A.ncols );
		if( A.nrows < max ) {
			sm_print( fp, A );
		}
	}

	static void sm_cleanup() {
if( Jespresso.FAST_AND_LOOSE ) {
		//for( Jsm_element pnext, p = sm_element_freelist; p != null; p = pnext ) {
		//	pnext = p.next_col;
		//	p = null;
		//}
		sm_element_freelist = null;

		//for( Jsm_row pnextrow, prow = sm_row_freelist; prow != null; prow = pnextrow ) {
		//	pnextrow = prow.next_row;
		//	prow = null;
		//}
		sm_row_freelist = null;

		//for( Jsm_col pnextcol, pcol = sm_col_freelist; pcol != null; pcol = pnextcol ) {
		//	pnextcol = pcol.next_col;
		//	pcol = null;
		//}
		sm_col_freelist = null;
}
	}
}