package espresso;

/** pair_t describes bit-paired variables. pair_struct, pair_t, *ppair */
final class Jpair {
	private static final int FIND_BEST_COST = 1;
	private static final int MINIMIZE_PAIR = 2;

	int cnt;
	int[] var1;
	int[] var2;

	static void set_pair(final JpPLA PLA) {
		set_pair1( PLA, true );
	}

	static void set_pair1(final JpPLA PLA, final boolean adjust_labels) {
		final Jcube_struct cube = Jglobals.cube;// java
		int newvar;
		int  new_mv_start;
		final Jpair pair = PLA.pair;

		if( adjust_labels ) {
			Jcvrout.makeup_labels( PLA );
		}

		/* Check the pair structure for valid entries and see which binary
		variables are left unpaired
		*/
		boolean[] paired = new boolean[ cube.num_binary_vars ];// java: all are already false
		//for( int var = 0; var < cube.num_binary_vars; var++ ) {
		//	paired[ var ] = false;
		//}
		for( int i = 0; i < pair.cnt; i++ ) {
			if( (pair.var1[ i ] > 0 && pair.var1[ i ] <= cube.num_binary_vars) &&
					(pair.var2[ i ] > 0 && pair.var2[ i ] <= cube.num_binary_vars) ) {
				paired[ pair.var1[ i ] - 1 ] = true;
				paired[ pair.var2[ i ] - 1 ] = true;
			} else {
				Jcvrmisc.fatal("can only pair binary-valued variables");
			}
		}

		PLA.F = delvar( pairvar( PLA.F, pair ), paired );
		PLA.R = delvar( pairvar( PLA.R, pair ), paired );
		PLA.D = delvar( pairvar( PLA.D, pair ), paired );

		/* Now painfully adjust the cube size */
		final int old_size = cube.size;
		final int old_num_vars = cube.num_vars;
		final int old_num_binary_vars = cube.num_binary_vars;
		final int old_mv_start = cube.first_part[ cube.num_binary_vars ];
		/* Create the new cube.part_size vector and setup the cube structure */
		int new_num_binary_vars = 0;
		for( int var = 0; var < old_num_binary_vars; var++ ) {
			if( ! paired[ var ] ) {
				new_num_binary_vars++;
			}
		}
		int new_num_vars = new_num_binary_vars + pair.cnt;
		new_num_vars += old_num_vars - old_num_binary_vars;
		final int[] new_part_size = new int[ new_num_vars ];
		for( int var = 0; var < pair.cnt; var++ ) {
			new_part_size[ new_num_binary_vars + var ] = 4;
		}
		for( int var = 0; var < old_num_vars - old_num_binary_vars; var++ ) {
			new_part_size[ new_num_binary_vars + pair.cnt + var ] = cube.part_size[ old_num_binary_vars + var ];
		}
		Jcubestr.setdown_cube();
		cube.part_size = null;
		cube.num_vars = new_num_vars;
		cube.num_binary_vars = new_num_binary_vars;
		cube.part_size = new_part_size;
		Jcubestr.cube_setup();

		/* hack with the labels to get them correct */
		if( adjust_labels ) {
			String[] oldlabel = PLA.label;
			PLA.label = new String[ cube.size ];
			for( int var = 0; var < pair.cnt; var++ ) {
				newvar = (cube.num_binary_vars << 1) + (var << 2);
				int pv = (pair.var1[var] - 1) << 1;
				final String var1bar = oldlabel[ pv ];
				final String var1 = oldlabel[ pv + 1 ];
				pv = (pair.var2[var] - 1) << 1;
				final String var2bar = oldlabel[ pv ];
				final String var2 = oldlabel[ pv + 1 ];
				PLA.label[ newvar ] = String.format( "%s+%s", var1bar, var2bar );
				PLA.label[ newvar + 1 ] = String.format( "%s+%s", var1bar, var2 );
				PLA.label[ newvar + 2 ] = String.format( "%s+%s", var1, var2bar );
				PLA.label[ newvar + 3 ] = String.format( "%s+%s", var1, var2 );
			}
			/* Copy the old labels for the unpaired binary vars */
			int i = 0;
			for( int var = 0; var < old_num_binary_vars; var++ ) {
				if( paired[ var ] == false ) {
					final int v2 = var << 1;
					PLA.label[ i++ ] = oldlabel[ v2 ];
					PLA.label[ i++ ] = oldlabel[ v2 + 1 ];
					oldlabel[ v2 ] = oldlabel[ v2 + 1 ] = null;
				}
			}
			/* Copy the old labels for the remaining unpaired vars */
			new_mv_start = cube.num_binary_vars * 2 + pair.cnt * 4;
			for( i = old_mv_start; i < old_size; i++ ) {
				PLA.label[ new_mv_start + i - old_mv_start ] = oldlabel[ i ];
				oldlabel[ i ] = null;
			}
			/* free remaining entries in oldlabel */
			oldlabel = null;
		}

		/* the paired variables should not be sparse (cf. mv_reduce/raise_in)*/
		for( int var = 0; var < pair.cnt; var++ ) {
			cube.sparse[ cube.num_binary_vars + var ] = false;
		}
		paired = null;
	}

	static Jset_family pairvar(Jset_family A, final Jpair pair) {
		final Jcube_struct cube = Jglobals.cube;// java

		final int insert_col = cube.first_part[ cube.num_vars - 1 ];

		/* stretch the cover matrix to make room for the paired variables */
		A = Jset.sf_delcol( A, insert_col, -(pair.cnt << 2) );

		/* compute the paired values */
		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			for( int pairnum = 0; pairnum < pair.cnt; pairnum++ ) {
				final int p1 = cube.first_part[ pair.var1[ pairnum ] - 1 ];
				final int p2 = cube.first_part[ pair.var2[ pairnum ] - 1 ];
				final boolean b1 = Jespresso.is_in_set( data, p, p2 + 1 );
				final boolean b0 = Jespresso.is_in_set( data, p, p2 );
				final int val = insert_col + (pairnum << 2);
				if( /* a0 */ Jespresso.is_in_set( data, p, p1 ) ) {
					if( b0 ) {
						Jespresso.set_insert( data, p, val + 3 );
					}
					if( b1 ) {
						Jespresso.set_insert( data, p, val + 2 );
					}
				}
				if( /* a1 */ Jespresso.is_in_set( data, p, p1 + 1 ) ) {
					if( b0 ) {
						Jespresso.set_insert( data, p, val + 1 );
					}
					if( b1 ) {
						Jespresso.set_insert( data, p, val );
					}
				}
			}
		}
		return A;
	}


	/** delvar -- delete variables from A, minimize the number of column shifts */
	static Jset_family delvar(Jset_family A, final boolean[] paired) {
		final Jcube_struct cube = Jglobals.cube;// java
		int first_run = -1;// java: fix var may have been not initialized
		int offset = 0;

		boolean run = false;
		int run_length = 0;
		for( int var = 0; var < cube.num_binary_vars; var++ ) {
			if( paired[ var ] ) {
				if( run ) {
					run_length += cube.part_size[ var ];
				} else {
					run = true;
					first_run = cube.first_part[ var ];
					run_length = cube.part_size[ var ];
				}
			} else
				if( run ) {
					A = Jset.sf_delcol( A, first_run - offset, run_length );
					run = false;
					offset += run_length;
				}
		}
		if( run ) {
			A = Jset.sf_delcol( A, first_run - offset, run_length );
		}
		return A;
	}

	/**
	find_optimal_pairing -- find which binary variables should be paired
	to maximally reduce the number of terms

	This is essentially the technique outlined by T. Sasao in the
	Trans. on Comp., Oct 1984.  We estimate the cost of pairing each
	pair individually using 1 of 4 strategies: (1) algebraic division
	of F by the pair (exactly T. Sasao technique); (2) strong division
	of F by the paired variables (using REDUCE/EXPAND/ IRREDUNDANT from
	espresso); (3) full minimization using espresso; (4) exact
	minimization.  These are in order of both increasing accuracy and
	increasing difficulty (!)

	Once the n squared pairs have been evaluated, T. Sasao proposes a
	graph covering of nodes by disjoint edges.  For now, I solve this
	problem exhaustively (complexity = (n-1)*(n-3)*...*3*1 for n
	variables when n is even).  Note that solving this problem exactly
	is the same as evaluating the cost function for all possible
	pairings.

	n       pairs

	1, 2           1
	3, 4           3
	5, 6          15
	7, 8         105
	9,10         945
	11,12      10,395
	13,14     135,135
	15,16   2,027,025
	17,18  34,459,425
	19,20 654,729,075
	*/
	@SuppressWarnings("boxing")
	static void find_optimal_pairing(final JpPLA PLA, final int strategy) {
		final Jcube_struct cube = Jglobals.cube;// java

		int[][] cost_arr = find_pairing_cost( PLA, strategy );

		if( Jglobals.summary ) {
			System.out.printf("    ");
			for( int i = 0; i < cube.num_binary_vars; i++ ) {
				System.out.printf("%3d ", i + 1 );
			}
			System.out.printf("\n");
			for( int i = 0; i < cube.num_binary_vars; i++ ) {
				System.out.printf("%3d ", i + 1 );
				for( int j = 0; j < cube.num_binary_vars; j++ ) {
					System.out.printf("%3d ", cost_arr[ i ][ j ] );
				}
				System.out.printf("\n");
			}
		}

		if( cube.num_binary_vars <= 14 ) {
			PLA.pair = pair_best_cost( cost_arr );
		} else {
			final Jpair temp[] = new Jpair[] { PLA.pair };// java helper
			greedy_best_cost( cost_arr, temp );
			PLA.pair = temp[0];// java
		}
		System.out.printf("# ");
		print_pair( PLA.pair );

		for( int i = 0; i < cube.num_binary_vars; i++) {
			cost_arr[i] = null;
		}
		cost_arr = null;

		set_pair( PLA );
		{
			final long t = System.currentTimeMillis();
			PLA.F = Jespresso.espresso( PLA.F, PLA.D, PLA.R );
			if( Jglobals.summary ) {
				Jcvrmisc.print_trace( PLA.F, "ESPRESSO  ", System.currentTimeMillis() - t );
			}
		}
	}

	static int[][] find_pairing_cost(final JpPLA PLA, final int strategy) {
		final Jcube_struct cube = Jglobals.cube;// java

		/* data is returned in the cost array */
		final int[][] cost_arr = new int[ cube.num_binary_vars ][];
		for( int i = 0; i < cube.num_binary_vars; i++ ) {
			cost_arr[ i ] = new int[ cube.num_binary_vars ];
		}
		/* for( int i = 0; i < cube.num_binary_vars; i++ ) {// java: already zeroed
			for( int j = 0; j < cube.num_binary_vars; j++ ) {
				cost_arr[ i ][ j ] = 0;
			}
		}*/

		/* Setup the pair structure for pairing variables together */
		PLA.pair = pair_new( 1 );
		PLA.pair.cnt = 1;

		for( int var1 = 0; var1 < cube.num_binary_vars - 1; var1++ ) {
			for( int var2 = var1 + 1; var2 < cube.num_binary_vars; var2++ ) {
				int[] xpart_size;
				/* if anything but simple strategy, perform setup */
				if( strategy > 0 ) {
					/* save the original covers */
					final Jset_family Fsave = Jset.sf_save( PLA.F );
					final Jset_family Dsave = Jset.sf_save( PLA.D );
					final Jset_family Rsave = Jset.sf_save( PLA.R );

					/* save the original cube structure */
					final int xnum_binary_vars = cube.num_binary_vars;
					final int xnum_vars = cube.num_vars;
					xpart_size = new int[ cube.num_vars ];
					for( int i = 0; i < cube.num_vars; i++ ) {
						xpart_size[i] = cube.part_size[i];
					}

					/* pair two variables together */
					PLA.pair.var1[0] = var1 + 1;
					PLA.pair.var2[0] = var2 + 1;
					set_pair1( PLA, /* adjust_labels */ false );

					int cost = 0;// java to suppress warning may be have been not initialized
					/* decide how to best estimate worth of this pairing */
					switch( strategy ) {// java: moved into if
					case 3:
						/*s = "exact minimization";*/
						PLA.F = Jexact.minimize_exact( PLA.F, PLA.D, PLA.R, true );
						cost = Fsave.count - PLA.F.count;
						break;
					case 2:
						/*s = "full minimization";*/
						PLA.F = Jespresso.espresso( PLA.F, PLA.D, PLA.R );
						cost = Fsave.count - PLA.F.count;
						break;
					case 1:
						/*s = "strong division";*/
						PLA.F = Jreduce.reduce( PLA.F, PLA.D );
						PLA.F = Jexpand.expand( PLA.F, PLA.R, false );
						PLA.F = Jirred.irredundant( PLA.F, PLA.D );
						cost = Fsave.count - PLA.F.count;
						break;
					/* case 0:
						// s = "weak division";
						int[] mask = Jespresso.new_cube();
						Jset.set_or( mask, cube.var_mask[var1], cube.var_mask[var2] );
						final Jset_family T = Jcontain.dist_merge( Jset.sf_save( PLA.F ), mask );
						cost = PLA.F.count - T.count;
						Jset.sf_free( T );
						mask = null; */
					}
					cost_arr[ var1 ][ var2 ] = cost;

					//if( strategy > 0 ) {
						/* restore the original cube structure -- free the new ones */
						Jcubestr.setdown_cube();
						cube.part_size = null;
						cube.num_binary_vars = xnum_binary_vars;
						cube.num_vars = xnum_vars;
						cube.part_size = xpart_size;
						Jcubestr.cube_setup();

						/* restore the original cover(s) -- free the new ones */
						PLA.F = Fsave;
						PLA.D = Dsave;
						PLA.R = Rsave;
					//}
				} else if( strategy == 0 ) {
					/*s = "weak division";*/
					int[] mask = Jespresso.new_cube();
					Jset.set_or( mask, cube.var_mask[var1], cube.var_mask[var2] );
					final Jset_family T = Jcontain.dist_merge( Jset.sf_save( PLA.F ), mask );
					final int cost = PLA.F.count - T.count;
					Jset.sf_free( T );
					mask = null;

					cost_arr[ var1 ][ var2 ] = cost;
				}
			}
		}

		pair_free( PLA.pair );
		PLA.pair = null;
		return cost_arr;
	}

	private static int best_cost;
	private static int[][] cost_array;
	private static Jpair best_pair;
	private static int[] best_phase;
	private static JpPLA global_PLA;
	private static Jset_family best_F, best_D, best_R;
	private static int pair_minim_strategy;

	@SuppressWarnings("boxing")
	static void print_pair(final Jpair pair) {
		System.out.printf("pair is");
		for( int i = 0; i < pair.cnt; i++ ) {
			System.out.printf(" (%d %d)", pair.var1[ i ], pair.var2[ i ] );
		}
		System.out.printf("\n");
	}


	static int greedy_best_cost(final int[][] cost_array_local, final Jpair[] pair_p) {
		final Jcube_struct cube = Jglobals.cube;// java
		int besti = -1, bestj = -1;// java: to suppress warning the var may have been not initialized

		final Jpair pair = pair_new( cube.num_binary_vars );
		int[] cand = Jset.set_full( cube.num_binary_vars );
		int total_cost = 0;

		while( Jset.set_ord( cand ) >= 2 ) {
			int maxcost = -1;
			for( int i = 0; i < cube.num_binary_vars; i++ ) {
				if( Jespresso.is_in_set( cand, i ) ) {
					for( int j = i + 1; j < cube.num_binary_vars; j++ ) {
						if( Jespresso.is_in_set( cand, j ) ) {
							final int m = cost_array_local[ i ][ j ];
							if( m > maxcost ) {
								maxcost = m;
								besti = i;
								bestj = j;
							}
						}
					}
				}
			}
			pair.var1[ pair.cnt ] = besti + 1;
			pair.var2[ pair.cnt ] = bestj + 1;
			pair.cnt++;
			Jespresso.set_remove( cand, besti );
			Jespresso.set_remove( cand, bestj );
			total_cost += maxcost;
		}
		cand = null;
		pair_p[ 0 ] = pair;
		return total_cost;
	}


	static Jpair pair_best_cost(final int[][] cost_array_local) {
		final Jcube_struct cube = Jglobals.cube;// java

		best_cost = -1;
		best_pair = null;
		cost_array = cost_array_local;

		final Jpair pair = pair_new( cube.num_binary_vars );
		int[] candidate = Jset.set_full( cube.num_binary_vars );
		generate_all_pairs( pair, cube.num_binary_vars, candidate, FIND_BEST_COST );// java find_best_cost );
		pair_free( pair );
		candidate = null;
		return best_pair;
	}

	@SuppressWarnings("boxing")
	static void find_best_cost(final Jpair pair) {
		int cost = 0;
		for( int i = 0; i < pair.cnt; i++) {
			cost += cost_array[ pair.var1[ i ] - 1 ][ pair.var2[ i ] - 1 ];
		}
		if( cost > best_cost ) {
			best_cost = cost;
			best_pair = pair_save( pair, pair.cnt );
		}
		if( (Jglobals.debug & Jespresso.MINCOV) != 0 && Jglobals.trace ) {
			System.out.printf("cost is %d ", cost );
			print_pair( pair );
		}
	}

	/**
	pair_all: brute-force approach to try all possible pairings

	pair_strategy is:
	2) for espresso
	3) for minimize_exact
	4) for phase assignment
	*/
	static void pair_all(final JpPLA PLA, final int pair_strategy) {
		final Jcube_struct cube = Jglobals.cube;// java

		global_PLA = PLA;
		pair_minim_strategy = pair_strategy;
		best_cost = PLA.F.count + 1;
		best_pair = null;
		best_phase = null;
		best_F = best_D = best_R = null;
		final Jpair pair = pair_new( cube.num_binary_vars );
		int[] candidate = Jset.set_new( cube.num_binary_vars );
		Jset.set_fill( candidate, cube.num_binary_vars );

		generate_all_pairs( pair, cube.num_binary_vars, candidate, MINIMIZE_PAIR );// java minimize_pair );

		pair_free( pair );
		candidate = null;

		PLA.pair = best_pair;
		PLA.phase = best_phase;
		/* not really necessary
		if( phase != NULL)
		(void) set_phase(PLA.phase);
		*/
		set_pair( PLA );
		System.out.printf("# ");
		print_pair( PLA.pair );

		PLA.F = best_F;
		PLA.D = best_D;
		PLA.R = best_R;
	}


	/**
	*  minimize_pair -- called as each pair is generated
	*/
	static void minimize_pair(final Jpair pair) {
		final Jcube_struct cube = Jglobals.cube;// java

		/* save the original covers */
		final Jset_family Fsave = Jset.sf_save( global_PLA.F );
		final Jset_family Dsave = Jset.sf_save( global_PLA.D );
		final Jset_family Rsave = Jset.sf_save( global_PLA.R );

		/* save the original cube structure */
		final int xnum_binary_vars = cube.num_binary_vars;
		final int xnum_vars = cube.num_vars;
		final int[] xpart_size = new int[ cube.num_vars ];
		for( int i = 0; i < cube.num_vars; i++ ) {
			xpart_size[i] = cube.part_size[i];
		}

		/* setup the paired variables */
		global_PLA.pair = pair;
		set_pair1( global_PLA, /* adjust_labels */ false );

		/* call the minimizer */
		if( Jglobals.summary ) {
			print_pair( pair );
		}
		switch( pair_minim_strategy ) {
		case 2:
			{
				final long t = System.currentTimeMillis();
				Jopo.phase_assignment( global_PLA, 0 );
				if( Jglobals.summary ) {
					Jcvrmisc.print_trace( global_PLA.F, "OPO       ", System.currentTimeMillis() - t );
				}
			}
			if( Jglobals.summary ) {
				System.out.printf("# phase is %s\n", Jcvrout.pc1( global_PLA.phase, 0 ) );
			}
			break;
		case 1:
			{
				final long t = System.currentTimeMillis();
				global_PLA.F = Jexact.minimize_exact( global_PLA.F, global_PLA.D, global_PLA.R, true );
				if( Jglobals.summary ) {
					Jcvrmisc.print_trace( global_PLA.F, "EXACT     ", System.currentTimeMillis() - t );
				}
			}
			break;
		case 0:
			{
				final long t = System.currentTimeMillis();
				global_PLA.F = Jespresso.espresso( global_PLA.F, global_PLA.D, global_PLA.R );
				if( Jglobals.summary ) {
					Jcvrmisc.print_trace( global_PLA.F, "ESPRESSO  ", System.currentTimeMillis() - t );
				}
			}
			break;
		default:
			break;
		}

		/* see if we have a new best solution */
		if( global_PLA.F.count < best_cost ) {
			best_cost = global_PLA.F.count;
			best_pair = pair_save( pair, pair.cnt );
			best_phase = global_PLA.phase != null ? Jset.set_save( global_PLA.phase ) : null;
			best_F = Jset.sf_save( global_PLA.F );
			best_D = Jset.sf_save( global_PLA.D );
			best_R = Jset.sf_save( global_PLA.R );
		}

		/* restore the original cube structure -- free the new ones */
		Jcubestr.setdown_cube();
		cube.part_size = null;
		cube.num_binary_vars = xnum_binary_vars;
		cube.num_vars = xnum_vars;
		cube.part_size = xpart_size;
		Jcubestr.cube_setup();

		/* restore the original cover(s) -- free the new ones */
		global_PLA.F = Fsave;
		global_PLA.D = Dsave;
		global_PLA.R = Rsave;
		global_PLA.pair = null;
		global_PLA.phase = null;
	}

	static void generate_all_pairs(final Jpair pair, final int n, final int[] candidate, final int action) {

		if( Jset.set_ord( candidate ) < 2 ) {
			// (*action)(pair);
			switch( action ) {
			case FIND_BEST_COST:
				find_best_cost( pair );
				return;
			case MINIMIZE_PAIR:
				minimize_pair( pair );
				return;
			}
			return;
		}

		final Jpair recur_pair = pair_save( pair, n );
		int[] recur_candidate = Jset.set_save( candidate );

		/* Find first variable still in the candidate set */
		int i;
		for( i = 0; i < n; i++ ) {
			if( Jespresso.is_in_set( candidate, i ) ) {
				break;
			}
		}

		/* Try all pairs of i with other variables */
		for( int j = i + 1; j < n; j++ ) {
			if( Jespresso.is_in_set( candidate, j ) ) {
				/* pair (i j) -- remove from candidate set for future pairings */
				Jespresso.set_remove( recur_candidate, i );
				Jespresso.set_remove( recur_candidate, j );

				/* add to the pair array */
				recur_pair.var1[recur_pair.cnt] = i + 1;
				recur_pair.var2[recur_pair.cnt] = j + 1;
				recur_pair.cnt++;

				/* recur looking for the end ... */
				generate_all_pairs( recur_pair, n, recur_candidate, action );

				/* now break this pair, and restore candidate set */
				recur_pair.cnt--;
				Jespresso.set_insert( recur_candidate, i );
				Jespresso.set_insert( recur_candidate, j );
			}
		}

		/* if odd, generate all pairs which do NOT include i */
		if( (Jset.set_ord( candidate ) % 2) == 1 ) {
			Jespresso.set_remove( recur_candidate, i );
			generate_all_pairs( recur_pair, n, recur_candidate, action );
		}

		Jpair.pair_free( recur_pair );
		recur_candidate = null;
	}

	static Jpair pair_new(final int n) {
		final Jpair pair1 = new Jpair();
		pair1.cnt = 0;
		pair1.var1 = new int[ n ];
		pair1.var2 = new int[ n ];
		return pair1;
	}

	static Jpair pair_save(final Jpair pair, final int n) {
		final Jpair pair1 = pair_new( n );
		pair1.cnt = pair.cnt;
		/* java replaced by arraycopy for( int k = 0, ke = pair.cnt; k < ke; k++ ) {
			pair1.var1[ k ] = pair.var1[ k ];
			pair1.var2[ k ] = pair.var2[ k ];
		}*/
		System.arraycopy( pair.var1, 0, pair1.var1, 0, pair.cnt );
		System.arraycopy( pair.var2, 0, pair1.var2, 0, pair.cnt );
		return pair1;
	}

	static void pair_free(Jpair pair) {
		pair.var1 = null;
		pair.var2 = null;
		pair = null;
	}
}