package espresso;
final class Jprimes {

	/** primes_consensus -- generate primes using consensus
	 *
	 * @param T T will be disposed of
	 */
	static Jset_family primes_consensus(final Jset[] T) {
		final Jset_family[] Tnew = new Jset_family[ 1 ];// java helper

		if( primes_consensus_special_cases( T, Tnew ) == Jespresso.MAYBE ) {
			int[] cl = Jespresso.new_cube();
			int[] cr = Jespresso.new_cube();
			final int best = Jcofactor.binate_split_select( T, cl, cr, Jespresso.COMPL );

			final Jset_family Tl = primes_consensus( Jcofactor.scofactor( T, cl, best ) );
			final Jset_family Tr = primes_consensus( Jcofactor.scofactor( T, cr, best ) );
			Tnew[ 0 ] = primes_consensus_merge( Tl, Tr, cl, cr );

			cl = null;
			cr = null;
			T[ 0 ] = null;
		}

		return Tnew[ 0 ];
	}

	/**
	 *
	 * @param T will be disposed if answer is determined
	 * @param Tnew returned only if answer determined
	 * @return 3 state boolean
	 */
	private static int primes_consensus_special_cases(final Jset[] T, final Jset_family[] Tnew) {
		final Jcube_struct cube = Jglobals.cube;// java
		final Jset cof = T[0];

		/* Check for no cubes in the cover */
		if( T[2] == null || T[ 2 ].base == null ) {// TODO java check T[ 2 ].base == null
			Tnew[ 0 ] = Jespresso.new_cover( 0 );
			T[0] = null;
			return Jespresso.TRUE;
		}

		/* Check for only a single cube in the cover */
		if( T[ 3 ] == null || T[ 3 ].base == null ) {// TODO java check T[ 3 ].base == null
			final Jset t = Jset.set_or( cof, cof.base, cof.offset, T[2].base, T[2].offset );
			Tnew[ 0 ] = Jset.sf_addset( Jespresso.new_cover( 1 ), t.base, t.offset );
			T[0] = null;
			return Jespresso.TRUE;
		}

		/* Check for a row of all 1's (implies function is a tautology) */
		Jset p;
		for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
			if( Jsetc.full_row( p, cof ) ) {
				Tnew[ 0 ] = Jset.sf_addset( Jespresso.new_cover( 1 ), cube.fullset );
				T[0] = null;
				return Jespresso.TRUE;
			}
		}

		/* Check for a column of all 0's which can be factored out */
		int[] ceil = Jset.set_save( cof );
		for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
			Jset.set_or( ceil, ceil, p.base, p.offset );
		}
		if( ! Jset.setp_equal( ceil, cube.fullset ) ) {
			int[] t = Jespresso.new_cube();
			Jset.set_diff( t, cube.fullset, ceil );
			Jset.set_or( cof.base, cof.offset, cof.base, cof.offset, t );
			t = null;

			final Jset_family A = primes_consensus( T );

			final int[] data = A.data;// java
		    for( int pi = 0, last = pi + A.count * A.wsize; pi < last; pi += A.wsize ) {
				Jset.set_and( data, pi, data, pi, ceil );
			}
			Tnew[ 0 ] = A;
			ceil = null;
			return Jespresso.TRUE;
		}
		ceil = null;

		/* Collect column counts, determine unate variables, etc. */
		Jcofactor.massive_count( T );

		/* If single active variable not factored out above, then tautology ! */
		if( Jglobals.cdata.vars_active == 1 ) {
			Tnew[ 0 ] = Jset.sf_addset( Jespresso.new_cover( 1 ), cube.fullset );
			T[0] = null;
			return Jespresso.TRUE;

		/* Check for unate cover */
		} else if( Jglobals.cdata.vars_unate == Jglobals.cdata.vars_active ) {
			final Jset_family A = Jcofactor.cubeunlist( T );
			Tnew[ 0 ] = Jcontain.sf_contain( A );
			T[0] = null;
			return Jespresso.TRUE;

		/* Not much we can do about it */
		} else {
			return Jespresso.MAYBE;
		}
	}

	private static Jset_family primes_consensus_merge(Jset_family Tl, Jset_family Tr, final int[] cl, final int[] cr) {
		Tl = and_with_cofactor( Tl, cl );
		Tr = and_with_cofactor( Tr, cr );

		Jset_family T = Jset.sf_new( 500, Tl.sf_size );
		int[] pt_base = T.data;
		int pt = 0;// java pt_base[ pt ]
		Jset_family Tsave = Jcontain.sf_contain( Jset.sf_join( Tl, Tr ) );

		final int[] datal = Tl.data;// java
		for( int pl = 0, lastl = pl + Tl.count * Tl.wsize; pl < lastl; pl += Tl.wsize ) {

			final int[] datar = Tr.data;// java
			for( int pr = 0, lastr = pr + Tr.count * Tr.wsize; pr < lastr; pr += Tr.wsize ) {
				if( Jsetc.cdist01( datal, pl, datar, pr ) == 1 ) {
					Jsetc.consensus( pt_base, pt, datal, pl, datar, pr );
					if( ++T.count >= T.capacity ) {
						Tsave = Jcontain.sf_union( Tsave, Jcontain.sf_contain( T ) );
						T = Jset.sf_new( 500, Tl.sf_size );
						pt_base = T.data;
						pt = 0;
					} else {
						pt += T.wsize;
					}
				}
			}
		}
		Jespresso.free_cover( Tl );
		Jespresso.free_cover( Tr );

		Tsave = Jcontain.sf_union( Tsave, Jcontain.sf_contain( T ) );
		return Tsave;
	}

	private static Jset_family and_with_cofactor(final Jset_family A, final int[] cof) {
		final Jcube_struct cube = Jglobals.cube;// java

		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			Jset.set_and( data, p, data, p, cof );
			if( Jsetc.cdist( data, p, cube.fullset, 0 ) > 0 ) {
				Jespresso.RESET( data, p, Jespresso.ACTIVE );
			} else {
				Jespresso.SET( data, p, Jespresso.ACTIVE );
			}
		}
		return Jset.sf_inactive( A );
	}
}