package espresso;

import java.io.PrintStream;

final class Jrows {

	/** allocate a new row vector */
	static Jsm_row sm_row_alloc() {
		Jsm_row prow;

if( Jespresso.FAST_AND_LOOSE ) {
		if( Jmatrix.sm_row_freelist == null ) {
			prow = new Jsm_row();
		} else {
			prow = Jmatrix.sm_row_freelist;
			Jmatrix.sm_row_freelist = prow.next_row;
		}
} else {
		prow = new Jsm_row();
}

		prow.row_num = 0;
		prow.length = 0;
		prow.first_col = prow.last_col = null;
		prow.next_row = prow.prev_row = null;
		prow.flag = false;
		prow.user_word = null;		/* for our user ... */
		return prow;
	}

	/*
	 *  free a row vector -- for FAST_AND_LOOSE, this is real cheap for rows;
	 *  however, freeing a column must still walk down the column discarding
	 *  the elements one-by-one; that is the only use for the extra '-DCOLS'
	 *  compile flag ...
	 */
	static void sm_row_free(final Jsm_row prow) {
if( Jespresso.FAST_AND_LOOSE && ! Jespresso.COLS ) {
		if( prow.first_col != null ) {
			/* Add the linked list of row items to the free list */
			prow.last_col.next_col = Jmatrix.sm_element_freelist;
			Jmatrix.sm_element_freelist = prow.first_col;
		}

		/* Add the row to the free list of rows */
		prow.next_row = Jmatrix.sm_row_freelist;
		Jmatrix.sm_row_freelist = prow;
} else {
		for( Jsm_element pnext, p = prow.first_col; p != null; p = pnext ) {
			pnext = p.next_col;
			Jmatrix.sm_element_free( p );
		}
		// prow = null;
}
	}

	/** duplicate an existing row */
	static Jsm_row sm_row_dup(final Jsm_row prow) {
		final Jsm_row pnew = sm_row_alloc();
		for( Jsm_element p = prow.first_col; p != null; p = p.next_col ) {
			sm_row_insert( pnew, p.col_num );
		}
		return pnew;
	}


	/** insert an element into a row vector */
	static Jsm_element sm_row_insert(final Jsm_row prow, final int col) {
		/* get a new item, save its address */
		final Jsm_element element = Jmatrix.sm_element_alloc();
		Jsm_element test = element;

		if( prow.last_col == null ) {
			test.col_num = col;
			prow.first_col = test;
			prow.last_col = test;
			test.next_col = null;
			test.prev_col = null;
			prow.length++;
		} else if( prow.last_col.col_num < col ) {
			test.col_num = col;
			prow.last_col.next_col = test;
			test.prev_col = prow.last_col;
			prow.last_col = test;
			test.next_col = null;
			prow.length++;
	    } else if( prow.first_col.col_num > col ) {
			test.col_num = col;
			prow.first_col.prev_col = test;
			test.next_col = prow.first_col;
			prow.first_col = test;
			test.prev_col = null;
			prow.length++;
	    } else {
	    	Jsm_element p;
	    	for( p = prow.first_col; p.col_num < col; p = p.next_col ) {
				;
			}
			if( p.col_num > col ) {
			    test.col_num = col;
			    p = p.prev_col;
			    p.next_col.prev_col = test;
			    test.next_col = p.next_col;
			    p.next_col = test;
			    test.prev_col = p;
			    prow.length++;
			} else {
			    test = p;
			}
	    }

		/* if item was not used, free it */
		if( element != test ) {
			Jmatrix.sm_element_free( element );
		}

		/* either way, return the current new value */
		return test;
	}


	/** remove an element from a row vector */
	static void sm_row_remove(final Jsm_row prow, final int col) {
		Jsm_element p;

		for( p = prow.first_col; p != null && p.col_num < col; p = p.next_col ) {
			;
		}
		if( p != null && p.col_num == col ) {
			if( p.prev_col == null ) {
				prow.first_col = p.next_col;
			} else {
				p.prev_col.next_col = p.next_col;
			}
			if( p.next_col == null ) {
				prow.last_col = p.prev_col;
			} else {
				p.next_col.prev_col = p.prev_col;
			}
			prow.length--;
			Jmatrix.sm_element_free( p );
		}
	}

	/** find an element (if it is in the row vector) */
	static Jsm_element sm_row_find(final Jsm_row prow, final int col) {
		Jsm_element p;

		for( p = prow.first_col; p != null && p.col_num < col; p = p.next_col ) {
			;
		}
		if( p != null && p.col_num == col ) {
			return p;
		}// else {
			return null;
		//}
	}

	/** return 1 if row p2 contains row p1; 0 otherwise */
	static boolean sm_row_contains(final Jsm_row p1, final Jsm_row p2) {
		Jsm_element q1 = p1.first_col;
		Jsm_element q2 = p2.first_col;
		while( q1 != null ) {
			if( q2 == null || q1.col_num < q2.col_num ) {
				return false;
			} else if( q1.col_num == q2.col_num ) {
				q1 = q1.next_col;
				q2 = q2.next_col;
			} else {
				q2 = q2.next_col;
			}
		}
		return true;
	}

	/** return 1 if row p1 and row p2 share an element in common */
	static boolean sm_row_intersects(final Jsm_row p1, final Jsm_row p2) {
		Jsm_element q1 = p1.first_col;
		Jsm_element q2 = p2.first_col;
		if( q1 == null || q2 == null ) {
			return false;
		}
		for( ;; ) {
			if( q1.col_num < q2.col_num ) {
				if( (q1 = q1.next_col) == null ) {
					return false;
				}
			} else if( q1.col_num > q2.col_num ) {
				if( (q2 = q2.next_col) == null ) {
					return false;
				}
			} else {
				return true;
			}
		}
	}

	/** compare two rows, lexical ordering */
	static int sm_row_compare(final Jsm_row p1, final Jsm_row p2 ) {
		Jsm_element q1 = p1.first_col;
		Jsm_element q2 = p2.first_col;
		while( q1 != null && q2 != null ) {
			if( q1.col_num != q2.col_num ) {
				return q1.col_num - q2.col_num;
			}
			q1 = q1.next_col;
			q2 = q2.next_col;
		}

		if( q1 != null ) {
			return 1;
		} else if( q2 != null ) {
			return -1;
		} else {
			return 0;
		}
	}

	/** return the intersection */
	static Jsm_row sm_row_and(final Jsm_row p1, final Jsm_row p2) {
		final Jsm_row result = sm_row_alloc();
		Jsm_element q1 = p1.first_col;
		Jsm_element q2 = p2.first_col;
		if( q1 == null || q2 == null ) {
			return result;
		}
		for( ; ; ) {
			if( q1.col_num < q2.col_num ) {
				if( (q1 = q1.next_col) == null ) {
					return result;
				}
			} else if( q1.col_num > q2.col_num ) {
				if( (q2 = q2.next_col) == null ) {
					return result;
				}
			} else {
				sm_row_insert( result, q1.col_num );
				if( (q1 = q1.next_col) == null ) {
					return result;
				}
				if( (q2 = q2.next_col) == null ) {
					return result;
				}
			}
		}
	}

	static int sm_row_hash(final Jsm_row prow, final int modulus) {
		int sum = 0;
		for( Jsm_element p = prow.first_col; p != null; p = p.next_col ) {
			sum = (sum * 17 + p.col_num) % modulus;
		}
		return sum;
	}

	/** remove an element from a row vector (given a pointer to the element) */
	static void sm_row_remove_element(final Jsm_row prow, final Jsm_element p) {
		if( p.prev_col == null ) {
			prow.first_col = p.next_col;
		} else {
			p.prev_col.next_col = p.next_col;
		}
		if( p.next_col == null ) {
			prow.last_col = p.prev_col;
		} else {
			p.next_col.prev_col = p.prev_col;
		}
		prow.length--;
		Jmatrix.sm_element_free( p );
	}

	@SuppressWarnings("boxing")
	static void sm_row_print(final PrintStream fp, final Jsm_row prow) {
		for( Jsm_element p = prow.first_col; p != null; p = p.next_col ) {
			fp.printf(" %d", p.col_num );
		}
	}
}