package espresso;

import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.util.Arrays;

/* LINTLIBRARY */
/**
 *   set.c -- routines for manipulating sets and set families
 */
final class Jset {
	private static Jset_family set_family_garbage = null;

	int[] base;
	int offset;

	Jset(final int[] b, final int off) {
		base = b;
		offset = off;
	}

	final int get(int i) {
		i += offset;
		return base[ i ];
	}

	/** bit_index -- find first bit (from LSB) in a word (MSB=bit n, LSB=bit 0) */
	static int bit_index(int a) {
		if( a == 0 ) {
			return -1;
		}
		int i;
		for( i = 0; (a & 1) == 0; a >>>= 1, i++ ) {
			;
		}
		return i;
	}

	/** set_ord -- count number of elements in a set */
	static int set_ord(final int[] a, final int off_a) {
		int sum = 0;
		int val;
		for( int i = Jespresso.LOOP( a, off_a ) + off_a; i > off_a; i-- ) {
			if( (val = a[ i ]) != 0 ) {
				sum += Jespresso.count_ones( val );
			}
		}
		return sum;
	}
	/** set_ord -- count number of elements in a set */
	static int set_ord(final int[] a) {
		int sum = 0;
		int val;
		for( int i = Jespresso.LOOP( a ); i > 0; i-- ) {
			if( (val = a[ i ]) != 0 ) {
				sum += Jespresso.count_ones( val );
			}
		}
		return sum;
	}

	/** set_dist -- distance between two sets (# elements in common) */
	static int set_dist(final int[] a, final int off_a, final int[] b) {
		int sum = 0;
		int val;
		for( int i = Jespresso.LOOP( a, off_a ), ai = off_a + i; i > 0; i--, ai-- ) {
			if( (val = a[ ai ] & b[ i ]) != 0 ) {
				sum += Jespresso.count_ones( val );
			}
		}
		return sum;
	}

	/** set_clear -- make "r" the empty set of "size" elements */
	static void set_clear(final int[] r, final int size) {
		int i = Jespresso.LOOPINIT( size );
		r[ 0 ] = i;
		do {
			r[ i ] = 0;
		} while( --i > 0 );
	}
	/** set_clear -- make "r" the empty set of "size" elements */
	static void set_clear(final int[] r, final int off, final int size) {
		int i = Jespresso.LOOPINIT( size );
		r[ off ] = i;
		i += off;
		do {
			r[ i ] = 0;
		} while( --i > off );
	}

	static int[] set_new(final int size) {
		final int[] b = new int[ Jespresso.SET_SIZE( size ) ];
		set_clear( b, size );
		return b;
	}

	/** set_fill -- make "r" the universal set of "size" elements */
	static void set_fill(final int[] r, final int off, final int size) {
		int i = Jespresso.LOOPINIT( size );
		r[ off ] = i;
		i += off;
		r[ i ] = (~0) >>> (i * Jespresso.BPI - size);
		while( --i > off ) {
			r[ i ] = ~0;
		}
	}
	/** set_fill -- make "r" the universal set of "size" elements */
	static void set_fill(final int[] r, final int size) {
		int i = Jespresso.LOOPINIT( size );
		r[ 0 ] = i;
		r[ i ] = (~0) >>> (i * Jespresso.BPI - size);
		while( --i > 0 ) {
			r[ i ] = ~0;
		}
	}
	/** set_fill -- make "r" the universal set of "size" elements */
	static int[] set_full(final int size) {
		final int[] b = new int[ Jespresso.SET_SIZE( size ) ];
		set_fill( b, size );
		return b;
	}

	/** set_copy -- copy set a into set r */
	static void set_copy(final int[] r, final int off_r, final int[] a, final int off_a) {
		int i = Jespresso.LOOPCOPY( a, off_a );
		int ai = off_a + i;
		i += off_r;
		do {
			r[ i ] = a[ ai-- ];
		} while( --i >= off_r );
	}
	/** set_copy -- copy set a into set r */
	static void set_copy(final int[] r, final int[] a, final int off_a) {
		int i = Jespresso.LOOPCOPY( a, off_a );
		int ai = off_a + i;
		do {
			r[ i ] = a[ ai-- ];
		} while( --i >= 0 );
	}
	/** set_copy -- copy set a into set r */
	static void set_copy(final int[] r, final int off_r, final int[] a) {
		int i = Jespresso.LOOPCOPY( a );
		int ri = off_r + i;
		do {
			r[ ri-- ] = a[ i ];
		} while( --i >= 0 );
	}
	/** set_copy -- copy set a into set r */
	static void set_copy(final int[] r, final int[] a) {
		int i = Jespresso.LOOPCOPY( a );
		do {
			r[ i ] = a[ i ];
		} while( --i >= 0 );
	}

	static int[] set_save(final Jset r) {
		final int[] b = new int[ Jespresso.SET_SIZE( Jespresso.NELEM( r ) ) ];
		set_copy( b, r.base, r.offset );
		return b;
	}
	static int[] set_save(final int[] r, final int off) {
		final int[] b = new int[ Jespresso.SET_SIZE( Jespresso.NELEM( r, off ) ) ];
		set_copy( b, r, off );
		return b;
	}
	static int[] set_save(final int[] r) {
		final int[] b = new int[ Jespresso.SET_SIZE( Jespresso.NELEM( r ) ) ];
		set_copy( b, r );
		return b;
	}

	/** set_and -- compute intersection of sets "a" and "b" */
	static void set_and(final int[] r, final int off_r, final int[] a, final int off_a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a, off_a );
		Jespresso.PUTLOOP( r, off_r, i );
		int ai = off_a + i;
		int bi = off_b + i;
		i += off_r;
		do {
			r[ i ] = a[ ai-- ] & b[ bi-- ];
		} while( --i > off_r );
    }
	/** set_and -- compute intersection of sets "a" and "b" */
	static void set_and(final int[] r, final int off_r, final int[] a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a );
		Jespresso.PUTLOOP( r, off_r, i );
		int bi = off_b + i;
		int ri = off_r + i;
		do {
			r[ ri-- ] = a[ i ] & b[ bi-- ];
		} while( --i > 0 );
    }
	/** set_and -- compute intersection of sets "a" and "b" */
	static void set_and(final int[] r, final int[] a, final int off_a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a, off_a );
		Jespresso.PUTLOOP( r, i );
		int ai = off_a + i;
		int bi = off_b + i;
		do {
			r[ i ] = a[ ai-- ] & b[ bi-- ];
		} while( --i > 0 );
    }
	/** set_and -- compute intersection of sets "a" and "b" */
	static void set_and(final int[] r, final int off_r, final int[] a, final int off_a, final int[] b) {
		int i = Jespresso.LOOP( a, off_a );
		Jespresso.PUTLOOP( r, off_r, i );
		int ai = off_a + i;
		int ri = off_r + i;
		do {
			r[ ri-- ] = a[ ai-- ] & b[ i ];
		} while( --i > 0 );
    }
	/** set_and -- compute intersection of sets "a" and "b" */
	static void set_and(final int[] r, final int off_r, final int[] a, final int[] b) {
		int i = Jespresso.LOOP( a );
		Jespresso.PUTLOOP( r, off_r, i );
		int ri = off_r + i;
		do {
			r[ ri-- ] = a[ i ] & b[ i ];
		} while( --i > 0 );
    }
	/** set_and -- compute intersection of sets "a" and "b" */
	static void set_and(final int[] r, final int[] a, final int off_a, final int[] b) {
		int i = Jespresso.LOOP( a, off_a );
		Jespresso.PUTLOOP( r, i );
		int ai = off_a + i;
		do {
			r[ i ] = a[ ai-- ] & b[ i ];
		} while( --i > 0 );
    }
	/** set_and -- compute intersection of sets "a" and "b" */
	static void set_and(final int[] r, final int[] a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a );
		Jespresso.PUTLOOP( r, i );
		int bi = off_b + i;
		do {
			r[ i ] = a[ i ] & b[ bi-- ];
		} while( --i > 0 );
    }
	/** set_and -- compute intersection of sets "a" and "b" */
	static void set_and(final int[] r, final int[] a, final int[] b) {
		int i = Jespresso.LOOP( a );
		Jespresso.PUTLOOP( r, i );
		do {
			r[ i ] = a[ i ] & b[ i ];
		} while( --i > 0 );
    }

	/** set_or -- compute union of sets "a" and "b" */
	static Jset set_or(final Jset r, final int[] a, final int off_a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a, off_a );
		Jespresso.PUTLOOP( r, i );
		final int[] rb = r.base;
		int ri = r.offset + i;
		int bi = off_b + i;
		i += off_a;
		do {
			rb[ ri-- ] = a[ i ] | b[ bi-- ];
		} while( --i > off_a );
		return r;
	}
	/** set_or -- compute union of sets "a" and "b" */
	static int[] set_or(final int[] r, final int[] a, final int off_a, final int[] b) {
		int i = Jespresso.LOOP( a, off_a );
		Jespresso.PUTLOOP( r, i );
		int ai = off_a + i;
		do {
			r[ i ] = a[ ai-- ] | b[ i ];
		} while( --i > 0 );
		return r;
	}
	/** set_or -- compute union of sets "a" and "b" */
	static void set_or(final int[] r, final int off_r, final int[] a, final int off_a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a, off_a );
		Jespresso.PUTLOOP( r, off_r, i );
		int ai = off_a + i;
		int bi = off_b + i;
		i += off_r;
		do {
			r[ i ] = a[ ai-- ] | b[ bi-- ];
		} while( --i > off_r );
	}
	/** set_or -- compute union of sets "a" and "b" */
	static void set_or(final int[] r, final int[] a, final int off_a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a, off_a );
		Jespresso.PUTLOOP( r, i );
		int ai = off_a + i;
		int bi = off_b + i;
		do {
			r[ i ] = a[ ai-- ] | b[ bi-- ];
		} while( --i > 0 );
	}
	/** set_or -- compute union of sets "a" and "b" */
	static void set_or(final int[] r, final int off_r, final int[] a, final int off_a, final int[] b) {
		int i = Jespresso.LOOP( a, off_a );
		Jespresso.PUTLOOP( r, off_r, i );
		int ai = off_a + i;
		int ri = off_r + i;
		do {
			r[ ri-- ] = a[ ai-- ] | b[ i ];
		} while( --i > 0 );
	}
	/** set_or -- compute union of sets "a" and "b" */
	static void set_or(final int[] r, final int[] a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a );
		Jespresso.PUTLOOP( r, i );
		int bi = off_b + i;
		do {
			r[ i ] = a[ i ] | b[ bi-- ];
		} while( --i > 0 );
	}
	/** set_or -- compute union of sets "a" and "b" */
	static void set_or(final int[] r, final int off_r, final int[] a, final int[] b) {
		int i = Jespresso.LOOP( a );
		Jespresso.PUTLOOP( r, off_r, i );
		int ri = off_r + i;
		do {
			r[ ri-- ] = a[ i ] | b[ i ];
		} while( --i > 0 );
	}
	/** set_or -- compute union of sets "a" and "b" */
	static void set_or(final int[] r, final int[] a, final int[] b) {
		int i = Jespresso.LOOP( a );
		Jespresso.PUTLOOP( r, i );
		do {
			r[ i ] = a[ i ] | b[ i ];
		} while( --i > 0 );
	}

	/** set_diff -- compute difference of sets "a" and "b" */
	static void set_diff(final int[] r, final int off_r, final int[] a, final int off_a, final int[] b) {
		int i = Jespresso.LOOP( a, off_a );
		Jespresso.PUTLOOP( r, off_r, i );
		int ai = off_a + i;
		int ri = off_r + i;
		do {
			r[ ri-- ] = a[ ai-- ] & ~ b[ i ];
		} while( --i > 0 );
	}
	/** set_diff -- compute difference of sets "a" and "b" */
	static void set_diff(final int[] r, final int off_r, final int[] a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a );
		Jespresso.PUTLOOP( r, off_r, i );
		int ri = off_r + i;
		int bi = off_b + i;
		do {
			r[ ri-- ] = a[ i ] & ~ b[ bi-- ];
		} while( --i > 0 );
	}
	/** set_diff -- compute difference of sets "a" and "b" */
	static void set_diff(final int[] r, final int[] a, final int off_a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a, off_a );
		Jespresso.PUTLOOP( r, i );
		int ai = off_a + i;
		int bi = off_b + i;
		do {
			r[ i ] = a[ ai-- ] & ~ b[ bi-- ];
		} while( --i > 0 );
	}
	/** set_diff -- compute difference of sets "a" and "b" */
	static void set_diff(final int[] r, final int[] a, final int off_a, final int[] b) {
		int i = Jespresso.LOOP( a, off_a );
		Jespresso.PUTLOOP( r, i );
		int ai = off_a + i;
		do {
			r[ i ] = a[ ai-- ] & ~ b[ i ];
		} while( --i > 0 );
	}
	/** set_diff -- compute difference of sets "a" and "b" */
	static void set_diff(final int[] r, final int[] a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a );
		Jespresso.PUTLOOP( r, i );
		int bi = off_b + i;
		do {
			r[ i ] = a[ i ] & ~ b[ bi-- ];
		} while( --i > 0 );
	}
	/** set_diff -- compute difference of sets "a" and "b" */
	static void set_diff(final int[] r, final int[] a, final int[] b) {
		int i = Jespresso.LOOP( a );
		Jespresso.PUTLOOP( r, i );
		do {
			r[ i ] = a[ i ] & ~ b[ i ];
		} while( --i > 0 );
	}

	/** set_xor -- compute exclusive-or of sets "a" and "b" */
	static void set_xor(final int[] r, final int[] a, final int[] b) {
		int i = Jespresso.LOOP( a );
		Jespresso.PUTLOOP( r, i );
		do {
			r[ i ] = a[ i ] ^ b[ i ];
		} while( --i > 0 );
	}

	/** set_merge -- compute "a" & "mask" | "b" & ~ "mask" */
	static void set_merge(final int[] r, final int off_r, final int[] a, final int[] b, final int[] mask) {
		int i = Jespresso.LOOP( a );
		Jespresso.PUTLOOP( r, off_r, i );
		int ri = off_r + i;
		do {
			r[ ri-- ] = (a[ i ] & mask[ i ]) | (b[ i ] & ~mask[ i ]);
		} while( --i > 0 );
	}
	static void set_merge(final int[] r, final int[] a, final int off_a, final int[] b, final int[] mask) {
		int i = Jespresso.LOOP( a, off_a );
		Jespresso.PUTLOOP( r, i );
		int ai = off_a + i;
		do {
			r[ i ] = (a[ ai-- ] & mask[ i ]) | (b[ i ] & ~mask[ i ]);
		} while( --i > 0 );
	}
	static void set_merge(final int[] r, final int[] a, final int[] b, final int off_b, final int[] mask) {
		int i = Jespresso.LOOP( a );
		Jespresso.PUTLOOP( r, i );
		int bi = off_b + i;
		do {
			r[ i ] = (a[ i ] & mask[ i ]) | (b[ bi-- ] & ~mask[ i ]);
		} while( --i > 0 );
	}

	/** set_andp -- compute intersection of sets "a" and "b" , TRUE if nonempty */
	static boolean set_andp(final int[] r, final int off_r, final int[] a, final int off_a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a, off_a );
		int x = 0;
		Jespresso.PUTLOOP( r, off_r, i );
		int ai = off_a + i;
		int bi = off_b + i;
		i += off_r;
		do {
			final int ri = a[ ai-- ] & b[ bi-- ];
			r[ i ] = ri;
			x |= ri;
		} while( --i > off_r );
		return x != 0;
	}

	/** set_orp -- compute union of sets "a" and "b" , TRUE if nonempty */
	static boolean set_orp(final int[] r, final int[] a, final int[] b) {
		int i = Jespresso.LOOP( a );
		int x = 0;
		Jespresso.PUTLOOP( r, i );
		do {
			r[ i ] = a[ i ] | b[ i ];
			x |= r[ i ];
		} while( --i > 0 );
		return x != 0;
	}

	/** setp_empty -- check if the set "a" is empty */
	static boolean setp_empty(final int[] a) {
		int i = Jespresso.LOOP( a );
		do {
			if( a[ i ] != 0 ) {
				return false;
			}
		} while( --i > 0 );
		return true;
	}

	/** setp_full -- check if the set "a" is the full set of "size" elements */
	static boolean setp_full(final int[] a, final int size) {
		int i = Jespresso.LOOP( a );
		int test = ~ 0;
		test >>>= i * Jespresso.BPI - size;
		if( a[ i ] != test ) {
			return false;
		}
		while( --i > 0 ) {
			if( a[ i ] != ~0 ) {
				return false;
			}
		}
		return true;
	}

	/** setp_equal -- check if the set "a" equals set "b" */
	static boolean setp_equal(final int[] a, final int off_a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a, off_a );
		int bi = off_b + i;
		i += off_a;
		do {
			if( a[ i ] != b[ bi-- ] ) {
				return false;
			}
		} while( --i > off_a );
		return true;
	}
	/** setp_equal -- check if the set "a" equals set "b" */
	static boolean setp_equal(final int[] a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a );
		int bi = off_b + i;
		do {
			if( a[ i ] != b[ bi-- ] ) {
				return false;
			}
		} while( --i > 0 );
		return true;
	}
	/** setp_equal -- check if the set "a" equals set "b" */
	static boolean setp_equal(final int[] a, final int off_a, final int[] b) {
		int i = Jespresso.LOOP( a, off_a );
		int ai = off_a + i;
		do {
			if( a[ ai-- ] != b[ i ] ) {
				return false;
			}
		} while( --i > 0 );
		return true;
	}
	/** setp_equal -- check if the set "a" equals set "b" */
	static boolean setp_equal(final int[] a, final int[] b) {
		int i = Jespresso.LOOP( a );
		do {
			if( a[ i ] != b[ i ] ) {
				return false;
			}
		} while( --i > 0 );
		return true;
	}

	/** setp_disjoint -- check if intersection of "a" and "b" is empty */
	static boolean setp_disjoint(final int[] a, final int off_a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a, off_a );
		int bi = off_b + i;
		i += off_a;
		do {
			if( (a[ i ] & b[ bi-- ]) != 0 ) {
				return false;
			}
		} while( --i > off_a );
		return true;
	}
	/** setp_disjoint -- check if intersection of "a" and "b" is empty */
	static boolean setp_disjoint(final int[] a, final int off_a, final int[] b) {
		int i = Jespresso.LOOP( a, off_a );
		int ai = off_a + i;
		do {
			if( (a[ ai-- ] & b[ i ]) != 0 ) {
				return false;
			}
		} while( --i > 0 );
		return true;
	}
	/** setp_disjoint -- check if intersection of "a" and "b" is empty */
	static boolean setp_disjoint(final int[] a, final int[] b) {
		int i = Jespresso.LOOP( a );
		do {
			if( (a[ i ] & b[ i ]) != 0 ) {
				return false;
			}
		} while( --i > 0 );
		return true;
	}

	/** setp_implies -- check if "a" implies "b" ("b" contains "a") */
	static boolean setp_implies(final int[] a, final int off_a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a, off_a );
		int bi = off_b + i;
		i += off_a;
		do {
			if( (a[ i ] & ~b[ bi-- ]) != 0 ) {
				return false;
			}
		} while( --i > off_a );
		return true;
	}
	/** setp_implies -- check if "a" implies "b" ("b" contains "a") */
	static boolean setp_implies(final int[] a, final int off_a, final int[] b) {
		int i = Jespresso.LOOP( a, off_a );
		int ai = off_a + i;
		do {
			if( (a[ ai-- ] & ~b[ i ]) != 0 ) {
				return false;
			}
		} while( --i > 0 );
		return true;
	}
	/** setp_implies -- check if "a" implies "b" ("b" contains "a") */
	static boolean setp_implies(final int[] a, final int[] b, final int off_b) {
		int i = Jespresso.LOOP( a );
		int bi = off_b + i;
		do {
			if( (a[ i ] & ~b[ bi-- ]) != 0 ) {
				return false;
			}
		} while( --i > 0 );
		return true;
	}
	/** setp_implies -- check if "a" implies "b" ("b" contains "a") */
	static boolean setp_implies(final int[] a, final int[] b) {
		int i = Jespresso.LOOP( a );
		do {
			if( (a[ i ] & ~b[ i ]) != 0 ) {
				return false;
			}
		} while( --i > 0 );
		return true;
	}

	/** sf_or -- form the "or" of all sets in a set family */
	static int[] sf_or(final Jset_family A) {
		final int[] or = set_new( A.sf_size );
		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			set_or( or, or, data, p );
		}
		return or;
	}

	/** sf_and -- form the "and" of all sets in a set family */
	static int[] sf_and(final Jset_family A) {
		final int[] and = set_new( A.sf_size );
		set_fill( and, A.sf_size );
		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			set_and( and, and, data, p );
		}
		return and;
	}

	/** sf_active -- make all members of the set family active */
	static Jset_family sf_active(final Jset_family A) {
		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			Jespresso.SET( data, p, Jespresso.ACTIVE );
		}
		A.active_count = A.count;
		return A;
	}

	/** sf_inactive -- remove all inactive cubes in a set family */
	static Jset_family sf_inactive(final Jset_family A) {
		final int[] data = A.data;// java
		int pdest = 0;// A.data;
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			if( Jespresso.TESTP( data, p, Jespresso.ACTIVE ) ) {
				if( pdest != p ) {
					set_copy( data, pdest, data, p );
				}
				pdest += A.wsize;
			} else {
				A.count--;
			}
		}
		return A;
	}

	/** sf_copy -- copy a set family */
	static Jset_family sf_copy(final Jset_family R, final Jset_family A) {
		R.sf_size = A.sf_size;
		R.wsize = A.wsize;
		/*R.capacity = A.count;*/
		/*R.data = REALLOC(unsigned int, R.data, (long) R.capacity * R.wsize);*/
		R.count = A.count;
		R.active_count = A.active_count;
		System.arraycopy( A.data, 0, R.data, 0, A.wsize * A.count );
		return R;
	}

	/** sf_join -- join A and B into a single set_family */
	static Jset_family sf_join(final Jset_family A, final Jset_family B) {
		Jset_family R;
		final int asize = A.count * A.wsize;
		final int bsize = B.count * B.wsize;

		if( A.sf_size != B.sf_size ) {
			Jcvrmisc.fatal("sf_join: sf_size mismatch");
		}
		R = sf_new(A.count + B.count, A.sf_size);
		R.count = A.count + B.count;
		R.active_count = A.active_count + B.active_count;
		System.arraycopy( A.data, 0, R.data, 0, asize );
		System.arraycopy( B.data, 0, R.data, asize, bsize );
		return R;
	}

	/** sf_append -- append the sets of B to the end of A, and dispose of B */
	static Jset_family sf_append(final Jset_family A, final Jset_family B) {
		final int asize = A.count * A.wsize;
		final int bsize = B.count * B.wsize;

		if( A.sf_size != B.sf_size ) {
			Jcvrmisc.fatal("sf_append: sf_size mismatch");
		}
		A.capacity = A.count + B.count;
		A.data = Arrays.copyOf( A.data, A.capacity * A.wsize );
		System.arraycopy( B.data, 0, A.data, asize, bsize );
		A.count += B.count;
		A.active_count += B.active_count;
		sf_free( B );
		return A;
	}

	/** sf_new -- allocate "num" sets of "size" elements each */
	static Jset_family sf_new(final int num, final int size) {
		Jset_family A;
		if( set_family_garbage == null ) {
			A = new Jset_family();
		} else {
			A = set_family_garbage;
			set_family_garbage = A.next;
		}
		A.sf_size = size;
		A.wsize = Jespresso.SET_SIZE( size );
		A.capacity = num;
		A.data = new int[ A.capacity * A.wsize ];
		A.count = 0;
		A.active_count = 0;
		return A;
	}

	/** sf_save -- create a duplicate copy of a set family */
	static Jset_family sf_save(final Jset_family A) {
		return sf_copy( sf_new( A.count, A.sf_size ), A );
	}

	/** sf_free -- free the storage allocated for a set family */
	static void sf_free(final Jset_family A) {
		A.data = null;
		A.next = set_family_garbage;
		set_family_garbage = A;
	}

	/** sf_cleanup -- free all of the set families from the garbage list */
	static void sf_cleanup() {
		//for( Jset_family pnext, p = set_family_garbage; p != null; p = pnext ) {
		//	pnext = p.next;
		//	p = null;
		//}
		set_family_garbage = null;
	}

	/** sf_addset -- add a set to the end of a set family */
	static Jset_family sf_addset(final Jset_family A, final int[] s, final int off) {
		if( A.count >= A.capacity ) {
			A.capacity = A.capacity + (A.capacity >>> 1) + 1;
			A.data = Arrays.copyOf( A.data, A.capacity * A.wsize );
		}
		set_copy( A.data, A.wsize * A.count++, s, off );
		return A;
	}
	/** sf_addset -- add a set to the end of a set family */
	static Jset_family sf_addset(final Jset_family A, final int[] s) {
		if( A.count >= A.capacity ) {
			A.capacity = A.capacity + (A.capacity >>> 1) + 1;
			A.data = Arrays.copyOf( A.data, A.capacity * A.wsize );
		}
		set_copy( A.data, A.wsize * A.count++, s );
		return A;
	}

	/** sf_delset -- delete a set from a set family */
	static void sf_delset(final Jset_family A, final int i) {
		final int[] data = A.data;
		set_copy( data, A.wsize * i, data, A.wsize * --A.count );
	}

	/** sf_print -- print a set_family as a set (list the element numbers) */
	@SuppressWarnings("boxing")
	static void sf_print(final Jset_family A) {
		final int[] data = A.data;// java
		for( int p = 0, i = 0; i < A.count; p += A.wsize, i++ ) {
			System.out.printf("A[%d] = %s\n", i, ps1( data, p ) );
		}
	}

	/** sf_bm_print -- print a set_family as a bit-matrix */
	@SuppressWarnings("boxing")
	static void sf_bm_print(final Jset_family A) {
		final int[] data = A.data;// java
		for( int p = 0, i = 0; i < A.count; p += A.wsize, i++ ) {
			System.out.printf("[%4d] %s\n", i, pbv1( data, p, A.sf_size ) );
		}
	}

	/** sf_write -- output a set family in an unintelligable manner */
	@SuppressWarnings("boxing")
	static void sf_write(final PrintStream fp, final Jset_family A) {
		fp.printf("%d %d\n", A.count, A.sf_size );
		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			set_write( fp, data, p );
		}
		fp.flush();
	}

	/** sf_read -- read a set family written by sf_write */
	static Jset_family sf_read(final InputStream fp) throws IOException {
		// final int i, j;
		// fscanf(fp, "%d %d\n", &i, &j);
		final int i = Jsupport.scanInt( fp, 10 );
		int j = Jsupport.scanInt( fp, 10 );
		final Jset_family A = sf_new( i, j );
		A.count = i;
		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			data[ p ] = Jsupport.scanInt( fp, 16 );// fscanf(fp, "%x", data[ p ] );
			for( j = 1; j <= Jespresso.LOOP( data, p ); j++ ) {
				data[ p + j ] = Jsupport.scanInt( fp, 16 );// fscanf(fp, "%x", data[ p + j ] );
			}
		}
		return A;
	}

	/** set_write -- output a set in an unintelligable manner */
	@SuppressWarnings("boxing")
	static void set_write(final PrintStream fp, final int[] a, int off_a) {
		final int n = Jespresso.LOOP( a, off_a );

		for( int j = 0; j <= n; j++, off_a++ ) {
			fp.printf("%x ", a[ off_a ] );
			if( (j + 1) % 8 == 0 && j != n ) {
				fp.printf("\n\t");
			}
		}
		fp.printf("\n");
	}

	/** sf_bm_read -- read a set family written by sf_bm_print (almost) */
	static Jset_family sf_bm_read(final InputStream fp) throws IOException {
		// fscanf(fp, "%d %d\n", &rows, &cols);
		final int rows = Jsupport.scanInt( fp, 10 );
		final int cols = Jsupport.scanInt( fp, 10 );
		final Jset_family A = sf_new( rows, cols );
		for( int i = 0; i < rows; i++ ) {
			final int[] pdest_base = A.data;
			final int pdest = A.wsize * A.count++;
			set_clear( pdest_base, pdest, A.sf_size );
			for( int j = 0; j < cols; j++ ) {
				switch( fp.read() ) {
				case '0':
					break;
				case '1':
					Jespresso.set_insert( pdest_base, pdest, j );
					break;
				default:
					Jcvrmisc.fatal("Error reading set family");
				}
			}
			if( fp.read() != '\n' ) {
				Jcvrmisc.fatal("Error reading set family (at end of line)");
			}
		}
		return A;
	}

	/** ps1 -- convert a set into a printable string */
	private static final int largest_string = 120;
	private static char s1[] = new char[ largest_string ];

	/** ps1 -- convert a set into a printable string */
	static String ps1(final int[] a, final int off_a) {
		int len = 0;
		final int n = Jespresso.NELEM( a, off_a );
		final char temp[] = new char[ 20 ];
		boolean first = true;

		s1[ len++ ] = '[';
		for( int i = 0; i < n; i++ ) {
			if( Jespresso.is_in_set( a, off_a, i ) ) {
				if( ! first ) {
					s1[ len++ ] = ',';
				}
				first = false;
				int num = i;
				/* Generate digits (reverse order) */
				int l = 0;
				do {
					temp[ l++ ] = (char)(num % 10 + '0');
				} while( (num /= 10) > 0 );
				/* Copy them back in correct order */
				do {
					s1[ len++ ] = temp[ --l ];
				} while( l > 0 );
				if( len > largest_string - 15 ) {
					s1[ len++ ] = '.'; s1[ len++ ] = '.'; s1[ len++ ] = '.';
					break;
				}
			}
		}

		s1[ len++ ] = ']';
		s1[ len++ ] = '\0';
		return new String( s1 );
	}

	/** pbv1 -- print bit-vector */
	static String pbv1(final int[] s, final int off_s, final int n) {
		for( int i = 0; i < n; i++ ) {
			s1[ i ] = Jespresso.is_in_set( s, off_s, i ) ? '1' : '0';
		}
		s1[ n ] = '\0';
		return new String( s1 );
	}

	/** set_adjcnt -- adjust the counts for a set by "weight" */
	static void set_adjcnt(final int[] a, final int off_a, final int[] count, final int weight) {
		for( int i = Jespresso.LOOP( a, off_a ); i > 0; ) {
			for( int val = a[ off_a + i ], base = --i << Jespresso.LOGBPI; val != 0; base++, val >>>= 1 ) {
				if( (val & 1) != 0 ) {
					count[ base ] += weight;
				}
			}
		}
	}

	/** sf_count -- perform a column sum over a set family */
	static int[] sf_count(final Jset_family A) {
		final int[] count = new int[ A.sf_size ];// java: already zeroed
		//for( int i = A.sf_size - 1; i >= 0; i-- ) {
		//	count[ i ] = 0;
		//}

		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			for( int i = Jespresso.LOOP( data, p ); i > 0; ) {
				for( int val = data[ p + i ], base = --i << Jespresso.LOGBPI; val != 0; base++, val >>>= 1 ) {
					if( (val & 1) != 0 ) {
						count[ base ]++;
					}
				}
			}
		}
		return count;
	}

	/* sf_count_restricted -- perform a column sum over a set family, restricting
	* to only the columns which are in r; also, the columns are weighted by the
	* number of elements which are in each row
	*/
	static int[] sf_count_restricted(final Jset_family A, final int[] r) {
		final int[] count = new int[ A.sf_size ];// java: already zeroed
		//for( int i = A.sf_size - 1; i >= 0; i-- ) {
		//	count[ i ] = 0;
		//}

		/* Loop for each set */
		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			final int weight = 1024 / (set_ord( data, p ) - 1);
			for( int i = Jespresso.LOOP( data, p ); i > 0; ) {
				for( int val = data[ p + i ] & r[ i ], base= --i << Jespresso.LOGBPI; val != 0; base++, val >>>= 1 ) {
					if( (val & 1) != 0 ) {
						count[ base ] += weight;
					}
				}
			}
		}
		return count;
	}

	/** sf_delc -- delete columns first ... last of A */
	static Jset_family sf_delc(final Jset_family A, final int first, final int last) {
		return sf_delcol( A, first, last - first + 1 );
	}

	/**
	*  sf_addcol -- add columns to a set family; includes a quick check to see
	*  if there is already enough room (and hence, can avoid copying)
	*/
	static Jset_family sf_addcol(final Jset_family A, final int firstcol, final int n) {
		/* Check if adding columns at the end ... */
		if( firstcol == A.sf_size ) {
			/* If so, check if there is already enough room */
			final int maxsize = Jespresso.BPI * Jespresso.LOOPINIT(A.sf_size);
			if( (A.sf_size + n) <= maxsize ) {
				A.sf_size += n;
				return A;
			}
		}
		return sf_delcol( A, firstcol, -n );
	}

	/**
	 * sf_delcol -- add/delete columns to/from a set family
	 *
	 *  if n > 0 then n columns starting from firstcol are deleted if n < 0
	 *  then n blank columns are inserted starting at firstcol
	 *      (i.e., the first new column number is firstcol)
	 *
	 *  This is done by copying columns in the array which is a relatively
	 *  slow operation.
	 */
	static Jset_family sf_delcol(final Jset_family A, final int firstcol, final int n) {
		final Jset_family B = sf_new( A.count, A.sf_size - n );
		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			final int[] pdest_base = B.data;
			final int pdest = B.wsize * B.count++;
			set_clear( pdest_base, pdest, B.sf_size );
			for( int i = 0; i < firstcol; i++ ) {
				if( Jespresso.is_in_set( data, p, i ) ) {
					Jespresso.set_insert( pdest_base, pdest, i );
				}
			}
			for( int i = n > 0 ? firstcol + n : firstcol; i < A.sf_size; i++ ) {
				if( Jespresso.is_in_set( data, p, i ) ) {
					Jespresso.set_insert( pdest_base, pdest, i - n );
				}
			}
		}
		sf_free( A );
		return B;
	}

	/**
	 *  sf_copy_col -- copy column "srccol" from "src" to column "dstcol" of "dst"
	 */
	static Jset_family sf_copy_col(final Jset_family dst, final int dstcol, final Jset_family src, final int srccol) {
		/* CHEAT! form these constants outside the loop */
		final int word_test = Jespresso.WHICH_WORD( srccol );
		final int bit_test = 1 << Jespresso.WHICH_BIT( srccol );
		final int word_set = Jespresso.WHICH_WORD( dstcol );
		final int bit_set = 1 << Jespresso.WHICH_BIT( dstcol );

		final int[] pdest_base = dst.data;// java
		int pdest = word_set;// dst.data;
		final int[] data = src.data;// java
		for( int p = word_test, last = p + src.count * src.wsize; p < last; p += src.wsize ) {
			if( (data[ p ] & bit_test) != 0 ) {
				pdest_base[ pdest ] |= bit_set;
			}
			/*
			 *  equivalent code for this is ...
			 *	if( is_in_set(p, srccol)) set_insert(pdest, destcol);
			 */
			pdest += dst.wsize;
		}
		return dst;
	}

	/** sf_compress -- delete columns from a matrix
	 * @param A will be feed
	 */
	static Jset_family sf_compress(final Jset_family A, final int[] c) {
		/* create a clean set family for the result */
		final Jset_family B = sf_new( A.count, set_ord( c ) );
		for( int i = 0; i < A.count; i++ ) {
			set_clear( B.data, B.wsize * B.count++, B.sf_size );
		}

		/* copy each column of A which has a 1 in c */
		int bcol = 0;
		for( int i = 0; i < A.sf_size; i++ ) {
			if( Jespresso.is_in_set( c, i ) ) {
				sf_copy_col( B, bcol++, A, i );
			}
		}
		sf_free( A );
		return B;
	}

	/**
	 *  sf_transpose -- transpose a bit matrix
	 *
	 *  There are trickier ways of doing this, but this works.
	 */
	static Jset_family sf_transpose(final Jset_family A) {
		final Jset_family B = sf_new( A.sf_size, A.count );
		B.count = A.sf_size;
		int[] data = B.data;// java
		for( int p = 0, i = 0; i < B.count; p += B.wsize, i++ ) {
			set_clear( data, p, B.sf_size );
		}

		data = A.data;// java
		for( int p = 0, i = 0; i < A.count; p += A.wsize, i++ ) {
			for( int j = 0; j < A.sf_size; j++ ) {
				if( Jespresso.is_in_set( data, p, j ) ) {
					Jespresso.set_insert( B.data, B.wsize * j, i );
				}
			}
		}
		sf_free( A );
		return B;
	}

	/**
	 *   sf_permute -- permute the columns of a set_family
	 *
	 *   permute is an array of integers containing column numbers of A which
	 *   are to be retained.
	 */
	static Jset_family sf_permute(final Jset_family A, final int[] permute, final int npermute) {
		final Jset_family B = sf_new( A.count, npermute );
		B.count = A.count;

		int[] data = B.data;// java
		for( int p = 0, last = p + B.count * A.wsize; p < last; p += B.wsize ) {
			set_clear( data, p, npermute );
		}

		final int[] pdest_base = B.data;// java
		int pdest = 0;// B.data;

		data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			for( int j = 0; j < npermute; j++ ) {
				if( Jespresso.is_in_set( data, p, permute[ j ] ) ) {
					Jespresso.set_insert( pdest_base, pdest, j );
				}
			}
			pdest += B.wsize;
		}
		sf_free( A );
		return B;
	}
}