package espresso;
/*
    sharp.c -- perform sharp, disjoint sharp, and intersection
*/

final class Jsharp {

	private static long start_time;

	/** cv_sharp -- form the sharp product between two covers */
	static Jset_family cv_sharp(final Jset_family A, final Jset_family B) {
		Jset_family T = Jespresso.new_cover( 0 );

		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			T = Jcontain.sf_union( T, Jsharp.cb_sharp( data, p, B ) );
		}
		return T;
	}

	/** cb_sharp -- form the sharp product between a cube and a cover */
	static Jset_family cb_sharp(final int[] c, final int off, Jset_family T) {
		if( T.count == 0 ) {
			T = Jset.sf_addset( Jespresso.new_cover( 1 ), c, off );
		} else {
			start_time = System.currentTimeMillis();
			T = cb_recur_sharp( c, off, T, 0, T.count - 1, 0 );
		}
		return T;
	}

	/** recursive formulation to provide balanced merging */
	@SuppressWarnings("boxing")
	static Jset_family cb_recur_sharp(final int[] c, final int off, final Jset_family T, final int first, final int last, final int level) {
		Jset_family temp;
		if( first == last ) {
			temp = sharp( c, off, T.data, T.wsize * first );
		} else {
			final int middle = (first + last) >> 1;
			final Jset_family left = cb_recur_sharp( c, off, T, first, middle, level + 1 );
			final Jset_family right = cb_recur_sharp( c, off, T, middle + 1, last, level + 1 );
			temp = cv_intersect( left, right );
			if( (Jglobals.debug & Jespresso.SHARP) != 0 && level < 4 ) {
				System.out.printf("# SHARP[%d]: %4d = %4d x %4d, time = %s\n",
						level, temp.count, left.count, right.count,
						Jespresso.print_time( System.currentTimeMillis() - start_time ) );
				System.out.flush();
			}
			Jespresso.free_cover( left );
			Jespresso.free_cover( right );
		}
		return temp;
	}


	/** sharp -- form the sharp product between two cubes */
	static Jset_family sharp(final int[] a, final int off_a, final int[] b, final int off_b) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int[] d = cube.temp[0], temp = cube.temp[1], temp1 = cube.temp[2];
		Jset_family r = Jespresso.new_cover( cube.num_vars );

		if( Jsetc.cdist0( a, off_a, b, off_b ) ) {
			Jset.set_diff( d, a, off_a, b, off_b );
			for( int var = 0; var < cube.num_vars; var++ ) {
				Jset.set_and( temp, d, cube.var_mask[var] );
				if( ! Jset.setp_empty( temp ) ) {
					Jset.set_diff( temp1, a, off_a, cube.var_mask[ var ] );
					Jset.set_or( r.data, r.wsize * r.count++, temp, temp1 );
				}
			}
		} else {
			r = Jset.sf_addset( r, a, off_a );
		}
		return r;
	}

	static Jset_family make_disjoint(final Jset_family A) {
		Jset_family R = Jespresso.new_cover( 0 );

		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			final Jset_family n = cb_dsharp( data, p, R );
			R = Jset.sf_append( R, n );
		}
		return R;
	}

	/** cv_dsharp -- disjoint-sharp product between two covers */
	static Jset_family cv_dsharp(final Jset_family A, final Jset_family B) {
		Jset_family T = Jespresso.new_cover( 0 );

		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			T = Jcontain.sf_union( T, cb_dsharp( data, p, B ) );
		}
		return T;
	}


	/** cb1_dsharp -- disjoint-sharp product between a cover and a cube */
	static Jset_family cb1_dsharp(final Jset_family T, final int[] c, final int off) {
		Jset_family R = Jespresso.new_cover( T.count );

		final int[] data = T.data;// java
		for( int p = 0, last = p + T.count * T.wsize; p < last; p += T.wsize ) {
			R = Jcontain.sf_union( R, dsharp( data, p, c, off ) );
		}
		return R;
	}

	/** cb_dsharp -- disjoint-sharp product between a cube and a cover */
	static Jset_family cb_dsharp(final int[] c, final int off, final Jset_family T) {
		Jset_family Y, Y1;

		if( T.count == 0 ) {
			Y = Jset.sf_addset( Jespresso.new_cover( 1 ), c, off );
		} else {
			Y = Jespresso.new_cover( T.count );
			Jset.set_copy( Y.data, Y.wsize * Y.count++, c, off );
			final int[] data = T.data;// java
			for( int p = 0, last = p + T.count * T.wsize; p < last; p += T.wsize ) {
				Y1 = cb1_dsharp( Y, data, p );
				Jespresso.free_cover( Y );
				Y = Y1;
			}
		}
		return Y;
	}

	/** dsharp -- form the disjoint-sharp product between two cubes */
	static Jset_family dsharp(final int[] a, final int off_a, final int[] b, final int off_b) {
		final Jcube_struct cube = Jglobals.cube;// java
		final int[] temp1 = cube.temp[0];

		Jset_family r = Jespresso.new_cover( cube.num_vars );

		if( Jsetc.cdist0( a, off_a, b, off_b ) ) {
			int[] diff = Jespresso.new_cube();
			Jset.set_diff( diff, a, off_a, b, off_b );
			int[] and = Jespresso.new_cube();
			Jset.set_and( and, a, off_a, b, off_b );
			int[] mask = Jespresso.new_cube();
			for( int var = 0; var < cube.num_vars; var++) {
				/* check if position var of "a and not b" is not empty */
				if( ! Jset.setp_disjoint(diff, cube.var_mask[var]) ) {

					/* coordinate var equals the difference between a and b */
					final int[] base = r.data;// GETSET( r, r.count++ );
					final int temp = r.wsize * r.count++;
					Jset.set_and( base, temp, diff, cube.var_mask[var] );

					/* coordinates 0 ... var-1 equal the intersection */
					Jset.set_and( temp1, and, mask );
					Jset.set_or( base, temp, base, temp, temp1 );

					/* coordinates var+1 .. cube.num_vars equal a */
					Jset.set_or( mask, mask, cube.var_mask[var] );
					Jset.set_diff( temp1, a, off_a, mask );
					Jset.set_or( base, temp, base, temp, temp1 );
				}
			}
			diff = null;
			and = null;
			mask = null;
		} else {
			r = Jset.sf_addset( r, a, off_a );
		}
		return r;
	}

	/* cv_intersect -- form the intersection of two covers */

	/** save 500 cubes before containment */
	private static int MAGIC = 500;

	static Jset_family cv_intersect(final Jset_family A, final Jset_family B) {
		Jset_family Tsave = null;

		/* How large should each temporary result cover be ? */
		Jset_family T = Jespresso.new_cover( MAGIC );
		int[] pt_base = T.data;
		int pt = 0;// java T.data;

		/* Form pairwise intersection of each cube of A with each cube of B */
		final int[] a_data = A.data;// java
		for( int pi = 0, lasti = pi + A.count * A.wsize; pi < lasti; pi += A.wsize ) {

			final int[] b_data = B.data;// java
			for( int pj = 0, lastj = pj + B.count * B.wsize; pj < lastj; pj += B.wsize ) {
				if( Jsetc.cdist0( a_data, pi, b_data, pj ) ) {
					Jset.set_and( pt_base, pt, a_data, pi, b_data, pj );
					if( ++T.count >= T.capacity ) {
						if( Tsave == null ) {
							Tsave = Jcontain.sf_contain( T );
						} else {
							Tsave = Jcontain.sf_union( Tsave, Jcontain.sf_contain( T ) );
						}
						T = Jespresso.new_cover( MAGIC );
						pt_base = T.data;
						pt = 0;
					} else {
						pt += T.wsize;
					}
				}
			}
		}

		if( Tsave == null ) {
			Tsave = Jcontain.sf_contain( T );
		} else {
			Tsave = Jcontain.sf_union( Tsave, Jcontain.sf_contain( T ) );
		}
		return Tsave;
	}
}