package espresso;

/** Module:sigma.c
 * Purpose:
 *	Contains routines for computing the signature cube of the given cube.
 * Routines:
 * pcube get_sigma():
 *	Computes signature cube of the given cube d.
 * void set_not():
 */
final class Jsigma {

	/**
	 * get_sigma:
	 * Objective: computes signature cube corresponding to the cube c
	 * Input:
	 *	R: OFFSET cover;
	 *	c: ONSET cube;
	 * Output:
	 * 	signature cube
	 */
	static int[] get_sigma(final Jset_family R, final int[] c, final int off_c) {
		final Jcube_struct cube = Jglobals.cube;// java

		int[] out_part_r = Jespresso.new_cube();
		final int[] s = Jespresso.new_cube();

		Jset_family BB = Jespresso.new_cover( R.count );
		BB.count = R.count;
		/* BB = get_blocking_matrix(R,c); */
		int[] data = R.data;// java
		for( int r = 0, i = 0; i < R.count; r += R.wsize, i++ ) {
			final int[] base = BB.data;// java
			final int b = BB.wsize * i;// GETSET(BB,i);
			final int last;
			if( (last = cube.inword) != -1 ) {
				/* Check the partial word of binary variables */
				int x = data[ r + last ] & c[ off_c + last];
				x = ~(x | x >>> 1) & cube.inmask;
				base[ b + last ] = data[ r + last ] & (x | x << 1);
				/* Check the full words of binary variables */
				for( int w = 1; w < last; w++ ) {
					x = data[ r + w ] & c[ off_c + w ];
					x = ~(x | x >>> 1) & Jespresso.DISJOINT;
					base[ b + w ] = data[ r + w ] & (x | x << 1);
				}
			}
			Jespresso.PUTLOOP( base, b, Jespresso.LOOP( data, r ) );
			Jset.set_and( base, b, base, b, cube.binary_mask );
			Jset.set_and( out_part_r, cube.mv_mask, data, r );
			if( ! Jset.setp_implies( out_part_r, c, off_c ) ) {
				Jset.set_or( base, b, base, b, out_part_r );
			}
		}
		out_part_r = null;
		BB = Junate.unate_compl( BB );

		Jset.set_copy( s, cube.emptyset );

		data = BB.data;// java
		for( int b = 0, i = 0; i < BB.count; b += BB.wsize, i++ ) {
			Jset.set_or( s, s, data, b );
		}
		Jespresso.free_cover( BB );
		set_not( s, 0 );
		return s;
	}

	/* set_not: flip 0 to 1 and 1 to 0 */
	static void set_not(final int[] c, final int off_c) {
		Jset.set_diff( c, off_c, Jglobals.cube.fullset, c, off_c );
	}

}