package espresso;

final class Jsignature_exact {

	/**
	 * signature_minimize_exact:
	 * What does it do: forms and solves the covering table whose rows are
	 *	essential signature cubes (ESCubes) and whose columns are
	 *	union of essential signature sets (ESSet)
	 * Input:
	 *	ESCubes: essential signature cubes
	 *	ESSet: union of essential signature sets
	 * Output:
	 *	COVER: exact cover
	 */
	static Jset_family signature_minimize_exact(final Jset_family ESCubes, final Jset_family ESSet) {
		Jsm_matrix table;

		/* number ESCubes, ESSet */
		int[] data = ESCubes.data;// java
		for( int p = 0, index = 0; index < ESCubes.count; p += ESCubes.wsize, index++ ) {
			Jespresso.PUTSIZE( data, p, index );
		}

		data = ESSet.data;// java
		for( int p = 0, index = 0; index < ESSet.count; p += ESSet.wsize, index++ ) {
			Jespresso.PUTSIZE( data, p, index );
		}

		/* form the covering table */
		{
			final long t = System.currentTimeMillis();
			table = signature_form_table( ESCubes, ESSet );
			Jutil_signature.s_totals( t, Jespresso.MINCOV_TIME );
		}

		/* solve the covering problem */
		final int[] weights = null; final boolean heur = false; final int level = 0;
		Jsm_row cover;
		{
			final long t = System.currentTimeMillis();
			cover = Jmincov.sm_minimum_cover( table, weights, heur, level );
			Jutil_signature.s_totals( t, Jespresso.MINCOV_TIME );
		}

		/* form the cover */
		Jset_family COVER = Jespresso.new_cover( 100 );
	    for( Jsm_element pe = cover.first_col; pe != null; pe = pe.next_col ) {
			COVER = Jset.sf_addset( COVER, ESSet.data, ESSet.wsize * pe.col_num );
		}

		Jmatrix.sm_free( table );
		Jrows.sm_row_free( cover );

		return COVER;
	}

	static Jsm_matrix signature_form_table(final Jset_family ESCubes, final Jset_family ESSet) {
		final Jsm_matrix table = Jmatrix.sm_alloc();

		final int[] data_cube = ESCubes.data;// java
		final int[] data_set = ESSet.data;// java
		for( int p = 0, column = 0; column < ESSet.count; p += ESSet.wsize, column++ ) {
			if( column % 1000 == 0 ) {
				Jdominate.sm_col_dominance( table, null );
			}
			for( int c = 0, row = 0; row < ESCubes.count; c += ESCubes.wsize, row++ ) {
				if( Jset.setp_implies( data_cube, c, data_set, p ) ) {
					Jmatrix.sm_insert( table, row, column );
				}
			}
		}
		Jdominate.sm_col_dominance( table, null );

		return table;
	}
}