package espresso;

import java.io.IOException;
import java.io.InputStream;

/**
 * The temporary class for some functions absent in Java
 */
final class Jsupport {
	/**
	 * Reads an integer value from an text input stream.
	 *
	 * @param fp the input stream to read.
	 * @param radix the radix, usually 10 or 16.
	 * @return the integer value.
	 * @throws IOException an error occurs while reading or parsing.
	 */
	static int scanInt(final InputStream fp, final int radix) throws IOException {
		final StringBuilder sb = new StringBuilder();
		int ch;
		do {
			ch = fp.read();
			if( ch < 0 ) {
				throw new IOException( "EOF" );
			}
		} while( Character.isWhitespace( ch ) );
		do {
			sb.append( (char)ch );
			ch = fp.read();
		} while( ch >= 0 && ! Character.isWhitespace( ch ) );
		try {
			return Integer.parseInt( sb.toString(), radix );
		} catch(final NumberFormatException ne) {
			throw new IOException( ne.getMessage() );
		}
		// return 0;
	}
}
