package espresso;

/**
 *  unate.c -- routines for dealing with unate functions
 */
final class Junate {
	static Jset_family map_cover_to_unate(final Jset[] T) {
		final Jcdata_struct cdata = Jglobals.cdata;// java

		final Jset_family A = Jset.sf_new( Jespresso.CUBELISTSIZE( T ), cdata.vars_unate );
		A.count = Jespresso.CUBELISTSIZE( T );

		final int[] data = A.data;// java
		for( int p = 0, i = 0; i < A.count; p += A.wsize, i++ ) {
			Jset.set_clear( data, p, A.sf_size );
		}
		int ncol = 0;

		for( int i = 0; i < Jglobals.cube.size; i++ ) {
			if( cdata.part_zeros[ i ] > 0 ) {
				// assert(ncol <= cdata.vars_unate);

				/* Copy a column from T to A */
				final int word_test = Jespresso.WHICH_WORD( i );
				final int bit_test = 1 << Jespresso.WHICH_BIT( i );
				final int word_set = Jespresso.WHICH_WORD( ncol );
				final int bit_set = 1 << Jespresso.WHICH_BIT( ncol );

				int pA = 0;// java A.data;
				Jset p;
				for( int T1 = 2; (p = T[ T1++ ]) != null; ) {
					if( (p.get( word_test ) & bit_test) == 0 ) {
						data[ pA + word_set ] |= bit_set;
					}
					pA += A.wsize;
				}
				ncol++;
			}
		}

		return A;
	}

	static Jset_family map_unate_to_cover(final Jset_family A) {
		final Jcube_struct cube = Jglobals.cube;// java
		final Jcdata_struct cdata = Jglobals.cdata;// java

		final Jset_family B = Jset.sf_new( A.count, cube.size );
		B.count = A.count;

		/* Find the unate variables */
		int[] unate = new int[ cube.num_vars ];
		int nunate = 0;
		for( int var = 0; var < cube.num_vars; var++ ) {
			if( cdata.is_unate[ var ] ) {
				unate[ nunate++ ] = var;
			}
		}

		/* Loop for each set of A */
		final int[] pB_base = B.data;
		int pB = 0;// java B.data;
		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {

			/* Initialize this set of B */
			Jset.set_fill( pB_base, pB, cube.size );

			/* Now loop for the unate variables; if the part is in A,
			* then this variable of B should be a single 1 in the unate
			* part.
			*/
			for( int ncol = 0; ncol < nunate; ncol++ ) {
				if( Jespresso.is_in_set( data, p, ncol ) ) {
					final int lp = cube.last_part[ unate[ ncol ] ];
					for( int i = cube.first_part[ unate[ ncol ] ]; i <= lp; i++ ) {
						if( cdata.part_zeros[ i ] == 0 ) {
							Jespresso.set_remove( pB_base, pB, i );
						}
					}
				}
			}
			pB += B.wsize;
		}

		unate = null;
		return B;
	}

	/* unate_compl */

	static Jset_family unate_compl(Jset_family A) {
		/* Make sure A is single-cube containment minimal */
		/*    A = sf_rev_contain(A);*/

		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			Jespresso.PUTSIZE( data, p, Jset.set_ord( data, p ) );
		}

		/* Recursively find the complement */
		A = unate_complement( A );

		/* Now, we can guarantee a minimal result by containing the result */
		A = Jcontain.sf_rev_contain( A );
		return A;
	}


	/**
	 *  Assume SIZE(p) records the size of each set
	 *
	 *  @param A disposes of A
	 */
	static Jset_family unate_complement(final Jset_family A) {
		Jset_family Abar;

		/* Check for no sets in the matrix -- complement is the universe */
		if( A.count == 0 ) {
			Jset.sf_free( A );
			Abar = Jset.sf_new( 1, A.sf_size );
			Jset.set_clear( Abar.data, Abar.wsize * Abar.count++, A.sf_size );

			/* Check for a single set in the maxtrix -- compute de Morgan complement */
		} else if( A.count == 1 ) {
			final int[] p = A.data;
			Abar = Jset.sf_new( A.sf_size, A.sf_size );
			for( int i = 0; i < A.sf_size; i++ ) {
				if( Jespresso.is_in_set( p, i ) ) {
					final int p1 = Abar.wsize * Abar.count++;
					Jset.set_clear( Abar.data, p1, A.sf_size );
					Jespresso.set_insert( Abar.data, p1, i );
				}
			}
			Jset.sf_free( A );

		} else {

			/* Select splitting variable as the variable which belongs to a set
			 * of the smallest size, and which has greatest column count
			 */
			int[] restrict = Jset.set_new( A.sf_size );
			int min_set_ord = A.sf_size + 1;
			int[] data = A.data;// java
			for( int p = 0, i = 0; i < A.count; p += A.wsize, i++ ) {
				if( Jespresso.SIZE( data, p ) < min_set_ord ) {
					Jset.set_copy( restrict, data, p );
					min_set_ord = Jespresso.SIZE( data, p );
				} else if( Jespresso.SIZE( data, p ) == min_set_ord ) {
					Jset.set_or( restrict, restrict, data, p );
				}
			}

			/* Check for no data (shouldn't happen ?) */
			if( min_set_ord == 0 ) {
				A.count = 0;
				Abar = A;

				/* Check for "essential" columns */
			} else if( min_set_ord == 1 ) {
				Abar = unate_complement( abs_covered_many( A, restrict ) );
				Jset.sf_free( A );
				data = Abar.data;// java
				for( int p = 0, i = 0; i < Abar.count; p += Abar.wsize, i++ ) {
					Jset.set_or( data, p, data, p, restrict );
				}

				/* else, recur as usual */
			} else {
				final int max_i = abs_select_restricted( A, restrict );

				/* Select those rows of A which are not covered by max_i,
				 * recursively find all minimal covers of these rows, and
				 * then add back in max_i
				 */
				Abar = unate_complement( abs_covered( A, max_i ) );
				data = Abar.data;// java
				for( int p = 0, i = 0; i < Abar.count; p += Abar.wsize, i++ ) {
					Jespresso.set_insert( data, p, max_i );
				}

				/* Now recur on A with all zero's on column max_i */
				data = A.data;// java
				for( int p = 0, i = 0; i < A.count; p += A.wsize, i++ ) {
					if( Jespresso.is_in_set( data, p, max_i ) ) {
						Jespresso.set_remove( data, p, max_i );
						final int j = Jespresso.SIZE( data, p ) - 1;
						Jespresso.PUTSIZE( data, p, j );
					}
				}

				Abar = Jset.sf_append( Abar, unate_complement( A ) );
			}
			restrict = null;
		}

		return Abar;
	}

	private static final class Jsuffice {
		Jset_family sf;
		int level;
	}

	@SuppressWarnings("boxing")
	static Jset_family exact_minimum_cover(Jset_family T) {
		final long start = System.currentTimeMillis();
		final Jsuffice stack[] = new Jsuffice[ 32 ];/* 32 suffices for 2 ** 32 cubes ! */

		if( T.count <= 0 ) {
			return Jset.sf_new( 1, T.sf_size );
		}
		int n, lev;
		for( n = T.count, lev = 0; n != 0; n >>= 1, lev++ ) {
			;
		}

		/* A simple heuristic ordering */
		T = Jcvrm.lex_sort( Jset.sf_save( T ) );

		/* Push a full set on the stack to get things started */
		n = 1;
		stack[0].sf = Jset.sf_new( 1, T.sf_size );
		stack[0].level = lev;
		Jset.set_fill( stack[0].sf.data, stack[0].sf.wsize * stack[0].sf.count++, T.sf_size );

		final int nlast = T.wsize * (T.count - 1);
		int[] data = T.data;// java
		for( int p = 0, last = p + T.count * T.wsize; p < last; p += T.wsize ) {

			/* "unstack" the set into a family */
			Jset_family temp = Jset.sf_new( Jset.set_ord( data, p ), T.sf_size );
			for( int i = 0; i < T.sf_size; i++ ) {
				if( Jespresso.is_in_set( data, p, i ) ) {
					final int p1 = temp.wsize * temp.count++;
					Jset.set_fill( temp.data, p1, T.sf_size );
					Jespresso.set_remove( temp.data, p1, i );
				}
			}
			stack[ n ].sf = temp;
			stack[ n++ ].level = lev;

			/* Pop the stack and perform (leveled) intersections */
			while( n > 1 && (stack[ n - 1 ].level == stack[ n - 2 ].level || p == nlast) ) {// java: p and nlast are pointers to T.data
				temp = unate_intersect( stack[ n - 1 ].sf, stack[ n - 2 ].sf, false );
				final int lvl = Math.min( stack[ n - 1 ].level, stack[ n - 2 ].level) - 1;
				if( (Jglobals.debug & Jespresso.MINCOV) != 0 && lvl < 10 ) {
					System.out.printf("# EXACT_MINCOV[%d]: %4d = %4d x %4d, time = %s\n",
							lvl, temp.count, stack[ n - 1 ].sf.count,
							stack[ n - 2 ].sf.count, Jespresso.print_time( System.currentTimeMillis() - start) );
					System.out.flush();
				}
				Jset.sf_free( stack[ n - 2 ].sf );
				Jset.sf_free( stack[ n - 1 ].sf );
				stack[ n - 2 ].sf = temp;
				stack[ n - 2 ].level = lvl;
				n--;
			}
		}

		final Jset_family temp = stack[0].sf;
		int[] p1 = Jset.set_new( T.sf_size );
		Jset.set_fill( p1, T.sf_size );
		data = temp.data;// java
		for( int p = 0, last = p + temp.count * temp.wsize; p < last; p += temp.wsize ) {
			Jset.set_diff( data, p, p1, data, p );
		}
		p1 = null;
		if( (Jglobals.debug & Jespresso.MINCOV1) != 0 ) {
			System.out.printf("MINCOV: family of all minimal coverings is\n");
			Jset.sf_print( temp );
		}
		Jset.sf_free( T );         /* this is the copy of T we made ... */
		return temp;
	}

	/** save 500 cubes before containment */
	private static final int MAGIC = 500;

	/**
	 *  unate_intersect -- intersect two unate covers
	 *
	 *  If largest_only is TRUE, then only the largest cube(s) are returned
	 */
	static Jset_family unate_intersect(final Jset_family A, final Jset_family B, final boolean largest_only) {
		int ord;

		/* How large should each temporary result cover be ? */
		Jset_family T = Jset.sf_new( MAGIC, A.sf_size );
		Jset_family Tsave = null;
		int maxord = 0;
		int[] pt_base = T.data;// java
		int pt = 0;// int[] pt = T.data;

		/* Form pairwise intersection of each set of A with each cube of B */
		final int[] a_data = A.data;// java
		for( int pi = 0, lasti = pi + A.count * A.wsize; pi < lasti; pi += A.wsize ) {

			final int[] b_data = B.data;// java
			for( int pj = 0, lastj = pj + B.count * B.wsize; pj < lastj; pj += B.wsize ) {

				boolean save = Jset.set_andp( pt_base, pt, a_data, pi, b_data, pj );

				/* Check if we want the largest only */
				if( save && largest_only ) {
					if( (ord = Jset.set_ord( pt_base, pt )) > maxord ) {
						/* discard Tsave and T */
						if( Tsave != null ) {
							Jset.sf_free( Tsave );
							Tsave = null;
						}
						pt_base = T.data;
						pt = 0;
						T.count = 0;
						/* Re-create pt (which was just thrown away) */
						Jset.set_and( pt_base, pt, a_data, pi, b_data, pj );
						maxord = ord;
					} else if( ord < maxord ) {
						save = false;
					}
				}

				if( save ) {
					if( ++T.count >= T.capacity ) {
						T = Jcontain.sf_contain( T );
						Tsave = (Tsave == null) ? T : Jcontain.sf_union( Tsave, T );
						T = Jset.sf_new( MAGIC, A.sf_size );
						pt_base = T.data;
						pt = 0;
					} else {
						pt += T.wsize;
					}
				}
			}
		}

		/* Contain the final result and merge it into Tsave */
		T = Jcontain.sf_contain( T );
		Tsave = (Tsave == null) ? T : Jcontain.sf_union( Tsave, T );

		return Tsave;
	}

	/**
	*  abs_covered -- after selecting a new column for the selected set,
	*  create a new matrix which is only those rows which are still uncovered
	*/
	private static Jset_family abs_covered(final Jset_family A, final int pick) {
		final Jset_family Aprime = Jset.sf_new( A.count, A.sf_size );
		final int[] pdest_base = Aprime.data;
		int pdest = 0;// java Aprime.data;
		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			if( ! Jespresso.is_in_set( data, p, pick ) ) {
				Jset.set_copy( pdest_base, pdest, data, p );
				Aprime.count++;
				pdest += Aprime.wsize;
			}
		}
		return Aprime;
	}


	/**
	*  abs_covered_many -- after selecting many columns for ther selected set,
	*  create a new matrix which is only those rows which are still uncovered
	*/
	private static Jset_family abs_covered_many(final Jset_family A, final int[] pick_set) {
		final Jset_family Aprime = Jset.sf_new( A.count, A.sf_size );
		final int[] pdest_base = Aprime.data;
		int pdest = 0;// java Aprime.data;
		final int[] data = A.data;// java
		for( int p = 0, last = p + A.count * A.wsize; p < last; p += A.wsize ) {
			if( Jset.setp_disjoint( data, p, pick_set ) ) {
				Jset.set_copy( pdest_base, pdest, data, p );
				Aprime.count++;
				pdest += Aprime.wsize;
			}
		}
		return Aprime;
	}

	/**
	*  abs_select_restricted -- select the column of maximum column count which
	*  also belongs to the set "restrict"; weight each column of a set as
	*  1 / (set_ord(p) - 1).
	*/
	private static int abs_select_restricted(final Jset_family A, final int[] restrict) {
		/* Sum the elements in these columns */
		int[] count = Jset.sf_count_restricted( A, restrict );

		/* Find which variable has maximum weight */
		int best_var = -1;
		int best_count = 0;
		for( int i = 0; i < A.sf_size; i++ ) {
			if( count[i] > best_count ) {
				best_var = i;
				best_count = count[ i ];
			}
		}
		count = null;

		if( best_var == -1 ) {
			Jcvrmisc.fatal("abs_select_restricted: should not have best_var == -1");
		}

		return best_var;
	}
}