package espresso;

import java.util.Comparator;

/** Module:util_signature.c
 * Purpose:
 *	contains miscellaneous utility routines
 * Routines:
 * void set_time_limit(seconds)
 *	sets the cpu time limit
 * int sf_equal():
 *	 checks equlaity of two set families.
 * int print_cover():
 *	prints cover.
 * int mem_usage():
 *	current Memory usage.
 *	Initialized on the first call.
 * int time_usage():
 *	current time usage.
 *	Initialized on the first call.
 */
final class Jutil_signature {
	/**
	 * TODO rewrite set_time_limit in Java
	 *
	 * @param seconds
	 */
	static void set_time_limit(final int seconds) {
		// final struct rlimit rlp_st, *rlp = &rlp_st;
		// rlp.rlim_cur = seconds;
		// setrlimit( RLIMIT_CPU, rlp );
	}

	@SuppressWarnings("boxing")
	static void print_cover(final Jset_family F, final String name) {
		System.out.printf("%s:\t %d\n", name, F.count );
		final int[] data = F.data;// java
		for( int p = 0, last = p + F.count * F.wsize; p < last; p += F.wsize ) {
			Jcvrout.print_cube( System.out, data, p, "~0" );
		}
		System.out.printf("\n\n");
	}

	/** sf_equal: Check equality of two set families */
	static boolean sf_equal(final Jset_family F1, final Jset_family F2) {
		final int count = F1.count;

		if( F1.count != F2.count ) {
			return false;
		}

		final Comparator<Jset> descend = new Jdescend_order();
		final Jset[] list1 = Jcontain.sf_sort( F1, descend );
		final Jset[] list2 = Jcontain.sf_sort( F2, descend );

		for( int i = 0; i < count; i++ ) {
			final Jset s1 = list1[ i ];
			final Jset s2 = list2[ i ];
			if( ! Jset.setp_equal( s1.base, s1.offset, s2.base, s2.offset ) ) {
				return false;
			}
		}

		return true;
	}

	private static long memory_init;
	private static boolean memory_init_flag = true;
	/** mem_usage:
	 * 	Initialized on first call. Prints current memory usage.
	 */
	@SuppressWarnings("boxing")
	static long mem_usage(final String name) {
		final Runtime r = Runtime.getRuntime();

		if( memory_init_flag ) {
			memory_init = r.totalMemory() - r.freeMemory();
			memory_init_flag = false;
		}

		final long memory_current = r.totalMemory() - r.freeMemory();

		System.out.printf("Memory %s\t %d\n", name, memory_current - memory_init );

		return memory_current;
	}

	private static long time_init;
	private static boolean time_usage_flag = true;

	/** time_usage:
	 * 	Initialized on first call. Prints current time usage.
	 */
	@SuppressWarnings("boxing")
	static long time_usage(final String name) {
		if( time_usage_flag ) {
			time_init = System.currentTimeMillis();
			time_usage_flag = false;
			return time_init;
		}

		final long time_current = System.currentTimeMillis();

		System.out.printf("%s\t %d\n", name, (time_current - time_init) / 1000);

		return time_current;
	}

	/** s_totals : add time spent in the function and update call count */
	static void s_totals(long time, final int i) {
		time = System.currentTimeMillis() - time;
		Jglobals.total_time[ i ] += time;
		Jglobals.total_calls[ i ]++;
	}

	@SuppressWarnings("boxing")
	static void s_runtime(final long total) {
		for( int i = 0; i < Jespresso.TIME_COUNT; i++ ) {
			if( Jglobals.total_calls[ i ] != 0 ) {
				final long temp = 100 * Jglobals.total_time[ i ];
				System.out.printf("# %s\t%2d call(s) for %s ( %2d.%01d%% )\n",
						Jglobals.total_name[ i ], Jglobals.total_calls[ i ], Jespresso.print_time( Jglobals.total_time[ i ] ),
						temp / total, (10 * (temp % total)) / total);
			}
		}
	}
}